/*
   SNCFTMSHELL SAS MACRO

    Author: Sally Picciotto <spicciot@hsph.harvard.edu>.

      Initial Version:     18 APRIL 2008
      Current Version:     15 JUNE 2010. 

        Estimate the probability of a cumulative failure time 
        dichotomous outcome under interventions using g-estimation
        and a Structural Nested Cumulative Failure Time Model.

        This is a shell that calls macro by Roger Logan to do the actual
        g-estimation, uses the estimated parameters to calculate risks
        and survival curves under hypothetical interventions, and then 
        gives output in the same format as Sarah Taubman's gformula and 
        msm macros.

    LOOSELY BASED ON MSM MACRO BY Sarah Taubman
         which was based on her GFORMULA macro and had
         some functionalities added by Roger Logan
    AND CALLING G-ESTIMATION MACRO by Roger Logan 
    (relevant part extracted below, 
    starting at %macro cumft, and including everything below that)
    (NOTE that changes from g-estimation macro by Roger Logan include 
    functionalities added for other blip functions, corrected censoring 
    weights, and possibly other changes, so replace/update with caution! 
    Extract has fewer functionalities than original: double-robustness 
    and efficient estimator options have been removed, among other things.)
                 
THIS VERSION IS WITHOUT THE GENPRED ETC. SUBMACROS: THE USER MUST PROVIDE THE 
LISTS OF COVARIATES TO BE ADJUSTED FOR, AND THE DATASET SHOULD CONTAIN WHATEVER
THOSE VARIABLES ARE. THUS, (UNLIKE IN THE G-FORMULA MACRO) THE USER WILL HAVE TO 
CREATE ANY TRANSFORMATIONS OF THE VARIABLES (INDICATORS FOR CATEGORIES, SPLINES,
ETC.) AND ANY INTERACTION TERMS DESIRED. 

THIS VERSION IS THE CURRENT "FINAL DRAFT". 

NOTE: OBSERVED RISK HERE USES GOOLEY METHOD, WHILE NATURAL COURSE AND 
INTERVENTIONS DO NOT BECAUSE WORKING IN PSEUDOPOPULATION WHERE THERE IS
NO CENSORING BY DEATH FROM OTHER CAUSES.

*/                                
    
%macro sncftmshell(
    data=,             /* dataset for analysis */
    id=,           /* unique identifier for subjects */

    time=,             /* time point */
    timepoints=,       /* number of time points */
    mintime=0,          /*minimum time point value (either 0 or 1)*/

    outc=,         /* outcome variable: 1 event, 0 no event */

    censdead=,      /* death from other causes variable: 1=censored, 0=uncensored */
    censlost=,      /* loss to follow up variable: 1=censored, 0=uncensored */
    censpred=,      /* full list of covariates to include in the model(s) for censoring*/

    treat1=,            /* treatment variable: 1 treated, 0 untreated */
    treat1skip=-1,      /* times when treatment variable not measured*/ 
    treat1class =  ,    /* class variable for separate treatment models */
    treat1pred=,        /* full list of covariates to include in the model for treatment.*/   
    regimen=,           /* treatment regime to blip up to; default is 1,1,1,1,... 
                           if specified, should be a list of &timepoints values, separated by spaces*/ 

    intlabel1=,         /* label for the "always treated" intervention (blipped up) */
    intlabel2=,         /* label for the "never treated" intervention (blipped down) */
        

    blipfunction=1,     /* There are several options for the blip function. 
                           Option 4 is now a more general version of option 5.
                            1 = exp[blip]=1+\frac{exp[psi1 A_m]-1}{k-m} (consistent with certain SNAFTM)
                            2 = exp[blip]=exp[psi A_m]
                            3 = exp[blip]=1+\frac{exp[psi1 A_m + psi2 A_m A_{m-1}]-1}{k-m}
                            4 = exp[blip]=1+\frac{exp[psi1 A_m + psi2 A_m X_0]-1}{k-m}
                            5 = exp[blip]=1+\frac{exp[psi1 A_m + psi2 A_m A_{-1}]-1}{k-m}
                            6 = exp[blip]=1+\frac{exp[psi1 A_m + psi2 A_m X_m]-1}{k-m}
                           Note: in blipfunction=4, must specify the name of the baseline effect
                            modifier X_0 (under parameter blem). 
                            In practice, the program will rename this to &treat1.preb
                            and proceed as if user had specified blipfunction=5.
                           Note: in blipfunction=5, A_{-1} must be provided in dataset by user
                            under the name &treat1.preb 
                           Note: in blipfunction=6, must specify the name of the time-varying effect 
                            modifier X_m (under paramter tvem); also, note that blipup will be set to no
                            since there is no closed-form way to blip up in this case.*/
    blem=,              /* only for blipfunction=4 */
    tvem=,              /* only for blipfunction=6 */

    psionly=0,          /* When psionly=1, program will calculate psi and then stop without blipping down and up.*/

    blipup=yes,         /* only for psionly=0: 
                            blipping up is only necessary to compare two different regimes, not for
                            for comparing always treat to natural course. If set to "yes", will 
                            blip down and then up, as originally programmed. If set to "no", will
                            only blip down: this is the preferred method if you want to know the 
                            effect of: 
                                (a) "never treat" compared to natural course (e.g. harmful exposure), or of 
                                (b) "always treat" compared to natural course if you have coded your 
                                    treatment variable to 0 for treated and 1 for untreated. 
                            [Option "no" still being developed.]*/
       
    dir=,               /* path of directory in which to save datasets permanently*/
    intervs_all=,       /* dataset to store psi estimates and the expected cumulative counterfactuals 
                            at each time under no treatment and under interventions*/
    weightdata=svwts,   /* dataset to store weights */
    resultsdata=,       /* dataset to store results */

    nointervdata=refdata,      /* base name to store datasets containing reference results--
                                    macro will add index 0: no intervention
                                    macro will add index 2: "never treat" (only when blipup=1)*/


    seed= 7834,     /* random numbers seed */
        
    nsamples=0,     /* number of sets of parameter estimates (default is 0: will not do bootstraps) */
    
    parallel=,      /*if running bootstraps in parallel, this gives the suffix to add to the 
                        end of the permanent datasets so they don't overwrite each other--
                        should be 1 through the total number of parallel runs*/
 
    outputs = yes,      /* whether to print results */

    survcurv = ,        /*name of pdf file in which to save survival plot: give full path but do not give suffix of .pdf*/
    survaxis = ,        /*for order option in axis statement for %survplot: 
                            allows user to define the vertical scale of survival plot
                            Format is: min to 1.000 by delta */

/** PARAMETERS FOR G-ESTIMATION **/
    srchbnd=3.0,        /* search band for optimization procedure in sncftm.*/
    starting_point=,    /* starting point for grid search and/or nelder-meade optimization procedure in sncftm.
                            If not specified, program assumes to start at 0 (or (0,0)).*/
    est_method = 2,     /** estimation method
                                           0 = no estimation done 
                                           1 = Nelder-Mead Simplex method
                                           2 = Newton's method
                                           3 = Simulated Annealing  
                                           4 = for one psi, use bisection method to find root of estimating equation 
                                         **/
    save_graph = 0,     /*whether to save graph of test statistic vs. psi*/
    graph_path = ,      /*path and file name to save graph under; must be specified if save_graph=1*/
    graph_path2= ,      /*only when two psis, path and file name to save contour graph under--
                            if not specified, contour graph will overwrite the 3D graph and 
                            be saved under the name given in graph_path. */
    );


    /*****************************************************************************/
    /* First we check whether a few essential parameters are correctly specified */
    /*****************************************************************************/
    %*Producing error message and exiting if no dataset is specified;
    %if %bquote(&data)= %then %do;
        %put ERROR: The DATA= argument must be specified.;
        %goto exit;
        %end;
    %put;   
    %*Producing error message and exiting if there will be an infinite loop in %lifeafterdeath below;
    %let maxtime=%eval(&timepoints-1+&mintime);
    data &data.temp;
        set &data;
        if _n_=1 then do;
            if &time ne . and &time<=&maxtime then call symput('runaway',trim(left(put("0",1.))));
            end;
        if &time=. or &time>&maxtime then do;
                                            call symput('runaway',trim(left(put("1",1.))));
                                            stop;
                                            end;
        run;
    %if %eval(&runaway)=1 %then %do;
    %put ERROR: Procedure cannot handle values of &time that are missing or larger than maximum time specified=&maxtime. --Program terminated.;
    %goto exit;
    %end;
    %*Producing error message and exiting if mintime specified by user is not 0 or 1;
    %if %eval(&mintime)^=0 and %eval(&mintime)^=1 %then %do;
        %put ERROR: Parameter mintime must be equal to 0 or 1. Program terminated.;
        %goto exit;
        %end;

/*Setting the directory in which to save datasets, if specified.*/
%global megan;
    %if %bquote(&dir)^= %then %do;
        %let megan=permadir.;
        libname permadir "&dir";
        %end;
    %else %do;
        %let megan=;
        %end;

   %*Setting local macro variables; 

    %local mint maxtime ssize obsp dimoutc dimcens dimcensw ninteract outcpred censnumpred /*which will be empty*/
            cens cens2 cumftparams save_weights ncens_var whereforlad boundarypsi pg arnold numint;

/*macro variables that are helpful when debugging-- can then enclose the problematic parts of
macro between &myoptions and &myoptions 2 to find out what macro is doing.*/
%let myoptions=options mlogic mprint symbolgen notes;
%let myoptions2=options nomlogic nomprint nosymbolgen nonotes;

/*initialize macro variables*/
%let boundarypsi=0;     /*this will change, keeping count of samples whose psi is on boundary*/
%let timekeeper=&time;                                  /*because name changes in %lifeafterdeath;*/


    %*Preparing data;
    %gestdataprep;

    %*Looping over bootstrap samples;
    %do boosample = 0 %to %eval(&nsamples);
       %let time=&timekeeper;                           /*because name changes in %lifeafterdeath, but the new samples*/
                                                        /*should start over with original variable name.*/

       %*Generating samples;                            
       %sncftmsamples;
       %if &psionly=0 %then %do;
           %*Generating weights;
           %weights;
           %end;
        
       %*Running interventions;


           %*Fitting structural models;
                %put Finding parameter for &outc using g-estimation;
                %put;
                %cumft(&cumftparams, 
                         /** VARIABLES INCLUDED IN THE MODEL OF INTEREST **/

                        outcome = &outc ,                /* outcome variable */         
                        treat = &treat1 ,                  /* treatment variable */
                        covA  = &treat1pred,
                        byA = &treat1class, /*NOTE: SALLY ADDED THIS PARAMETER!*/
                        cens = &cens,
                        covC = &censpred,                       
                        cens2= &cens2,
                        covC2= &censpred,
                        save_weights=&save_weights,
                        plot_graph = &pg,             /* plot results of grid search in one or two dimensions */
                        extra=SNCFTM for &outc -- Treatment &treat1 -- Sample &boosample
                        ); 

                title;

            /*THE OUTPUT PSI WILL BE IN A DATASET CALLED _SNMM1_ ...*/
            %global npsi;
            %if &blipfunction=1 or &blipfunction=2 %then %do;
                %let npsi=1;
                %end;
            %else %let npsi=2;; 
            %do iiii=1 %to &npsi;   
                %global psi&iiii;
                %end;

            data _snmm1_&outc;
                set _snmm1_ (drop=_bsample_ sample);
                %do iiii=1 %to &npsi;
                    call symput("psi&iiii",trim(left(put(psi&iiii,best12.))));
                    %end;
                _boosample_=%eval(&boosample);
                %do tloop=%eval(&mintime) %to %eval(&maxtime);
                    nonmono&tloop=0;
                    %end;
                run;

            %if &boosample=0 %then %do; 

                %do iiii=1 %to &npsi;
                    %global psi&iiii.pe;
                    %let psi&iiii.pe=&&psi&iiii; 
                    %end;/*npsi*/
                %end;/*boosample=0*/


            title;

        %if &psionly=0 %then %do;
            /*Checking whether psi estimate(s) on boundary, 
            and blipping down and up if estimate(s) valid.*/
            %processpsi;
            %if &arnold=1 %then %goto exit;;/*Terminator in case of invalid estimate in boosample 0*/


            %if &boosample = 0 %then %do;
                data &megan._snmm1_&outc._all&parallel;
                    set _snmm1_&outc;
                    run;
                %end;
            %else %do;
                proc append base=&megan._snmm1_&outc._all&parallel data=_snmm1_&outc;
                    run;
                %end;
           %end;    /*&psionly=0*/

        %*Deleting no longer needed datasets;
        proc datasets nolist library=work;
            delete psiprob psiprob2 /*msmparam*/  
                    %if &boosample>0 %then %do;
                    weights&boosample %end;
                    %if &psionly = 0 %then %do; 
                    /*blipdown*/ meany0 meany %end;
                    /*and Roger's datasets*/ _snmm1_
                    for_repeated ycount _cero_ _eligcount_ _goodcnt_ _just_y_ _tcount_ ycount temp tmp0  ;
            quit;

        data &megan.bdypsi&parallel;
            bdypsi=&boundarypsi;
            run;

        %*Changing seed for next run;
        %let seed = %eval(&seed+3);
    %end; /*bsample loop*/

    
    %*Outputting results;
    %if &nointerv = 0 and &psionly=0 %then %do;
        title ;
        %if &blipup=yes %then %do;
            %do evens=0 %to 2 %by 2;
                %sncftmresults(&evens);
                %end;
            %end;
        %else %do;
            %sncftmresults(0);
            %end;
        %cleanup;
        title ;
        %survplot;
        %end;


    %exit:;


    %mend sncftmshell;


%macro gestdataprep;

%if &blipfunction=6 %then %let blipup=no;

%if &blipup=yes %then %let numint=2;
                %else %let numint=1;

/*just to make sure, because a specified value for &blem or &tvem will be
how we retrace the original value of &blipfunction: all effect modification
models are treated as blipfunction=5 by the macro.*/
%if &blipfunction ne 4 %then %let blem=;    
%if &blipfunction ne 6 %then %let tvem=;

/*obsolete parameters whose values still get checked at some point in program*/ 
%global nointerv lagsetto0at0 bliptype treat1otype ;
    %let nointerv=0;        
    %let lagsetto0at0=1;
    %let bliptype=0;

/*so as to avoid problems when censoring does not occur in data*/
    %if %bquote(&censdead)^= | %bquote(&censlost)^= %then %let nocens=0;
                                                    %else %let nocens=1;

    %************ARRAYING REGRESSION PREDICTORS;


    %if &nocens=0 %then %do;
            %put Censoring denominator predictors are &censpred; 
        %put ;
        %end;/*censoring weights needed*/
    
    %else /*no weights needed*/ %do;
        %let censpred2= ;
        %end;

    %if &blipfunction ne 2 %then %do;
        %let outcpred = &time ;
        %end;
    %else %do;
        %let outcpred = ;
        %end;    

    %let t_index = 1 ;
    %let outcpred = &outcpred &&treat&t_index; 
    %if &blipfunction = 3 %then %do;
       %let outcpred = &outcpred &&treat&t_index.._l1;
       %end;
    %if &blipfunction = 5 %then %do;
        %let outcpred = &outcpred &&treat&t_index..preb;
        %end;

    %if &blipfunction ne 4 %then %do;
        %put Outcome predictors in structural nested cumulative failure time model are &outcpred ;
        %put ; 
        %end;

    %if &blipfunction = 4 %then %do; /*we will rename &blem, but this should be invisible to the user*/
        %put Outcome predictors in structural nested cumulative failure time model are &outcpred &blem;
        %put ; 
        %let outcpred = &outcpred &&treat&t_index..preb;
        %end;
    %if &blipfunction = 6 %then %do; /*we will rename &tvem, but this should be invisible to the user*/
        %put Outcome predictors in structural nested cumulative failure time model are &outcpred &tvem;
        %put ; 
        %let outcpred = &outcpred &&treat&t_index..preb;
        %end;



        /*Predictors of the treatment variable to send to g-estimation macro.*/


    %unique(&treat1pred);
    %let treat1pred = &unique;
            
    %put Treatment &treat1 predictors are &treat1pred;
    %put ;
           
    %*********** PREPARING DATA;
    
    %* Creating data set _msmparamdata_ sorted by id and time;
    %* Creating data set indiv with one record per id;
    proc sort data=&data out=_msmparamdata_; by &id descending &time; run;    
    proc sort nodupkey out=indiv; by &id; run;
        
    %* Calculating obs pr of outcome (by time points);    
    %obscuminc(data=&data,time=&time,timepoints=&timepoints,event=&outc,censdead=&censdead,censlost=&censlost);
 
    %*Finding min and max observed times;                
    proc means data=&data min max noprint;
        var &time;
        output out=minmax (keep= mint maxt) min=mint max=maxt;
    run;

        %if &blipfunction=5 or &blipfunction=4 /*or &strat=1*/ %then %do;
        %global prevalence;
        proc means data=indiv mean noprint; 
            var %if &blipfunction=5 %then %do;
                    &treat1.preb; 
                    %end;
                %else %do;
                    &blem;
                    %end;
            output out=preval (keep=prevale) mean=prevale;
            run;
        data preval;
            set preval;
            call symput('prevalence',trim(left(put(prevale,best12.))) );
            run;
        %put Prevalence of baseline effect modifier=&prevalence;
        %end;




     %*Generating _orig_ dataset with ID number, a new sequential ID and sample (=0);
     %*Also generating macros for obsp, ssize, min, max, and dim arrays;
    data _orig_ (keep= _sample_ newid &id);
        merge indiv cuminc minmax end=_end_;

        newid=_n_;
        _sample_=0;
        
        if _n_=1 then do;
                    
            array aoutc &outcpred;
            dimoutc=dim(aoutc);
            call symput('dimoutc',trim(left(put(dimoutc,8.))) );


            %if %bquote(&censlost)^= or %bquote(&censdead)^= %then %do; 
                array acensw &censpred;
                dimcensw=dim(acensw);
                call symput('dimcensw',trim(left(put(dimcensw,8.))) );
            %end;
            call symput('obsp',cuminc);
            call symput('mint',mint);
            call symput('maxt',maxt);
            
            end;
            
        if _end_ then do;
            call symput('ssize',trim(left(put(newid,8.))) );
            end;
    run;


    proc sort; by &id; run;

    %* Adds the sequential ID and sample (=0) to the _msmparamdata_ data;
    data _msmparamdata_;
        merge _orig_ _msmparamdata_;
        by &id;
    proc sort; by newid &time; run;

    %* Adding necessary variables for models;
    data _msmparamdata_;
        set _msmparamdata_;

        &time._l1 = &time-1;
        &time._l2 = &time-2;
        &time._l3 = &time-3;

    /*if blipfunction is 4, it's essentially the same idea as blipfunction=5 but the variable
    name is different; we'll just give it the same name and switch over.*/
        %if &blipfunction=4 %then %do;
            &treat1.preb=&blem;
            %end;
        %if &blipfunction=6 %then %do;
            &treat1.preb=&tvem;
            %end;


        run;

        /*we have changed the name of the baseline effect modifier to the name we use for prebaseline
            treatment when blip function=5, so from now on we will just use blipfunction=5.*/
    %if &blipfunction=4 or &blipfunction=6 %then %do;
        %let blipfunction=5;
        %end;


        /*Preparing parameters for calling g-estimation macro*/
    %if &blipfunction=3 %then %do;
        %let psivar=&treat1._l1;
        %end;/*blipfunction=3*/
    /*%else %do;
        %if &blipfunction=4 %then %do;
            %let psivar=&blem;
            %end;/*blipfunction=4 but actually before this we will have transformed to blipfunction=5.*/
        %else %do;
            %if &blipfunction=5 %then %do;
                %let psivar=&treat1.preb;
                %end;/*blipfunction=5*/
            %else %do;
                %let psivar=;
                %end;/*blipfunction ne 3,4,5*/
            /*%end;/*blipfunction ne 3,4*/
        %end;/*blipfunction ne 3*/
    %let cumftparams=data=msmparam,
                         /** VARIABLES INCLUDED IN THE MODEL OF INTEREST **/

                        id =  uniqueid ,                            /*careful, this will depend on what dataset used*/
                        time =  &time ,
                        eligible= , 
                        psivar= &psivar,        /*other factor (besides A_m) in product term in structural model*/
 
                         /*** DATA ANALYSIS SPECIFICATIONS **/

                        dr = 0 ,
                        conf_interval = 0 ,
                        starting_point = &starting_point,       /* starting point for nelder mead simplex and center for gridsearch region */
                        user_range =   ,         /* user defined range for gridsearch and optimization region */
                        srchbnd= &srchbnd ,         /* used for creating interval constraints in the optimization routines for multiplicative model */
                        grid_search = 1,           /** use grid search **/
                        pts1 = 100,                /* number of points used in grid search */
                        graph_path = &graph_path,   /*path and file name to save graph under; must be specified if save_graph=1*/
                        gdevice = pdf , 
                        ftype = pdf,
                        countmax = 100,           /* maximum number of steps to find rough estimates for confidence bounds */

                        do_graph = 0,              /** will be used for plotting graphs **/
                        save_graph = &save_graph,   /* if plot_graph = 1 and save_graph = 1 then output graph in some format
                                                        given by variables  gdevice and ftype .
                                                        default type will be pdf */

                        est_method = &est_method,            /** estimation method
                                           0 = no estimation done 
                                           1 = Nelder-Mead Simplex method
                                           2 = Newton's method
                                           3 = Simulated Annealing  
                                           4 = for one psi, use bisection method to find root of estimating equation 
                                         **/
                        optn_level = 0,             /* level of output for the iml optimization routines 0 = none 5 = all */

                        bootstrap = 0,
                        nboot = 50, 
                 
                        zero_one = 1,
                        failure_time = 1,

                        debug = 0;

    %let save_weights=1;
    %if %bquote(&censlost)^= %then %do;
            %let cens=&censlost;
            %if %bquote(&censdead)^= %then %do;
                %let cens2=&censdead;
                %let ncens_var=2;
                /*%let whereforlad=(&censlost=0 and &censdead=0);*/
                %end;
            %else %do;
                %let cens2=;
                %let ncens_var=1;
                /*%let whereforlad=(&censlost=0);*/
                %end;
            %end;

    %else %do;
            %let cens2=;
            %if %bquote(&censdead)^= %then %do;
                %let cens=&censdead;
                %let ncens_var=1;
                /*%let whereforlad=(&censdead=0);*/
                %end;
            %else %do;
                %let cens=;
                %let ncens_var=0;
                /*%let whereforlad=(1=1);*/
                %let save_weights=0;
                %end;
            %end;



    %* Deleting no longer needed datasets;
    proc datasets nolist library=work; 
    delete /*cuminc*/ minmax indiv _orig_ &data.temp preval;
    quit;
    
    %mend gestdataprep;


%macro sncftmsamples;
   
    %************** CREATING BOOTSTRAP SAMPLES;

        %if %eval(&boosample) = 0 %then %do;

            %*Base data for estimation and simulation (sample 0);
            
            data msmparam;
                set _msmparamdata_;
                _sample_ = 0;
                uniqueid=newid;
                keep  _sample_ newid uniqueid &time 
                    &censpred 
                    %do n = 1 %to 1;
                        &&treat&n &&treat&n.pred &&treat&n.class
                        %end;
                     &outc &outcpred %if &blipfunction=3 %then %do; &treat1.preb %end;
                     &censlost &censdead 
                     ;
            run;
            

            /*FOR %CUMFT CALL--whether to plot graph of estimating equation*/
            %let pg=1;
            
            %end;

        %else %do;
            /*FOR %CUMFT CALL--whether to plot graph of estimating equation*/
            %let pg=0;

            %put   ;
            %put   ;
            %put Creating bootstrap sample &boosample;
            %put   ;

        %*Generating random sample;
            data msmparamsample (keep = _sample_ newid uniqueid &time);
                call streaminit(&seed);
                do uniqueid=1 to &ssize;
                    U = rand('uniform');
                    newid=int(U*&ssize)+1;
                    _sample_ = &boosample;
                    do &time= &mintime to &maxtime;
                        output msmparamsample;
                        end;
                    end;
            run;

        %*Merging with parameter estimation data;
        proc sort data = msmparamsample;
            by newid &time;
        run;

        data msmparam;
            merge _msmparamdata_ (in=p) msmparamsample (in=ps);
            by newid &time;
            if p ne 0 and ps=1 and _sample_ ne 0;
            keep _sample_  newid uniqueid &time 
                &censpred 
                %do n = 1 %to 1/*&ntreat*/;
                    &&treat&n  &&treat&n.pred &&treat&n.class
                    %end;
                %if &blipfunction=3 %then %do; &treat1.preb %end;
                &outc &outcpred
                &censlost &censdead ;
        run;



            %*Cleaning up no longer needed datasets;
            proc datasets nolist library=work;
                delete msmparamsample ;
            quit;
            
            %end;

    %mend sncftmsamples;

%macro weights;
    %local printop;

    %let printop=noprint;
    
    proc sort data=msmparam;
        by uniqueid &time;
    run;
    
   %*** ESTIMATING WEIGHTS FOR IPW ANALYSIS;
    
    %* Censoring;
    %if %bquote(&censlost)^= %then %do;
        proc logistic data=msmparam noprint;
            model &censlost = &censnumpred;
            output out=clweight_0 p=p_uncensl_0;
        run;

        proc logistic data=clweight_0 (drop=_level_) &printop;
            model &censlost =  &censpred;
            output out=clweight_k p=p_uncensl_k;
        run;
        %end;

    %else %do;
        data clweight_k;
            set msmparam(keep=uniqueid &time);
            p_uncensl_0=1;
            p_uncensl_k=1;
        run;
        %end;    

    %if %bquote(&censdead)^= %then %do;
      proc logistic data=msmparam noprint;
            model &censdead = &censnumpred;
            output out=cdweight_0 p=p_uncensd_0;
        run;

        proc logistic data=cdweight_0 (drop=_level_) &printop;
            model &censdead =  &censpred;
            output out=cdweight_k p=p_uncensd_k;
        run;
        %end;

    %else %do;
        data cdweight_k;
            set msmparam(keep=uniqueid &time);
            p_uncensd_0=1;
            p_uncensd_k=1;
        run;
        %end;    

    
    %*** Creating pseudo-population;
    
    data weighted;
        merge msmparam
            clweight_k (keep=uniqueid &time p_uncensl_k)
            cdweight_k (keep=uniqueid &time p_uncensd_k)
            ;
        
        by uniqueid &time;

        retain clwght_k cdwght_k;
        if first.uniqueid then do; 
            clwght_k = 1;         
            cdwght_k = 1; 
            end;

        clwght_k = clwght_k * p_uncensl_k;

        cdwght_k = cdwght_k * p_uncensd_k;
                
        unstabw_censd =1/(cdwght_k);
        unstabw_censl =1/(clwght_k);

        unstabw=1/(cdwght_k*clwght_k);

     /*   %if &trim = yes %then %do;
           if stabw > 50 then stabw = 50;
           if stabw < 0.02 then stabw = 0.02;
           %end; /*we use only unstabilized weights. option to trim weights not implemented.*/

        ipcw=unstabw;
       
    run;
    
 
    %*Summarizing weights;    
    proc univariate data=weighted noprint;
        var ipcw unstabw_censd unstabw_censl;
        output out = sampleweights
            mean = unstabw_mean cd_wgt_mean cl_wgt_mean 
            max =  unstabw_max  cd_wgt_max  cl_wgt_max 
            min =  unstabw_min  cd_wgt_min  cl_wgt_min 
            ;
    run;

    data sampleweights;
        set sampleweights;
        _sample_ = &boosample;
        attrib
            unstabw_mean 
            unstabw_max  
            unstabw_min  
           label = ' '; 
    run;

    proc append base=_weights_ data=sampleweights;
    run;

    %if &outputs = yes %then %do;        
        %if &boosample = &nsamples %then %do;
            proc means data=_weights_;
            run;
            %end;
        %end;

    %*Outputting primary data weights;
    %if &boosample = 0 %then %do;
       %put Saving original dataset weights to &weightdata;
       data &weightdata;
       %end;       
    %else %do;
       %let weightdata=weights&boosample;
       data &weightdata;
       %end;
          set weighted (keep = uniqueid &time ipcw );
          run;
          
    %*Cleaning up no longer needed datasets;
    %if %bquote(&censdead)^= or %bquote(&censlost)^= %then %do;
    proc datasets nolist library=work;
         delete 
            %if %bquote(&censdead)^= %then %do; 
                cdweight_k  %end;
            %if %bquote(&censlost)^= %then %do;
                clweight_k  %end;
            ;
        quit;%end;
            
%mend weights;
%macro processpsi;
/*WE ALSO DON'T WANT TO USE PSI ESTIMATES WHERE THE ALGORITHM DID NOT CONVERGE.*/
    /*SO FIRST CHECK IF PSI DOES NOT EXIST OR IS ON BOUNDARY*/
        /* IF THERE ARE TWO PSIS, boundary is any edge of rectangle */
            %let arnold=0;  /*This is the "Terminator" to send to "exit" when =1.*/
            %if %bquote(&psi1)= %then %do;
                %if &boosample=0 %then %do;
                    %put ERROR: g-estimation failed. Program will exit.;
                    %let arnold=1;/*%goto exit;*/
                    %end;
                %else %do;
                    %psionboundary;
                    %end;
                %end;
            %else %do;

                %local x y;
                %if %bquote(&starting_point)= %then %do;
                    %let x=0.0;
                    %if &npsi=2 %then %let y=0.0;;
                    %end;
                %else %do;
                    %let x=%scan(&starting_point,1," ");
                    %let y=%scan(&starting_point,2," ");/*will be blank if only one word in &starting_point*/
                    %end;

                %if %sysevalf(&psi1>&x) %then %do;
                    %if %sysevalf(&x+&srchbnd-&psi1<1.e-11) %then %do;
                        %if &boosample=0 %then %do;
                            %put ERROR: Optimization did not converge. Program will exit.; 
                            %put You might try specifying a different value of parameter SRCHBND and/or STARTING_POINT.;
                            %let arnold=1;/*%goto exit;*/
                            %end;
                        %else %do;  
                            %psionboundary;
                            %end; /*boosample ne 0*/
                        %end; /*psi1 on upper boundary*/
                    %else %do;
                        %if &npsi=2 %then %do;
                    /*if psi1 is not on boundary, check whether psi2 is...*/
    /**/                    %if %sysevalf(&psi2>&y) %then %do;
                                %if %sysevalf(&y+&srchbnd-&psi2<1.e-11) %then %do;
                                    %if &boosample=0 %then %do;
                                        %put ERROR: Optimization did not converge. Program will exit.; 
                                        %put You might try specifying a different value of parameter SRCHBND and/or STARTING_POINT.;
                                        %let arnold=1;/*%goto exit;*/
                                        %end;
                                    %else %do;  
                                        %psionboundary;
                                        %end; /*boosample ne 0*/
                                    %end; /*psi2 on upper boundary*/
                                %else %do;/*both parameters are okay!*/     
                                    %validpsi; /*includes calls to %lifeafterdeath and %blipdownup*/
                                    %end;
                                %end; /*psi2 above starting_point*/
                            %else %do;
                                %*Checking that &psi2 is not on negative boundary.;
                                %if %sysevalf(&psi2-&y+&srchbnd<1.e-11) %then %do;
                                    %if &boosample=0 %then %do;
                                        %put ERROR: Optimization did not converge. Program will exit.; 
                                        %put You might try specifying a different value of parameter SRCHBND and/or STARTING_POINT.;
                                        %let arnold=1;/*%goto exit;*/
                                        %end;
                                    %else %do; 
                                        %psionboundary;
                                        %end; /*boosample ne 0*/
                                    %end; /*psi2 on negative boundary*/
                                %else %do;/*both parameters are okay!*/
                                    %validpsi; /*includes calls to %lifeafterdeath and %blipdownup*/
                                    %end; /*psi2 not on boundary*/
                                %end;/*psi2 below starting_point*/

                            %end; /*npsi=2*/
                        %else /*npsi=1*/ %do;
                            %validpsi; /*includes calls to %lifeafterdeath and %blipdownup*/
                            %end;/*npsi=1*/

                    /**/
                        %end;/*psi1 not on positive boundary*/
                    %end; /*psi1 to the right of starting_point*/
                %else %do;
                        %*Checking that &psi1 is not on negative boundary.;
                    %if %sysevalf(&psi1-&x+&srchbnd<1.e-11) %then %do;
                        %if &boosample=0 %then %do;
                            %put ERROR: Optimization did not converge. Program will exit.; 
                            %put You might try specifying a different value of parameter SRCHBND and/or STARTING_POINT.;
                            %let arnold=1;/*%goto exit;*/
                            %end;
                        %else %do; 
                            %psionboundary;
                            %end; /*boosample ne 0*/
                        %end; /*psi1 on negative boundary*/
                    %else %do;
                        %if &npsi=2 %then %do;
                    /*if psi1 is not on boundary, check whether psi2 is...*/
                            %if %sysevalf(&psi2>&y) %then %do;
                                %if %sysevalf(&srchbnd+&y-&psi2<1.e-11) %then %do;
                                    %if &boosample=0 %then %do;
                                        %put ERROR: Optimization did not converge. Program will exit.; 
                                        %put You might try specifying a different value of parameter SRCHBND and/or STARTING_POINT.;
                                        %let arnold=1;/*%goto exit;*/
                                        %end;
                                    %else %do;  
                                        %psionboundary;
                                        %end; /*boosample ne 0*/
                                    %end; /*psi2 on positive boundary*/
                                %else %do;/*both parameters are okay!*/
                                    %validpsi; /*includes calls to %lifeafterdeath and %blipdownup*/
                                    %end;
                                %end; /*psi2 above starting_point*/
                            %else %do;
                                    %*Checking that &psi2 is not on negative boundary.;
                                %if %sysevalf(&psi2+&srchbnd<1.e-11) %then %do;
                                    %if &boosample=0 %then %do;
                                        %put ERROR: Optimization did not converge. Program will exit.; 
                                        %put You might try specifying a different value of parameter SRCHBND and/or STARTING_POINT.;
                                        %let arnold=1;/*%goto exit;*/
                                        %end;
                                    %else %do; 
                                        %psionboundary;
                                        %end; /*boosample ne 0*/
                                    %end; /*psi2 on negative boundary*/
                                %else %do;/*both parameters are okay!*/
                                    %validpsi; /*includes calls to %lifeafterdeath and %blipdownup*/
                                    %end; /*psi2 not on boundary*/
                                %end; /*psi2 below starting_point*/
                            %end; /*npsi=2*/
                        %else /*npsi=1*/ %do;
                            %validpsi; /*includes calls to %lifeafterdeath and %blipdownup*/
                            %end;/*npsi=1*/

                                /**/
                        %end; /*psi1 not on boundary*/
                    %end; /*psi1 to left of starting_point*/
                %end; /*psi1 has a value*/
%mend processpsi;

%macro psionboundary;
    %put WARNING: Algorithm did not converge, and/or psi estimated on boundary. Cumulative risks under interventions will not be calculated;
    %let boundarypsi=%eval(&boundarypsi+1);
    data psiprob;
        _sample_=&boosample;
        %if %bquote(&psi1)= %then %do; 
            %do iiii=1 %to &npsi;
                psi&iiii=.;
                %end;           
            %end;
        %else %do;
            %do iiii=1 %to &npsi;
                psi&iiii=&&&psi&iiii;
                %end;
            %end;
        run;
    %if %eval(&boundarypsi)=1 %then %do;
        data noncnvrg;
            set psiprob;
            run;
        %end;
    %else %do;
        proc append base=noncnvrg data=psiprob;
            run;
        %end;
    data _snmm1_&outc (drop=ar);
        set _snmm1_&outc (keep=%do iiii=1 %to &npsi; psi&iiii %end; 
                                objfunc _rc_overall _boosample_
                                grid_min grid_psi1 grid_psi2);
        array nonmonos nonmono&mintime-nonmono&maxtime;
        do ar=1 to %eval(&timepoints);
            nonmonos(ar)=.;
            end;
        run;
%mend psionboundary;
%macro validpsi;
        %*Changing dataset structure to include observations after death, with appropriate values for treatment;        
    %lifeafterdeath(dataset=msmparam,id=uniqueid);
        %*Looping over interventions;;
   /* %do ints = 1 %to %eval(&numint);/*this loop is not necessary for current version*/
    %let ints=1;
        /*blip down and back up*/
    %if &blipup=yes %then %do;
            /*first for outcome of interest*/
        %blipdownup(datain=weighted2,time=&time,timepoints=&timepoints,id=uniqueid,
                        treat=&treat1,theoutcome=&outc);
        %end; /*blipup=yes*/
    %else %do;
        %if &ints=1 %then %do;
            %let datain=weighted2;
            %let time=&time;
            %let timepoints=&timepoints;
            %let id=uniqueid;
            %let treat=&treat1;
            %let theoutcome=&outc;
            %blipdown;
            %end; /*first intervention*/
        %end; /*blipup=no*/
        /*%end; /*ints loop*/
%mend validpsi;
/****************************************************************************************/
/****************************************************************************************/
/****************************************************************************************/
/* SARAH'S DATA IS STANDARD FORMAT WITH NO DATA AFTER PERSON DIES. HOWEVER FOR BLIPPING */
/* UP AND DOWN WE NEED OBSERVATIONS UNTIL END OF FOLLOW-UP FOR EVERYONE EVEN AFTER THEY */
/* FAIL. SINCE ONLY THE UNCENSORED SO-FAR WILL BE BLIPPED DOWN, WE EXCLUDE THE CENSORED */
/****************************************************************************************/
/****************************************************************************************/
/****************************************************************************************/
%macro lifeafterdeath (dataset=,id=);

%put Adding observations after failure;
/* FIRST: NEED TO WORK IN THE (WEIGHTED) PSEUDOPOPULATION.*/
proc sort data=&dataset; by &id &time; run;
%if %sysevalf(&ncens_var>0) %then %do;
    proc sort data=&weightdata; by &id &time; run;/*these weights are not from the g-estimation procedure*/
    %end;

    data weighted;
        merge &dataset
        %if %sysevalf(&ncens_var>0) %then %do; &weightdata %end; ;
        by &id &time;
        %if &ncens_var=0 %then %do; ipcw=1; %end;
    run;

    data last (keep=&id censever);
        set weighted;
        by &id;
        if &outc=. then censever=1;
                   else censever=0;
        if last.&id;
        run;

    data weighteda;
        merge weighted last;
        by &id ;
        run;



            /*include only the observations with nonmissing outcome, 
                and add observations if they died before end of follow-up*/
    data weighted2 (drop=&time);
        set weighteda ; 
        by &id &time;
        &outc._l1=0;
        newtime=&time;
        if &outc ne . then output;
        if last.&id and &time ne %eval(&maxtime) and &outc=1 then do;       
            do until (newtime=%eval(&maxtime)) ; /*if this loop would not terminate, program has already stopped */
                newtime=newtime+1;
                &outc._l1=&outc;
                &treat1._l1=&treat1;
                &treat1=0;

                output;
                end; /*while*/
            end; /*if failure before end*/
        run;

/*****************************************/
proc sort data=weighted2 out=weighted3; by newtime; run;
data weighted4;
    set weighted3;
    by newtime;
    keep newtime n sumwt;
    retain n sumwt;
    if first.newtime then do;
        n=0;
        sumwt=0;
        end;
    n=n+1;
    sumwt=sumwt+ipcw;
    if last.newtime then output;
    run;

%if &boosample=0 %then %do; proc print; run; %end;
/******************************************/

proc datasets nolist;
    delete last weighteda /**/weighted4 weighted3/**/;
    quit;

    %let time=newtime;
%mend lifeafterdeath;

/*  BLIPPING DOWN IS NOT THE REVERSE OF BLIPPING UP BECAUSE WE BLIP DOWN FOR 
    INDIVIDUALS, GETTING SOMETHING THAT HAS NO CLEAR INTERPRETATION UNTIL WE 
    TAKE ITS MEAN; BUT WE BLIP UP THE MEANS, IE NOT FOR INDIVIDUALS. */

%macro blipdown;

/****************************************************************/
/****************************************************************/
/****************************************************************/
/****** THIS MACRO WILL CALCULATE COUNTERFACTUAL RISK ***********/
/****** UNDER REFERENCE REGIME "NEVER TREATED" WHEN   ***********/
/****** THERE IS NO INTERACTION BETWEEN TREATMENT AND ***********/
/****** COVARIATES.                                   ***********/
/****************************************************************/
/****************************************************************/
/****************************************************************/

    /*  THIS MACRO WILL USE THE DATASET CREATED BY %LIFEAFTERDEATH,
        WHICH ONLY INCLUDES THE UNCENSORED PEOPLE AND HAS OBSERVATIONS
        FROM BEGINNING TO END OF FOLLOW-UP FOR EVERYONE ELSE, EVEN THOSE 
        WHO FAILED BEFORE THE END OF FOLLOW-UP.     */

    /*NEED TO BLIP DOWN ONLY THE PEOPLE WHO DIDN'T GET CENSORED,*/
    /*WITH WEIGHTS.*/



%put Calculating counterfactual risks under no treatment; 
%put;

%if &npsi=2 %then %do;
    %do zerone=0 %to 1;
        %global sumwgtpreb&zerone;
        %end;
    %end;
%do timeloop=%eval(&mintime) %to %eval(&maxtime); 
    %global &theoutcome.0avg&timeloop; 
    %if &npsi=2 %then %do;
        %do zerone=0 %to 1;
            %global &theoutcome.0avgpreb&zerone._&timeloop ; 
            %end;
        %end;
    %end; 

/*  THE BLIP FUNCTION: option 1
    EXP[\GAMMA_{K,M}(A_M)]=1+[(EXP[PSI*A_M]-1)/(K-M)]*/
/*  EXCEPT SINCE HERE Yk IS MEASURED AFTER Ak, I NEED TO ADD ONE TO MY K'S IN THE FORMULA
    (that is, Y_{m+1} appears on same row with time m!)*/
/*  GENERAL FORMULA FOR COUNTERFACTUAL OUTCOME AT TIME k AFTER FINAL TREATMENT BLIP AT TIME m:
    H_{m,k}(\psi)= (1-Y_m)*Y_k*\prod_{j=m}^{k-1} \frac{k-j}{k-j+exp[\psi*A_j]-1} + Y_m */

/*  THE BLIP FUNCTION: option 2
    EXP[\GAMMA_{K,M}(A_M)]=EXP[PSI*A_M]*/
/*  THE BLIP FUNCTION: option 3
    EXP[\GAMMA_{K,M}(A_M)]=1+[(EXP[PSI1*A_M+PSI2*A_{M-1}*A_M]-1)/(K-M)]*/
/*  THE BLIP FUNCTION: option 5
    EXP[\GAMMA_{K,M}(A_M)]=1+[(EXP[PSI1*A_M+PSI2*A_{-1}*A_M]-1)/(K-M)]*/

/* NOTE: OPTION 4 NO LONGER SUPPORTED BECAUSE ARBITRARY WHEN THE CUMULATIVE AVERAGE STARTS...*/

/* blip down to treatment=0, then take average. */
proc sort data=&datain; by &id &time; run;
data blipdown (drop=idz);
    set &datain;
    by &id;
    retain idz 0 blip0cum_&mintime-blip0cum_&maxtime 1;
    %do timeloop=%eval(&mintime) %to %eval(&maxtime); /*these are the k's, ish. sort of (k-1)ish*/
        if idz ne &id then do;
            blip0cum_&timeloop=1;
            end;
        if &time>%eval(&timeloop) then do;/*since &time=m=treatment time, and &timeloop=k=outcome time*/
            blip0cum_&timeloop=1;       /*so as to be sure not to affect the product*/
            blip0_&timeloop=1;
            end;
        else do;            /*here m<=k.*/
            if censever=0 then do;
                %if &blipfunction=1 %then %do;
                    blip0_&timeloop=(%eval(&timeloop)+1-&time)/(%eval(&timeloop)-&time+exp(&psi1*&treat));
                    %end;
                %if &blipfunction=2 %then %do;
                    blip0_&timeloop=1/(exp(&psi1*&treat));
                    %end;
                %if &blipfunction=3 %then %do;
                    blip0_&timeloop=(%eval(&timeloop)+1-&time)/(%eval(&timeloop)-&time+exp((&psi1+&psi2*&treat._l1)*&treat));
                    %end;
                %if &blipfunction=5 /*includes 4*/ %then %do;
                    blip0_&timeloop=(%eval(&timeloop)+1-&time)/(%eval(&timeloop)-&time+exp((&psi1+&psi2*&treat.preb)*&treat));
                    %end;
                end;
            else do;
                blip0_&timeloop=0;
                end;
            end;
        %if &bliptype ne 1 %then %do;   /*NOTE: NO LONGER USING BLIPTYPE=1!!*/
            blip0cum_&timeloop = blip0_&timeloop * blip0cum_&timeloop;
            if &time=%eval(&timeloop) then do;
                if censever=0 then do;
                    &theoutcome.0 = &theoutcome * blip0cum_&timeloop;
                    end;
                else do;
                    &theoutcome.0 = 0;
                    end;
                end;
            %end;
        %end;
    idz=&id;
    run;


/*FOR 2-PARAMETER MODELS, WE NEED THE BLIPPED DOWN AVERAGES SEPARATELY FOR THOSE WHO WERE AND WEREN'T
    ADHERING TO THE TREATMENT AT PRE-BASELINE TIME.*/
%if (&blipfunction=3 and &lagsetto0at0=0) or &blipfunction=5 %then %do;
    proc sort data=blipdown; by &treat.preb &time; run;
    proc means data=blipdown mean sumwgt noprint;
            weight ipcw;
            var &theoutcome.0 ; by &treat.preb &time;
            output out=meanY0 (keep=&time &treat.preb &theoutcome.0avg sumwgt) mean=&theoutcome.0avg sumwgt=sumwgt;
            run;

    data meanY0preb0 meanY0preb1;
        set meanY0 ;
        %do timeloop=%eval(&mintime) %to %eval(&maxtime);
            if &theoutcome.0avg ne . then do;
                if &time=&timeloop then do;
                    &theoutcome.0avg&timeloop=&theoutcome.0avg;
                    if &treat.preb=0 then do;
                        call symput("&theoutcome.0avgpreb0_&timeloop",trim(left(put(&theoutcome.0avg&timeloop,best12.))));
                        if &time=0 then do;
                            call symput("sumwgtpreb0",trim(left(put(sumwgt,best12.))));
                            end;
                        output meanY0preb0;
                        end;
                    if &treat.preb=1 then do;
                        call symput("&theoutcome.0avgpreb1_&timeloop",trim(left(put(&theoutcome.0avg&timeloop,best12.))));
                        if &time=1 then do;
                            call symput("sumwgtpreb1",trim(left(put(sumwgt,best12.))));
                            end;
                        output meanY0preb1;
                        end;
                    end;
                end;
            %end;
        run;
    %end;/*npsi=2*/

/*NOTE: WE ALSO NEED OVERALL Y0 AVERAGE AT EACH TIME POINT.*/
proc sort data=blipdown; by &time; run;
proc means data=blipdown mean noprint;
        weight ipcw;
        var &theoutcome.0 &theoutcome; by &time;
        output out=meanY0 (keep=&time &theoutcome.0avg &theoutcome.avg) mean=&theoutcome.0avg &theoutcome.avg;
        run;

data meanY0;
    set meanY0 ;
    %do timeloop=%eval(&mintime) %to %eval(&maxtime);
        if &theoutcome.0avg ne . then do;
            if &time=&timeloop then do;
                &theoutcome.0avg&timeloop=&theoutcome.0avg;
                &theoutcome.avg&timeloop=&theoutcome.avg;
                call symput("&theoutcome.0avg&timeloop",trim(left(put(&theoutcome.0avg&timeloop,best12.))));
                call symput("&theoutcome.avg&timeloop",trim(left(put(&theoutcome.avg&timeloop,best12.))));
                end;
            end;
        %end;
    run;

/* Y0AVG&timeloop IS THE AVERAGE BLIPPED DOWN RISK AT TIME &TIMELOOP UNDER NO TREATMENT. SAME FOR EVERYONE! */

%put Observed cumulative risks (under actual treatment) at each time:;
%put Yavg_&mintime=&&&theoutcome.avg&mintime ;
%do timeloop=%eval(&mintime+1) %to %eval(&maxtime);
    %put Yavg_&timeloop=&&&theoutcome.avg&timeloop ;
    %end;
%put;

%put Counterfactual cumulative risks under no treatment (blipped down) at each time:;
%put Y0avg_&mintime=&&&theoutcome.0avg&mintime ;
%do timeloop=%eval(&mintime+1) %to %eval(&maxtime);
    %let loopm1=%eval(&timeloop-1);
    %put Y0avg_&timeloop=&&&theoutcome.0avg&timeloop ;
    %if &&&theoutcome.0avg&timeloop<&&&theoutcome.0avg&loopm1 %then %do;
        %put WARNING: Counterfactual risks under no treatment are not monotonically increasing between &loopm1 and &timeloop;
        data _snmm1_&outc;
            set _snmm1_&outc;
            %do tloop=%eval(&mintime) %to %eval(&maxtime);
                %if &timeloop=&tloop %then nonmono&tloop=1; ;
                %end;
        run;
        %end;
    %end;

%put;
%let tp1=%eval(&timepoints+1);
data interv0_all (keep=int intnum int2 pD pchd3-pchd&tp1 _sample_) ;
    set meanY0 (keep=&time &theoutcome.avg);
    int=0;
    intnum=0;
    int2="Natural Course                             ";
    _sample_=&boosample;
    pd= &theoutcome.avg;
    array Ys pchd3 - pchd&tp1;
    retain pchd3 - pchd&tp1;
    do today=1 to %eval(&timepoints-1);
        if _N_=today then do;
            Ys(today)=&theoutcome.avg;
            end;
        end;
    if _N_ = &timepoints then output ;
    run;

%local twelve;
%let twelve=1;
%if &blipup=yes %then %let twelve=2; /*called twelve because takes value 1 or 2!*/

data interv&twelve._all (keep=int intnum int2 pD pchd3-pchd&tp1 _sample_);
    format int2 $43.;
    set meanY0 (keep=&time &theoutcome.0avg);
    int=&twelve;
    intnum=&twelve;
    int2="&intlabel2"; 
    _sample_=&boosample;
    pd= &theoutcome.0avg;
    array Ys pchd3 - pchd&tp1;
    retain pchd3 - pchd&tp1;
    do today=1 to %eval(&timepoints-1);
        if _N_=today then do;
            Ys(today)=&theoutcome.0avg;
            end;
        end;
    if _N_ = &timepoints then output ;
    run;

%if %eval(&boosample)=0 %then %do;
        data &megan&nointervdata.0&parallel;
            set interv0_all;
            run;           
        data &megan&nointervdata.2&parallel;
            set interv&twelve._all;
            run;
        %end;
            
    %else %do;
        proc append base = &megan&nointervdata.0&parallel data=interv0_all;
        run;
        proc append base = &megan&nointervdata.2&parallel data=interv&twelve._all;
        run;
        %end;

%if &blipup=no %then %do;
    /*put the interv1 stuff here.*/
    %if %eval(&boosample)=0 %then %do;
        data &megan.interv1_all&parallel;
            set interv&twelve._all;
            run;
        %end;
    %else %do;
        proc append base = &megan.interv1_all&parallel data=interv&twelve._all;
            run;
        %end;
    %end;
%mend blipdown;

/*  METHOD BASED ON THEOREM 3 FROM PAGE PAPER */

/*  BLIPPING UP IS DONE FOR THE GROUP AS A WHOLE, CALCULATED USING THE MEANS. THUS IT
    SHOULD NOT BE VERY COMPUTATIONALLY INTENSIVE. NOTE HOWEVER THAT THE NUMBER OF 
    CALCULATIONS FOR THE T-FUNCTIONS IS STILL CUBIC IN THE NUMBER OF TIMEPOINTS. */

%macro blipdownup(datain=,id=,time=,timepoints=,treat=,theoutcome=,
    intno= 1,
    intlabel=&intlabel1,
    nintvar = 1,
    int1var = &treat1,
    int1type = targ,
    int1value = 1);

/****************************************************************/
/****************************************************************/
/****************************************************************/
/****** THIS MACRO WILL CALCULATE COUNTERFACTUAL RISK ***********/
/****** UNDER REGIME "ALWAYS TREATED" WHEN THRE IS NO ***********/
/****** INTERACTION BETWEEN TREATMENT AND COVARIATES. ***********/
/****** REFERENCE REGIME IS "NATURAL COURSE". SIMPLE. ***********/
/****************************************************************/
/****************************************************************/
/****************************************************************/

/* structure of dataset: person-time */

%if &intno=1 %then %do;     /*only do this once*/
    proc sort data=&datain; by &id &time; run; 
    %blipdown;
    %end;
%put Computing intervention &intno for &theoutcome;
%put;


/*  this proc IML calculates the t_\overline{a}'s: There are &timepoints such matrices, 
    each is a square matrix (one each of 1x1, 2x2, ..., &timepoints x &timepoints) containing 0's 
    below the antidiagonal/counterdiagonal. In the notation of the paper, we would find
    t_\overline{a}(k,j,k-1-[s+1]) located in matrix tg_k, in the (j+1)th row and the (k-s)th column.

    Then it multiplies the first column (i.e. k-1-{s+1}=-1) by the row of E[Y0]'s produced in %blipdown, 
    (inner product of two vectors, actually in the other order) for each time point to give E[Yg(k)]. */

proc IML;
    always=J(1,%eval(&timepoints),%eval(&int1value));                           
    %if %bquote(&regimen)= %then %do;
        regime=always; 
        %end;
    %else %do;
        regime={&regimen}`;
        %end;
        /*use "miguel" as "k" which equals m+1...*/
    %do miguel=1 %to %eval(&timepoints);
            /*create a matrix full of zeros*/
        %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
            tg_&miguel=J(&miguel,&miguel,0);
            %end;
        %else %do;
            %do zerone=0 %to 1;
                tg_preb&zerone._&miguel=J(&miguel,&miguel,0);
                %end;
            %end;
        
        %let miguelm1=%eval(&miguel-1);
        %let miguelm3=%eval(&miguel-3);

            /*initialize upper right hand corner of matrix*/
        %if &blipfunction=1 | &blipfunction=2 | 
            (&miguel=1 and /*(&blipfunction=4 |*/ (&blipfunction=3 and &lagsetto0at0=1))/*)*/ %then %do;
            tg_&miguel[1,&miguel]=exp(&psi1*regime[&miguel]);
            %end;/*1 parameter needed; covers 6 of 12 cases*/
        %else %do;
            %if (&blipfunction=3 /*| &blipfunction=4*/) and &miguel^=1 %then %do;
                /*if blipfunction is 3 or 4 then (as long as this is not the first matrix) 
                we want exp[psi1+psi2] (for always treat) here*/
                /*blipfunction 4 is only for populations with no pre-baseline treatment.*/
                regimet=regime`;
                psi={&psi1 &psi2};
                subreg=regimet[{&miguel,&miguelm1}];
                subreg[2]=subreg[1]#subreg[2];
                psireg=psi*subreg;
                %if &lagsetto0at0=0 %then %do;
                    %do zerone=0 %to 1;
                        tg_preb&zerone._&miguel[1,&miguel]=exp(psireg);
                        %end;/*zerone*/
                        %end;
                %else %do;
                    tg_&miguel[1,&miguel]=exp(psireg);
                    %end;
                %end;/*3 cases--includes both values of lagsetto0at0.*/
            %else %do;
                /*if blipfunction=5 or (blipfunction=3 and miguel=1 and lagsetto0at0=0) then 
                    we will blip up separately in the two groups stratified by 
                    pre-baseline treatment.*/
                %do zerone=0 %to 1;
                    psi={&psi1 &psi2};
                    subreg=regime[&miguel]||&zerone; /*in previous version used prevalence here*/
                    subreg[2]=subreg[1]#subreg[2];
                    subregt=subreg`;
                    psireg=psi*subregt;
                    tg_preb&zerone._&miguel[1,&miguel]=exp(psireg);
                    %end;/*zerone*/
                %end;/*last 3 cases*/
            %end;/*2 parameters*/

            /*use "colin" as the column index, which equals m-(s+1) and goes from -1 to m-1; but in the other order;
            /*but name "jamie" as the IML-approved indices=colin+2, going from 1 to m+1, to use as subscripts*/
            /*similarly, use "rowin" as the row index, which equals j, and goes from 0 to m. 
            /*but similarly name "maile" as the IML-approved indices=rowin+1*/
        %do colin=%eval(&miguelm3) %to -1 %by -1;/*top R corner set above, start from 2nd-to-last column*/
            %let jamie=%eval(&colin+2);
            %let jamiep1=%eval(&jamie+1);
            %let colinp1=%eval(&colin+1);
            %do rowin=0 %to %eval(&miguelm1);
                %let maile=%eval(&rowin+1);
                %if &rowin<%eval(&miguel-&jamie) %then %do; /*want j<s, ie rowin<m-1-colin*/
                                            
                    %if &blipfunction=1 /*| (&blipfunction=4 and &colin=-1)*/ 
                            | (&blipfunction=3 and &lagsetto0at0=1 and &colin=-1) %then %do;
                        tg_&miguel[&maile,&jamie]=
                            tg_&miguel[&maile,&jamiep1]*(((exp(&psi1*regime[&jamiep1])-1)/%eval(&miguel-&rowin-1-&colin))+1);
                        %end;               
                    %if &blipfunction=2 %then %do;
                        tg_&miguel[&maile,&jamie]=tg_&miguel[&maile,&jamiep1]*exp(&psi1*regime[&jamiep1]);
                        %end;
                    /*NOTE: for always treat (except at first time point), 
                            blip functions 3 and 4 should be equivalent at
                            later times since the previous treatment and previous cumulative  
                            average treatment should be always be 1 */
                    %if (&blipfunction=3 /*| &blipfunction=4*/) and &colin^=-1 %then %do; 
                        %if &lagsetto0at0=0 %then %do;
                            %do zerone=0 %to 1;
                                tg_preb&zerone._&miguel[&maile,&jamie]=                     
                                    tg_preb&zerone._&miguel[&maile,&jamiep1]*(((exp((&psi1+&psi2*regime[&jamie])*regime[&jamiep1])-1)/%eval(&miguel-&rowin-1-&colin))+1);
                                %end;
                            %end;
                        %else %do;
                            tg_&miguel[&maile,&jamie]=                      
                                    tg_&miguel[&maile,&jamiep1]*(((exp((&psi1+&psi2*regime[&jamie])*regime[&jamiep1])-1)/%eval(&miguel-&rowin-1-&colin))+1);
                            %end;
                        %end;
                    /*NOTE: WE CANNOT CONTROL &TREAT1PREB ! SO WE DO THIS TWICE FOR TWO POSSIBLE VALUES IT CAN TAKE.*/  
                    %if &blipfunction=5 | (&blipfunction=3 and &colin=-1 and &lagsetto0at0=0) %then %do; 
                        tg_preb0_&miguel[&maile,&jamie]=                        
                            tg_preb0_&miguel[&maile,&jamiep1]*(((exp(&psi1*regime[&jamiep1])-1)/%eval(&miguel-&rowin-1-&colin))+1);
                        tg_preb1_&miguel[&maile,&jamie]=                        
                            tg_preb1_&miguel[&maile,&jamiep1]*(((exp((&psi1+&psi2)*regime[&jamiep1])-1)/%eval(&miguel-&rowin-1-&colin))+1);
                        %end;               

                    %end;/*j<s*/
                %else %if &rowin=%eval(&miguel-&jamie) %then %do;
                    %if &blipfunction=1 | &blipfunction=2 /*| (&colin=-1 and &blipfunction=4)*/ 
                                                    | (&colin=-1 and &blipfunction=3 and &lagsetto0at0=1) %then %do;
                        sumcol=tg_&miguel[+,&jamiep1];
                        tg_&miguel[&maile,&jamie]=(1-sumcol)*exp(&psi1*regime[&jamiep1]); /*denom of blip reduces to 1*/
                        %end;
                    %if (&blipfunction=3 /*| &blipfunction=4*/) and &colin^=-1 %then %do;
                        %if &lagsetto0at0=0 %then %do;
                            %do zerone=0 %to 1;
                                sumcol=tg_preb&zerone._&miguel[+,&jamiep1];
                                tg_preb&zerone._&miguel[&maile,&jamie]=(1-sumcol)*exp((&psi1+&psi2*regime[&jamie])*regime[&jamiep1]); /*denom of blip reduces to 1*/
                                %end;
                            %end;
                        %else %do;
                                sumcol=tg_&miguel[+,&jamiep1];
                                tg_&miguel[&maile,&jamie]=(1-sumcol)*exp((&psi1+&psi2*regime[&jamie])*regime[&jamiep1]); /*denom of blip reduces to 1*/
                            %end;
                        %end;
                    %if &blipfunction=5 | (&blipfunction=3 and &colin=-1 and &lagsetto0at0=0) %then %do;
                        sumcol=tg_preb0_&miguel[+,&jamiep1];
                        tg_preb0_&miguel[&maile,&jamie]=(1-sumcol)*exp(&psi1*regime[&jamiep1]); /*denom of blip reduces to 1*/
                        sumcol=tg_preb1_&miguel[+,&jamiep1];
                        tg_preb1_&miguel[&maile,&jamie]=(1-sumcol)*exp((&psi1+&psi2)*regime[&jamiep1]); /*denom of blip reduces to 1*/
                        %end;
                    %end;
            %end; /*rowin*/
        %end; /*colin*/
    %end; /*miguel*/
/*NOTE! IN THE LAST/LARGEST MATRIX, THE FIRST COLUMN IS NOT MONOTONE.*/
/*however manual check seems to show it is correct.*/

    %if &boosample=0 and &intno=1 %then %do; 
        %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do; 
            print tg_&timepoints; 
            %end;
        %else %do;
            %do zerone=0 %to 1;
                print tg_preb&zerone._&timepoints;
                %end;
            %end;
        %end;


/*now I have calculated the tg_k's, so I can blip up!*/
    
        /*create a 0 row vector to fill in with the expected counterfactuals at each time under no treatment*/
    %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
        Y0avgs=J(1,%eval(&timepoints),0); 
        %end;
        /*but we need two of them if there are two parameters, because we've stratified according to pre-baseline treatment*/
    %else %do;
        %do zerone=0 %to 1;
            Y0avgspreb&zerone=J(1,%eval(&timepoints),0); 
            %end;
        %end;
        /*now fill in: NOTE: reverse order from k=matin down to 1.*/  
            /*upper limit is timepoints if mintime=1, timepoints-1 if mintime=0*/
    %let backwup=%eval(&timepoints-(1-&mintime));
    %if %sysevalf(&backwup>=&mintime) %then %do;
        %do backw=%eval(&backwup) %to %eval(&mintime) %by -1;
            %let forth=%eval(&backwup-&backw+1);    
            %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
                Y0avgs[&forth]=&&&theoutcome.0avg&backw;
                %end;/*single parameter*/
            %else %do;
                %do zerone=0 %to 1;
                    Y0avgspreb&zerone[&forth]=&&&theoutcome.0avgpreb&zerone._&backw;
                    %end;/*zerone*/
                %end;/*2 parameters*/ 
            %end;/*backw*/
        %end;
    /*%if &boosample=0 %then %do; print Y0avgs; %end;*/

        /*create a 0 row vector to fill in with the expected counterfactuals at each time under regime g*/
    %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
        EYgs=J(1,%eval(&timepoints),0);
        %end;
        /*or two of them in case there are two parameters, as above.*/
    %else %do;
        %do zerone=0 %to 1;
            EYgspreb&zerone=J(1,%eval(&timepoints),0); 
            %end;
        %end;

    %do matin=1 %to %eval(&timepoints);     /*matin=matrix index (which tg_&matin to use); will also index entries of EYgs.*/
            /*now we are ready to multiply Y0avgs by the first column of the appropriate tg_k!*/
        %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
            EYgs[&matin]=Y0avgs[1,{%do ind=%eval(&timepoints-&matin+1) %to %eval(&timepoints); &ind %end;}] * tg_&matin[,1];
            %end;       /*matin*/
        %else %do;
            %do zerone=0 %to 1;
                EYgspreb&zerone[&matin]=
                    Y0avgspreb&zerone[1,{%do ind=%eval(&timepoints-&matin+1) %to %eval(&timepoints); &ind %end;}] * tg_preb&zerone._&matin[,1];
                %end;
            %end;
        %end;       /*matin*/


    /*now put the matrix of EYgs, which contains the blipped up values, into 
      an output dataset in format for Sarah's macro to understand it.*/
    %put Saving blipped up values;
    %put;
        /*%if &boosample=0 %then %do; print EYgs; %end;*/

        /*create row vector containing names of variables for output dataset*/

    %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
        %if &theoutcome=&outc %then %do;
            pchds={%do matin=3 %to %eval(&timepoints+2); p&outc.cum&matin %end;};
            %end;
        %else %if &theoutcome=&censdead %then %do;          /*not developed*/
            pchds={%do matin=3 %to %eval(&timepoints+2); pcensdcum&matin %end;};
            %end;
        varnames={intnum _sample_}||pchds;                                              
        premat={&intno &boosample};

            /*create row vector containing the observation I want to add*/
        myrow=premat||EYgs;
            /*create or append to the dataset*/
        %if &boosample=0 %then %do;                                                 
            create inter&intno._most&theoutcome from myrow[colname=varnames];
            %end;
        %else %do;
            edit inter&intno._most&theoutcome;
            %end;
        append from myrow;
        close inter&intno._most&theoutcome;

            /*save the psi, E[Y_0(k)] values too.*/
        %if &theoutcome=&outc %then %do;
            othervarnames={intnum _sample_ %do iiii=1 %to &npsi; psi&iiii %end; 
                                            %do k=%eval(&mintime) %to %eval(&maxtime); Y0cum&k %end;};
            %end;
          /*next bit not developed yet*/
        %else %if &theoutcome=&censdead %then %do;
            othervarnames={intnum _sample_ psicomprisk %do k=%eval(&mintime) %to %eval(&maxtime); censd0cum&k %end;};
            %end;

        EY0s=J(1,%eval(&timepoints),0);
        psimat={&intno &boosample %do iiii=1 %to &npsi; &&psi&iiii %end;};
        %do timeloop=%eval(&mintime) %to %eval(&maxtime);
            %let tloop2=%eval(&timeloop+1-&mintime);
            EY0s[&tloop2]=&&&theoutcome.0avg&timeloop;
        %end;
        newrow=psimat||EY0s;

        %if &boosample=0 %then %do;
            create inter&intno.other&theoutcome from newrow[colname=othervarnames];
            %end;
        %else %do;
            edit inter&intno.other&theoutcome;
            %end;
        append from newrow;
        close inter&intno.other&theoutcome;

        %end;/*1 parameter or lagsetto0at0=1*/
    %else %do;
        %do zerone=0 %to 1;
            %if &theoutcome=&outc %then %do;
                pchds&zerone={%do matin=3 %to %eval(&timepoints+2); p&outc.cumpreb&zerone._&matin %end;};
                %end;
            %else %if &theoutcome=&censdead %then %do;/*not developed*/
                pchds&zerone={%do matin=3 %to %eval(&timepoints+2); pcensdcumpreb&zerone._&matin %end;};
                %end;
            varnames={intnum _sample_}||pchds&zerone;                                               
            premat={&intno &boosample};

                /*create row vector containing the observation I want to add*/
            myrow&zerone=premat||EYgspreb&zerone;
                /*create or append to the dataset*/
            %if &boosample=0 %then %do;                                                 
                create inter&intno._most&theoutcome.preb&zerone from myrow&zerone[colname=varnames];
                %end;
            %else %do;
                edit inter&intno._most&theoutcome.preb&zerone;
                %end;
            append from myrow&zerone;
            close inter&intno._most&theoutcome.preb&zerone;

            /*save the psi, E[Y_0(k)] values too, along with the sumwgt for prebaseline treatment levels 0 and 1.*/
                /*variable names*/
            %if &theoutcome=&outc %then %do;
                othervarnames={intnum _sample_ %do iiii=1 %to &npsi; psi&iiii %end; 
                                                %do k=%eval(&mintime) %to %eval(&maxtime); Y0cum&k %end; /*want the overall Y0s*/
                                                sumwgtpreb&zerone};
                %end;
              /*next bit not developed yet*/
            %else %if &theoutcome=&censdead %then %do;
                othervarnames={intnum _sample_ psicomprisk /*should have npsi of them, but not developed yet*/
                                                    %do k=%eval(&mintime) %to %eval(&maxtime); censd0cum&k  %end; sumwgtpreb&zerone};
                %end;

                /*create row vector containing observations (in chunks)*/
            EY0s=J(1,%eval(&timepoints),0);
            psimat={&intno &boosample %do iiii=1 %to &npsi; &&psi&iiii %end;};
            sumwgtpreb&zerone={&&sumwgtpreb&zerone};
            %do timeloop=%eval(&mintime) %to %eval(&maxtime);
                %let tloop2=%eval(&timeloop+1-&mintime);
                EY0s[&tloop2]=&&&theoutcome.0avg&timeloop;
            %end;
            newrow&zerone=psimat||EY0s||sumwgtpreb&zerone;

                /*create or append to the dataset*/
            %if &boosample=0 %then %do;
                create inter&intno.other&theoutcome.preb&zerone from newrow&zerone[colname=othervarnames];
                %end;
            %else %do;
                edit inter&intno.other&theoutcome.preb&zerone;
                %end;
            append from newrow&zerone;
            close inter&intno.other&theoutcome.preb&zerone;

            %end;/*zerone*/

        %end;/*2 parameters*/


        /*create scalar to append to end of labels column*/
    varname={int2};
    intlabels="&intlabel";
        /*create or append to the dataset containing the labels*/
    %if &boosample=0 %then %do;                                                 
        create inter&intno.label&theoutcome from intlabels[colname=varname];
        %end;
    %else %do;
        edit inter&intno.label&theoutcome;
        %end;
    append from intlabels;
    close inter&intno.label&theoutcome;



quit; 

    %if &npsi=1 | (&blipfunction=3 and &lagsetto0at0=1) %then %do;
        data interv&intno; 
            merge inter&intno._most&theoutcome inter&intno.label&theoutcome;    
            %do timeloop=3 %to %eval(&timepoints+2);
                %if &timeloop=%eval(&timepoints+2) %then rename p&outc.cum&timeloop=pD;
                %else rename p&outc.cum&timeloop=pchd&timeloop;;
                %end;
            run;    
        /*%if %eval(&boosample) = 0 %then %do;*/
            data &megan.interv&intno._all&parallel;
                set interv&intno;
            run;           
            /*%end;
        %else %do;
            proc append base = interv&ints._all data = interv&ints;
            run;
            %end;     */    
        %end;/*1 parameter or &blipfunction=3 with &lagsetto0at0=1*/
    %else %do;
        %let tpp2=%eval(&timepoints+2);
        data inter&intno._most&theoutcome (drop=j);
            merge   inter&intno._most&theoutcome.preb0 inter&intno.other&theoutcome.preb0
                    inter&intno._most&theoutcome.preb1 inter&intno.other&theoutcome.preb1;
            by _sample_;
            /*now this dataset should have all the info I need to get the weighted average at each time point for each sample.*/
            array zeros p&outc.cumpreb0_3 - p&outc.cumpreb0_&tpp2;
            array ones p&outc.cumpreb1_3 - p&outc.cumpreb1_&tpp2;
            array allsn p&outc.cumbyn3 - p&outc.cumbyn&tpp2;
            array allspseudo p&outc.cum3 - p&outc.cum&tpp2;
            do j=1 to %eval(&timepoints);
                allspseudo(j)= (zeros(j)*sumwgtpreb0+ones(j)*sumwgtpreb1)/(sumwgtpreb0+sumwgtpreb1);
                allsn(j)=(zeros(j)*(1-&prevalence) + ones(j)*&prevalence);
                end;
            run;
        data interv&intno;
            merge inter&intno._most&theoutcome (keep=p&outc.cum3 - p&outc.cum&tpp2 _sample_ intnum) inter&intno.label&theoutcome;
            %do timeloop=3 %to %eval(&timepoints+2);
                %if &timeloop=%eval(&timepoints+2) %then rename p&outc.cum&timeloop=pD;
                %else rename p&outc.cum&timeloop=pchd&timeloop;;
                %end;
            run;    
        data &megan.interv&intno._all&parallel;
            set interv&intno;
            run;           

        
        %end;/*2 parameters*/


%if ((&blipfunction=3 and &lagsetto0at0=0) or &blipfunction=5) and &boosample=%eval(&nsamples) %then %do;
    proc datasets nolist;
        delete inter&intno._most&theoutcome.preb0 inter&intno._most&theoutcome.preb1 
              inter&intno.other&theoutcome.preb0 inter&intno.other&theoutcome.preb1
              meany0preb0 meany0preb1
                                    
                ;
    quit;
    %end;
%mend blipdownup;




%macro sncftmresults(evens);


    %*********** SUMMARIZING AND OUTPUTTING RESULTS;
%if &evens=0 %then %do;
    %put Preparing results using natural course as reference;
    %end;
%if &evens=2 %then %do;
    %put Preparing results using never treated as reference;
    %end;

%if &blipfunction=5 %then %do;
    %if %bquote(&tvem)^= %then %let blipfunction=6;
    %if %bquote(&blem)^= %then %let blipfunction=4;
    %end; 

    data fin;    

    %*Reading in no intervention data;
    data &megan.interv&evens._all&parallel;
        set &megan&nointervdata&evens&parallel;
    run;

    %if &blipup=yes and &evens=0 %then %do;
        data &megan.interv2_all&parallel;
            set &megan&nointervdata.2&parallel;
            run;
        %end;

    %*Generating reference cumulative incidence;
    data _ref_;
        set &megan.interv&evens._all&parallel;
        pDref=pD;
        keep _sample_ pDref;
    run;

    %*Printing point estimate for psi and its bootstrap confidence interval;
    %if &nsamples>0 and &evens=0 /*only want to do this once*/ %then %do;
        proc univariate data = &megan._snmm1_&outc._all&parallel;
                where _boosample_ ne 0;
                var %do iiii=1 %to &npsi; psi&iiii %end;;
                output out = psistats
                    pctlpre = %do iiii=1 %to &npsi; psi&iiii._ %end;
                    pctlname = llim95 ulim95  pctlpts = 2.5 97.5;
            run;
        data psistats;
            format %do iiii=1 %to &npsi; psi&iiii psi&iiii._llim95 psi&iiii._ulim95 %end;;
            set psistats;
            %do iiii=1 %to &npsi;
                label psi&iiii._llim95='Lower Limit 95% CI';
                label psi&iiii._ulim95='Upper limit 95% CI';
                psi&iiii=&&&psi&iiii.pe;
                %end;
            run;
        title "SNCFTM parameter point estimates and bootstrap 95% CIs";
        proc print data=psistats noobs label; run; title;
    %end;

    %local by02;

    %if &evens=0 %then %do;
        %let by02=1;
    %end;
    %if &evens=2 %then %do;
        %let by02=-1;
        %let numint=1;
        %end;

    %do i=&evens %to &numint %by &by02;

        %*Outputting summary of intervention;
        %if &outputs = yes and &evens=0 %then %do;
            proc means data = &megan.interv&i._all&parallel;
            run;
            %end;

        %*Comparing intervention to reference;
        proc sort data=&megan.interv&i._all&parallel; by _sample_ ; run;
        proc sort data=_ref_; by _sample_; run;
        data interv&i; 
            merge &megan.interv&i._all&parallel _ref_;
            by _sample_; 

            if pDref^=0 then rr=pD/pDref;
            if rr>0 then logrr=log(rr);

            rd = pD - pDref;
            if rd^=. and rd^=0 then nnt = 1/rd;
        run;

        %*Appending intervention datasets;
        data fin; 
            set fin interv&i; 
            if _sample_=0;
            if int=. then int=&i;
        run;

        %*Calculating bootstrap mean, variance and confidence intervals;
        proc univariate data=interv&i noprint;
            where _sample_ ne 0;
            var pD rr rd nnt;
            output out = temp&i
                mean = pD_mean RR_mean RD_mean NNT_mean
                std =  pD_std  RR_std  RD_std  NNT_std
                pctlpre = pD_  RR_     RD_     NNT_
                pctlname = llim95 ulim95  pctlpts = 2.5 97.5;
        run;

        data temp&i;
            set temp&i;
            int = &i;
        run;
          
        proc sort data=fin; by %if &evens=2 %then descending; int; run;
        data fin;
            merge fin temp&i;
            by %if &evens=2 %then descending; int;
        run;

        %*Deleting no longer needed datasets;
        proc datasets nolist library=work; 
            delete interv&i /*interv&i._all*/ inter&i.label inter&i._most temp&i;
        quit;
        
        %end;

    %*Cleaning up results to print nicely;
    data finfin;
        set fin;

        obsp= round(&obsp*10000)/100;
        call symput('obspm',obsp);
        
        pD = round(pD*10000)/100;
        
        pD_mean = round(pD_mean*10000)/100;
        pD_std  = round(pD_std*10000)/100;

        pD_ulim95 = round(pD_ulim95*10000)/100;
        pD_llim95 = round(pD_llim95*10000)/100;

        RR = round(RR*100)/100;

        if int =0 then do;
            RR_ulim95 = .;
            RR_llim95 = .;
            end;
        
        if int^=0 then do;
            RR_ulim95 = round(RR_ulim95*100)/100;
            RR_llim95 = round(RR_llim95*100)/100;
            end;

        RD = round(RD*10000)/100;

        if int =0 then do;
            RD_ulim95 = .;
            RD_llim95 = .;
            end;
        
        if int^=0 then do;
            RD_ulim95 = round(RD_ulim95*10000)/100;
            RD_llim95 = round(RD_llim95*10000)/100;
            end;

        NNT = round(NNT);

        if int =0 then do;
            NNT_ulim95 = .;
            NNT_llim95 = .;
            end;
        
        if int^=0 then do;
            NNT_ulim95 = round(NNT_ulim95);
            NNT_llim95 = round(NNT_llim95);
            end;
        
        label pD        = 'Risk (%)';
        label pD_ulim95 = 'Upper limit 95% CI';
        label pD_llim95 = 'Lower limit 95% CI';
        label pD_std    = 'Bootstrap Risk SE';
        label pD_mean   = 'Bootstrap Risk Mean';
        label rr        = 'Risk ratio';
        label RR_ulim95 = 'Upper limit 95% CI';
        label RR_llim95 = 'Lower limit 95% CI';
        label RD        = 'Risk difference';
        label RD_ulim95 = 'Upper limit 95% CI';
        label RD_llim95 = 'Lower limit 95% CI';
        label NNT        = '# Needed to Treat';
        label NNT_ulim95 = 'Upper limit 95% CI';
        label NNT_llim95 = 'Lower limit 95% CI';
        label int       = 'Interv.';
        label int2      = 'Description';
    run;

    %*Outputting results dataset;
    %if %bquote(&resultsdata)^= %then %do;
        %put;
        %put Outputting results to &resultsdata;
        data &resultsdata;
            set finfin;
        run;
        %end;
    %put;
    
    %*Printing results;
    %if &outputs = yes %then %do;        
        title4 "PREDICTED RISK UNDER SEVERAL INTERVENTIONS using blip function &blipfunction";
        %if &evens=2 %then %do;
            title5 'comparing to never treat';
            %end;
        proc print data=finfin noobs label double;
            var int int2;
        run;

        %let nsamplesadj=%eval(&nsamples-&boundarypsi);

        title6 "Observed risk= %sysevalf(&obspm) %";
        title7 "Data= &data, Sample size= &ssize"; ;
        title8 "Number of bootstrap samples included= &nsamplesadj";
        title9 "Number of bootstrap samples excluded because optimization did not converge= &boundarypsi";

        proc print data=finfin noobs label double; 
            var int pD pD_llim95 pD_ulim95 rr rr_llim95 rr_ulim95 pD_mean pD_std; 
        run;

        proc print data=finfin noobs label double; 
            var int pD pD_llim95 pD_ulim95 rd rd_llim95 rd_ulim95 nnt nnt_llim95 nnt_ulim95;
        run;

        title;
        %end;
            
    %* Deleting no longer needed datasets;
    proc datasets nolist library=work; 
        delete fin finfin  _msmbeta_  _ref_ psistats;
    quit;

%mend sncftmresults;

%macro cleanup;

    %put Cleaning up;
    %if &blipup=yes %then %let numint=2; /*(it was set to 1 in sncftmresults when comparing to never treat)*/ 
        data &outc.psiandblippeddown;
            set %do ints=1 %to 1 /*%eval(&numint)*/;
                %if &blipup=yes %then %do;
                    %if &npsi=1 %then %do;
                        inter&ints.other&outc 
                        %end;
                    %else %do;
                            inter&ints._most&outc
                        %end;
                    %end;
                %else %do;
                    &megan&nointervdata.2&parallel;
                    %end;
                %end;;
            run;
        proc sort; by intnum _sample_; run;


        data intervs_p&outc.cum (drop=int);
            set %do ints=0 %to %eval(&numint);
                    &megan.interv&ints._all&parallel                    
                    %end;;
            run;
        proc sort; by intnum _sample_; run;

        data intervs_all;
            merge &outc.psiandblippeddown 
                    intervs_p&outc.cum ;
            by intnum _sample_;
            run;

        %if %bquote(&intervs_all)^= %then %do;
            %put Outputting counterfactual outcomes to &megan&intervs_all&parallel;
             data &megan&intervs_all&parallel;
                set intervs_all;
                run;
            %end;

            data &megan._snmm1_&outc._all&parallel;
                set &megan._snmm1_&outc._all&parallel;
                nonmonocount=0;
                %do tloop=%eval(&mintime) %to %eval(&maxtime);
                    nonmonocount= nonmonocount + nonmono&tloop;
                    %end;
                run;

            data &outc.psiandblippeddown&parallel (drop=intnum);
                set &outc.psiandblippeddown %if &boundarypsi ne 0 %then noncnvrg;
                ;
                if intnum>1 then delete;
                run;
            proc sort; by _sample_; run; 



    %*Deleting no longer needed datasets;
    proc datasets nolist;
        delete _msmparamdata_ _weights_ 
            weighted2 weighted _snmm1_&outc _snmm1_comprisk
            %do ints=1 %to %eval(&numint); 
                inter&ints.label&outc inter&ints._most&outc inter&ints.other&outc 
                interv&ints._&outc interv&ints 
            %end;
        max_followup idcount /*_graph_ 
        /*split0*/ tmp10;
    quit;
%mend cleanup;
%macro survplot;
/*********************************************/
/* THIS MACRO CREATES A CLUMSY SURVIVAL PLOT.*/
/*********************************************/
    %put Preparing survival plot;

    %local tp1;
    %let tp1=%eval(&timepoints+1);

    data pointest;
        set intervs_all;
        where _sample_=0;
        %do intloop=0 %to %eval(&numint);
            if intnum=&intloop then do;
                call symput("intlabel&intloop",trim(left(int2)));
                end;
            %end;
        run;

    data pointestnew ;
        set pointest;
        array pchds (&timepoints) pchd3-pchd&tp1 pd;
        do timeloop=1 to %eval(&timepoints);
            %do intloop=0 %to %eval(&numint);
                if intnum=&intloop then do;
                    cuminc&intloop=pchds(timeloop);
                    tpoint=timeloop-1;
                    end;
                %end;
                output;
            end;
        run;
    proc sort; by tpoint intnum; run;

    data pointesttest 
        (keep=%do intloop=0 %to %eval(&numint); cuminc&intloop cumsurv&intloop %end; 
            tpoint /*%do iiii=1 %to &npsi; psi&iiii %end;*/);
        set pointestnew;
        by tpoint intnum;
        array cumincs cuminc0-cuminc&numint;
        array cumsurv cumsurv0-cumsurv&numint;
        array cumincsz cumincz0-cumincz&numint;
        array cumsurvz cumsurvz0-cumsurvz&numint;
        retain cumincz0-cumincz&numint cumsurvz0-cumsurvz&numint;
        do j=1 to %eval(&numint+1);
            if cumsurv(j)=. then cumsurv(j)=cumsurvz(j);
            if cumincs(j)=. then cumincs(j)=cumincsz(j);

            cumsurv(j)=1-cumincs(j);

            cumsurvz(j)=cumsurv(j);
            cumincsz(j)=cumincs(j);
            end;
        if last.tpoint;
        run;

    data addme;
        tpoint=-1;
        %do intloop=0 %to %eval(&numint);
            cumsurv&intloop=1;
            %end;
        run;

    data pointestnewer;
        set addme pointesttest;
        %do intloop=0 %to %eval(&numint);
            attrib
                cumsurv&intloop
                label="&&intlabel&intloop";
            %end;
        tpoint=tpoint+1;
        run;

    filename grafout "&survcurv..pdf";
    goptions reset=global 
             device = pdf
             gsfname = grafout
             gsfmode = replace
                        
             /*gunit=pct */
             border 
             cback=white
             colors=(black green red blue)
             htitle=2.8
             /*ftitle=swissb ftext=swiss htitle=4 htext=2*/;

             symbol1 
                value=dot 
                interpol=stepLJ
                color=black
                line=1
                       ;
             symbol2 
                value=triangle 
                interpol=stepLJ
                color=green
                line=2
                       ; 
             symbol3 
                value=star 
                interpol=stepLJ
                color=red
                line=3
                       ;

             axis2 label=none
                    %if %bquote(&survaxis) ne %then %do;
                        order=&survaxis
                        %end;
                        ;
                    
             title 'Survival proportion under hypothetical interventions';
        
    proc gplot data=pointestnewer uniform ;
        plot (cumsurv0-cumsurv&numint)*tpoint /overlay legend vaxis=axis2; 
        run;
        quit;
    title;
    proc datasets nolist;
        delete addme pointest pointestnew pointesttest /*pointestnewer*/;
    quit;
%mend survplot;


/* THIS MACRO IS FOR BINARY SURVIVAL TIME ANALYSIS. */

%macro obscuminc(data=,time=,timepoints=,event=,censdead= censdead,censlost= censlost);

    proc sort data=&data(keep=&time &event &censdead &censlost) out=limit;
        by &time;
    run;

    data summed;
        set limit;
        by &time;
        keep &time n e r c ;
        retain n e r c ;
        if first.&time then do;
            n=0;
            e=0;
            r=0;
            c=0;
            end;
        n = n+1;
        if &event = 1 then e= e+1;
        %if &nocens=0 %then %do;
            %if %bquote(&censdead)^= %then %do; if &censdead = 1 then r = r+1; %end;
            %if %bquote(&censlost)^= %then %do; if &censlost = 1 then c = c+1; %end;
            %end;
        if last.&time then output;
    run;

    /*proc print;
    run;*/
        
    data cuminc;
        set summed;
        keep cuminc;

        by &time;
        retain cumsurv 1 cuminc 0;
   
        inc = cumsurv*(e/n);
        cuminc = cuminc + inc;

        surv = (1 - (e/n))*(1 - (r/n));
        cumsurv = cumsurv * surv;

        if _N_ = &timepoints then output;
    run;

/*proc print;
   run;*/
   
    proc datasets library=work nolist;
        delete limit summed;
    quit;
    
    %mend;

%macro numargs2(arg, delimit);
   %local n ;   
   %if %bquote(&delimit) = %then %let delimit = %str( ) ;
   %if %nrstr(&arg)= %then %do;
        0
   %end;
   %else %do;
     %let n=1;
     %do %until (%qscan(%quote(&arg), %eval(&n), &delimit )=%str());
        %let n=%eval(&n+1);
        %end;
        %eval(&n-1)
   %end;
%mend numargs2;


%macro check_list(word1,list);
    %local n  test_word in_list wordc ; 

    %let in_list = 0;
    
   %let wordc = %cmpres(%upcase(&word1));
   
    %let n = 1;
     %if %bquote(&list)^= %then %do;
          %do %until (%qscan(&list,%eval(&n),%str( ))=%str() OR (%eval(&in_list) = 1));
           
               %let test_word = %cmpres(%upcase(%qscan(&list,&n,%str( ))));
               %if &wordc = &test_word  %then %let in_list = 1;                          
               %let n = %eval(&n + 1 );
          %end;

     %end;
     %eval(&in_list) /* there is no ; here since it will be used as %let a = %check_list(&word,&list) ;
                      and the ; is included at the end of this line  */  
%mend ; 
/*SAS MACRO CODE TO REMOVE DUPLICATES FROM A LIST*/
/* Taken from www2.sas.com/proceedings/sugi26/p090-26.pdf */

%macro unique(fldlist);
    %global unique;
    %local loop1 loop2 count cnt word;
    %let cnt = 0;
    %let unique =;
    %let count = 0;

    %if &fldlist ne %then %do;
        %let word = %scan(&fldlist,1);
        %do %while (&word ne);
            %let cnt = %eval(&cnt+1);
            %let word = %scan(&fldlist,&cnt+1);
            %end;
        %end;

    %do loop1=1 %to &cnt;
        %let word=%scan(&fldlist,&loop1);
        %if &count = 0 %then %do;
            %let unique = &word;
            %let count=1;
            %end;
        %else %do;
            %do loop2=1 %to &count;
                %if &word=%scan(&unique,&loop2) %then %let word=;
                %end;
            %if &word ne %then %do;
                %let unique = &unique &word;
                %let count=%eval(&count+1);
                %end;
            %end;
        %end;

    
    %mend unique;
/********************************************/
/********************************************/
/********************************************/
/** THIS MACRO COMBINES THE VARIOUS        **/
/** RESULTS OBTAINED IN PARALLEL RUNS OF   **/
/** SNCFTMSHELL (FOR USE WITH MULTIPLE     **/
/** BOOTSTRAPS). AFTER ALL BOOTSTRAP RUNS  **/
/** HAVE FINISHED, THIS MACRO CAN BE USED  **/
/** TO CREATE THE FINAL RESULTS.           **/
/********************************************/
/********************************************/
/********************************************/


/********************************************/
/* Note! NECESSARY TO "%INCLUDE" THE WHOLE */
/* SNCFTMSHELL MACRO BEFORE RUNNING THIS.  */
/*******************************************/

%macro combineparallels(
        nsets=,     /*nsets is a number counting the datasets of each type saved 
                permanently for use in %sncftmresults. The suffixes on the names 
                will be 1-&nsets. */
        npsi=,      /*number of parameters*/
        outc=,      /*outcome variable name*/
        nointervdata=,  /*name given for saving the reference data*/ 
        nsamples=,  /*total number of bootstrap samples in the combined dataset to be created*/
        blipup=yes, /*if yes then compares always treat to both never treat and natural course*/ 
                /*if no then compares never treat to natural course only*/
        outputs=yes,    /*user can decide whether to print results*/
        resultsdata=,   /*name given for saving the results data*/
        simcensdead=0,
        censdead=,
        censlost=,
        data=,      /*name of original dataset*/
        time=,
        timepoints=,
        nocens=0,
        dir=,       /*where to save the datasets*/
        deldata=no, /*whether to delete all the datasets created in the parallel runs*/
        ssize=      /*size of bootstrap samples*/
            );


    %if &blipup=yes %then %let numint=2;
            %else %let numint=1;


/*First, calculate the observed risk.*/
    %obscuminc(data=&data,time=&time,timepoints=&timepoints,event=&outc,censdead=&censdead,censlost=&censlost);
    data cuminc;
        set cuminc;
        call symput('obsp',cuminc);
        run;

/*Set up a libname if user has specified a directory.*/
    %if %bquote(&dir)^= %then %do;
        %let megan=permadir.;
        libname permadir "&dir";
            %end;
    %else %do;
        %let megan=;
        %end;

/*Must give empty value to the macro variable parallel.*/
    %let parallel=;

/*Prepare the datasets that were saved in parallel runs*/ 
/*by correcting bootstrap sample numbers*/ 
/*and getting rid of duplicate sample 0's.*/
    %do j=1 %to %eval(&nsets);
        data _snmm1_&outc._all&j;
            set &megan._snmm1_&outc._all&j;
            if _boosample_=0 and %sysevalf(&j>1) then delete;
            _boosample_=_boosample_+%eval(&j-1)*%eval(&nsamples/&nsets);
            run;
        data &nointervdata.0&j;
            set &megan&nointervdata.0&j;
                if _sample_=0 and %sysevalf(&j>1) then delete;
                _sample_=_sample_+%eval(&j-1)*%eval(&nsamples/&nsets);
                run;
        %if &blipup=yes %then %do;
            data &nointervdata.2&j;
                set &megan&nointervdata.2&j;
                    if _sample_=0 and %sysevalf(&j>1) then delete;
                    _sample_=_sample_+%eval(&j-1)*%eval(&nsamples/&nsets);
                    run;
            %let up=2;
            %end;
        %else %do;
            %let up=1;
            %end;
        %do jj=0 %to %eval(&up);
            data interv&jj._all&j;
                set &megan.interv&jj._all&j;
                if _sample_=0 and %sysevalf(&j>1) then delete;
                _sample_=_sample_+%eval(&j-1)*%eval(&nsamples/&nsets);
                run;
            %end;
        %end;

/*Combine the datasets.*/
    data &megan._snmm1_&outc._all;
        set %do j=1 %to %eval(&nsets); _snmm1_&outc._all&j %end;; 
        if _boosample_=0 then do;
            %do iiii=1 %to %eval(&npsi);
                call symput("psi&iiii.pe",trim(left(put(psi&iiii,best12.))));
                %end;
            end;
        run;

    %do jj=0 %to %eval(&up);    
        data &megan.interv&jj._all;
            set %do j=1 %to %eval(&nsets); interv&jj._all&j %end;;
            run;
        %end;

    %do jj=0 %to %eval(&up) %by 2;
        data &megan.&nointervdata&jj;
            set %do j=1 %to %eval(&nsets); &nointervdata&jj&j %end;;
            run;
        %end;

    data bdypsi ;
        set %do j=1 %to %eval(&nsets); &megan.bdypsi&j %end;;
        retain totbdy 0;
        totbdy=bdypsi+totbdy;
        if _n_=%eval(&nsets) then do;
            call symput("boundarypsi",trim(left(put(totbdy,best12.))));
            end;
        run;

/*Print the results from the combined datasets.*/
        %if &blipup=yes %then %do;
            %do evens=0 %to 2 %by 2;
                %sncftmresults(&evens);
                %end;
            %end;
        %else %do;
            %sncftmresults(0);
            %end;
            


/*Delete the extra datasets, if user specifies to do so.*/
%if &deldata=yes %then %do;

    %if %bquote(&dir)^= %then %do;
        proc datasets nolist library=permadir; %end;
    %else %do;
        proc datasets nolist library=work; %end;
            delete %do j=1 %to %eval(&nsets); 
                bdypsi&j 
                _snmm1_&outc._all&j
                %do jj=1 %to %eval(&numint);
                    &nointervdata&jj&j 
                    interv&jj._all&j
                    %end;
                %end;   ;
            quit;
%end;


%mend combineparallels; 



/**********************************************************************    
     16JUN2005
     Structural Nested Mean Model with optimal regime for a single outcome
    or with repeated outcome measurements per subject.

    Authors: Roger Logan, Miguel Hernan, James Robins

    Contact: Roger Logan rwlogan@hsph.harvard.edu
 
    Reference: Optimal Structural Nested Models, James M. Robins


   SNMM0.6a.SAS is a set of SAS macros to run a structural nested
   mean model with either a time varying outcome or a single
   outcome for each subject in a study.

   %REALD, the main dirver, inputs data, analysis and computing
   specifications   

   %SETDATAM prepares the data from REALD or GENDATAM to be 
   analyzed by the macro SNMM_NO_REP or SNMM_ADD or SNMM_MULT

   %SNMM_NO_REP estimates the parameters of a structural mean
   model using either a multiplicative model or an additive model 
   in the case where there is only a single output recorded at the
   end of follow-up for each subjects in the study (excluding those
   subjects who are censored prior to end of follow-up). This macro
   can use a variety of methods for finding the solutions to  various 
   estimating equations including doubly-robust and efficient estimators. 
   IPCW is used to deal with censoring by loss to follow-up.

   %SNMM_ADD estimates the parameters of a structural mean model using
   an additive model where the outcome of interest is observed at each
   time point for each subject until the end of follow-up or that time
   the subject is lost to follow-up. The same methods listed above in
   the macro SNMM_NO_REP are available in this macro.

   %SNMM_MULT estimates the parameters of a structural mean model using
   a multiplicative model where the outcome of interest is observed at each
   time point for each subject until the end of follow-up or that time
   the subject is lost to follow-up. The same methods listed above in
   the macro SNMM_NO_REP are available in this macro. 
  
   %SNMM calls one of the above SNMM_ macros to obtain the desired estimates.

   %DEFAULT_RHO is an example of what a user defined IML module should look like 
   for the user to define their own blip function. Will have input a vector psi and
   will output the blip function values at each observed time for each subject 
   contributing to the estimating equations. Will also output a vector containing
   the derivative of the blip fucntion with respect to psi at each observed time.

   %BIGBOOT, boostrap estimates of variance

   %FIND_ROOT, bisection method to find root of estimating equation 
   of one-dimensional parameters in SNMM

   %SIM_ANN, simulated annealing to find minimum of objective function
   in SNMM
 
   %GRIDSEARCH is a general macro for performing a grid search to find an
   estimate of the minimum of the objective function in SNMM. The results of
   the gridsearch can also be used for plotting either the objective function
   or the estimating equation when there is only a single parameter of interest.

   %CONF_REGION is an experimental macro for trying to find an estimate of the
   confidence region for the parameters of interest when there are at most two
   parameters. Uses an adaptive grid search and can take lots of time depending
   on the behavior of the objective function.

   %SIMULATIONS is a driver for generating and reporting the results from a set of 
   simulations.

   %GENDATAM creates either a single simulated data set to be used in the testing of
   the various macros making up the SNMM file.

   %PLOT_GRAPH, auxillary macro for graphing the results from a grid search
   when the number of parameters is at most two. 

   %NUMARGS, auxillary macro for finding number of entries in a macro list

   %OUTPUT_RESULTS2, macro for displaying final results
 
   %NEWCLASS, a macro to convert a list of class variables to a new list of 0-1 
   variables. Will convert all class variables that are listed in the call of REALD
   to 0-1 varriable and remove the class statements in the proc statements in SETDATAM.
 
   %REMOVE_CLASS,  remove entries in one list  from a second list

 
   NOTE:  When using any of the censoring varibles, CENS, CENS2 or CENS3, if there are
          any missing values the program will set the value  pC_t = Pe(CENS = 0 | covariates ) = 1


***********************************************************/


%macro cumft(data=,

                         /** VARIABLES INCLUDED IN THE MODEL OF INTEREST **/

    outcome = ,                /* outcome variable */
    psivar =  ,                /* variables used in the blip function. By default the
                                     treatment variable is included. The form is
                                     blip = treatment *(psi0 + psi1 * L + psi3*Am1) */
    intercept = 1,              /* use intercept, psi0, in the blip function above */
    blip_type = 0 ,            /* indicator of type of blip function which will be used in
                                   the calculation of H_{m,k+1}. PRESENTLY ONLY WORKS WITH 
                                   MULTIPLICATIVE MODEL.
                                   blip_type = 0  use the original form given in the definition of psivar
                                   blip_type = 1 use the form H = Y_{k+1} exp( -psi * A_k) */
        
     user_rho = 0,              /* let the user define the blip functions. If this is 0, then
                                   the program will use the defualt definition of rho which
                                   uses the optimal treatment regime form of the blip funtion.
                                   If user_rho = 1 then the user must provide a iml module named
                                   rho_d_dstar which has the form rho_d_dstar(psi,blip,dblip) where
                                    blip is an array which conatins the quantity
                                            gamma(A_{k-1},L_k,d(A_{k-1},L_k,psi),psi) - 
                                                   gamma(A_{k-1},L_k,d_star(A_{k-1},L_k),psi)
                                            and is evaluated at each observation for each subject
                                     dblip is an array with dimensions p_time x dim(psi) where
                                            p_time is the person time for those subjects who are
                                            contribute to the estimating equations.
                                            dblip hold the derivative of the above blip function with
                                            respect to psi. 
                                     dblip only needs to be evaluated when blip_Q = 0. */
     rho_macro = default_rho,    /* name of the macro that holds iml module rho_d_dstar */
     user_rho_var =         ,    /* list of variables used in the user defined rho_d_dstar */ 

     user_npsi = 0,              /* number of psi in users definition of blip function */                                
         
    treat = ,                  /* treatment variable */
    id =   ,
    time = ,
    covA  = ,                 /* variables in the model of treatment. Will use a logistic model which includes an intercept */
    byA = ,                     /* by variable for separate models of A [ADDED BY SALLY]*/
    classA = ,                /* class variables used in model for A. used in proc logistic */
     eligible = ,               /* 0-1 indicator if measurements is eligible for inclusion in model for probabilty of treatment */  
    covEH = ,                  /* variables in the model for E(H_m). can not include the current treatment A_m */
    classH = ,                 /* class variables used in model for E(H)  */
    covVH = ,                  /* variables in the model for the var(H) which is used in constructing the efficient estimator. An intercept  is added */
    classVH= ,                 /* class variables used in model for var(H)  */
    covDH = ,                  /* variables used in the model for the derivative of H with repsect to psi. An intercept is added */
    classDH= ,                 /* class variables used in model for derivative of H with repsect to psi */
    Cens = ,                   /* 0-1 variable for indicating censoring, missing values will have P(Cens = 0) = 1 */
    covC =,                   /* variables used in finding probability of being censored at time m */
    whereC=,
    classC =,                 /* class variables used in model for censoring. Used in proc logistic */
     Cens2 = ,                   /* 0-1 variable for indicating censoring, missing values will have P(Cens = 0) = 1 */
    covC2 =,                   /* variables used in finding probability of being censored at time m */
    whereC2=,
    classC2 =,                 /* class variables used in model for censoring. Used in proc logistic */
     Cens3 = ,                   /* 0-1 variable for indicating censoring , missing values will have P(Cens = 0) = 1*/
    covC3 =,                   /* variables used in finding probability of being censored at time m */
    whereC3=,
    classC3 =,                 /* class variables used in model for censoring. Used in proc logistic */

 
                         /*** DATA ANALYSIS SPECIFICATIONS **/
 
    just_basic_models = 0,     /* output results from setdata alone  */                                 
    useopt=0 ,                  /* use optimal treatment formulation in the estimation of psi */
    user_regimes = 0,           /* use user defined regimes in calculation of rho_d_dstar, variables must be defined
                                   in user_defined_regimes . The macro will loop through various regimes. The user is
                                   required to make sure the user defined rho_d_dstar module accesses the
                                  user defined regimes properly by using the macro variable "&regime" */
    user_defined_regimes = ,    /* variables to used as the variable dm in the blip function calculation in rho_macro 
                                    rho_d_dstar = [ dm*(bliparg * psi ) - Am * (bliparg * psi) ] */
                               
     k_step =   ,               /* in repeated measurement case k_step is the number of periods that W_m influences 
                                   H_{m,k+1}: H_{m,m+1}, ..., H_{m,m+k_step} */                
    weight =   ,               /* user defined weight variable to be used in estimating equations */
    itt = 0,                   /* indicator of intent to treat, only valid with one psi in multiplicative model */
    repeated = 0,              /* are the measurements repeated at each time point */
    mult = 0,                  /* use a multiplicative model */
    zero_one = 0,              /* indicator of using 0-1 outcome data  in the multiplicative model */
    dr = 1,                    /* use the doubly robust estimating equations including a model for  E(H) */ 
    estprob=1,                 /* use true probability from data generation: 1/(1+exp(-logitA)) 
                                  if 0 then user must supply variable */
    est_pcens = 1,             /* use true probability from data generation 
                                  if 0 then user must supply variables named pC_t, pC2_t and pC3_t for Pr(CensX = 0) */
    useCov = 1,                /* Indicator of using the covariance matrix in constructing the estimating equations */
                               /* use simulated annealing to find starting point for nlpnms  NOT INCLUDED IN ALL VERSIONS */
    new_parts = 0,             /* use the efficient pieces of estimating equations */               
    blip_Q = 0,                /* use user defined variables in newQ: can be taken to be the variables in the blip function */
    conf_min = 0.0001,
    conf_interval = 1,         /* invert the test for estimating a confidence interval when npsi = 1 */
    bootstrap = 0,             /* use bootstrap for estimating variance of psi */
    nboot = 200,               /* number of bootstrap samples */
    bseed = 12345,
    mean_of_h = 0 ,            /* calculate the average value of Hd */

                    /** COMPUTATIONAL SPECIFICATIONS **/

    analytic = 0,            /* use analytic solution when dr = 1 , new_parts = 0, blip_q = 1 , mult = 0 and
                                user_rho = 0
                                */
 
    starting_point = ,       /* starting point for nelder mead simplex and center for gridsearch region */
    user_range =   ,         /* user defined range for gridsearch and optimization region */
     srchbnd= 5,    /* used for creating interval constraints in the optimization routines for multiplicative model */
    grid_search = 0,           /** use grid search **/
    pts1 = 100,                /* number of points used in grid search */
     plot_graph = 1,             /* plot results of grid search in one or two dimensions */
     save_graph = 0,             /* if plot_graph = 1 and save_graph = 1 then output graph in some format
                                    given by variables  gdevice and ftype .
                                    default type will be pdf */
    graph_path = ,  /*path and file name to save graph under; must be specified if save_graph=1*/
    gdevice = pdf , 
     ftype = pdf,
    countmax = 100,           /* maximum number of steps to find rough estimates for confidence bounds */

    do_graph = 0,              /** will be used for plotting graphs **/
    check_func = 0 ,           /* evaluate the test statistic at the point given in starting_point  */
    est_method = 1,            /** estimation method
                                           0 = no estimation done 
                                           1 = Nelder-Mead Simplex method
                                           2 = Newton's method
                                           3 = Simulated Annealing  
                                            4 = for one psi, use bisection method to find root of estimating equation 
                                         **/
    optn_level = 0,             /* level of output for the iml optimization routines 0 = none 5 = all */
    use_loop = 0,              /* in non-repeated additive model use smoothing of indicator function using normal distribution. Loops
                                  through parameters used in smoothing:  bandwidth */

                         /* parameters for Simulated Annealing */

     t0_eval = 15,  /* number of iterations used in finding an initial temperature */
     sim_eval = 20, /* number of iterations or steps of decreasing the temperature  */
      ddwell=25,     /* number of evaluations used in the equilbrium module */ 
     rho=0.1,       /* used in calculating new point x = rho * t * tan(theta) where theta is
                        a random point in the inerval [-rrange,rrange] **/
     rrange=1.57,   /* range of points used in evaluating tangent for finding new test point.
                        default value is pi/2 */
     tscale = 1.0,  /* used in the calculation of the next temperatur t = t0/(1+i*tscale) */
     K=1.0,         /* used in the probability of accepting a new point */    

    
       
                 /** PARAMETERS FOR SIMULATED DATA **/
     
    subjects=1000 ,            /* number of subjects in the simulated data sets */
    tpoints=10 ,               /* maximum number of measurements for each subject in the simulated data */
    mcsamples=1 ,                /* number of simulated samples to run */
    use_stream = 0,            /* option for generating data sets (what random number generators to use */
    use_new_stream = 1,        /* use new stream with new initial seeds  for non-overlapping streams of random numbers */
    usebig = 0,                /* generate all data sets at once */
    seed=1234567,              /* seed for random number generators */
    useCens1 = 0,
    useCens2 = 0,             /* option for using stagered entry in the simulated data generation */
    use_pos = 0,               /* generate positive Y for simulated data */
    h0offset = 0,         /* offset used in calculating H0 in simulated data for multiplicative model  */
    fixedpy = 0,               /* used fixed probability of generating Y = 1 in multiplicative model */
    py = 0.1,                  /* P(Y=1) */
    failure_time = 0,          /* generate failure time data in 0-1 data. once Y = 1 stop generating data for a subject */
    mu = 0,std=1.0,
    psi1=0,psi2=0,psi3=0,      /* parameters for blip function */
    gamma0 = -3,gamma1 = 2 , gamma2 = 0.05, gamma3 = 0.3, gamma4 = -0.7,
                              /* gamma =parameters for L: logitL = -3+2*U+.05*j+.3*cumLar(j-1)-.7*cumAar(j-1);*/
    alpha0=-3 ,alpha1=0.1 , alpha2 = 1.0, alpha3 = 0.5,  alpha4 = 0.0,                      
                              /* alpha = parameters for A:  logitA=-3+.1*j+1*Aarray(j-1)+.5*Lar(j); */
    theta0 = -3.0, theta1 = 1.0, theta2 = 1.5,    
                              /* theta = parameters for censoring : -3 + 1.0 * A + 1.5 * L  */
    delta0 = -1, delta1 = 0.5 ,  delta2 = 0.5,
                                   /*  ####   delta = parameters for stage second censoring:  -4 + 1*cumL*/
    pElig = 1.0,               /* probability for a measurement at time m being eligible for inclusion in
                                          the estimating equations */ 

     t0_mean= 100 ,          /* mean for exponential distribution for failure time data generation */


            /** OUTPUT SPECIFICATIONS **/

    save_weights = 0 ,         /* save weights used in estimating equations = product of censoring weights and any
                                   user supplied weights */
    print_sim = 0,             /* print out results for inividual simulations */
    analysis = 1,               /* perfor proc univariate on results of simulations : on psi1, psi2 ,..., min */
    print_boot = 1,             /* print individual bootstrap sample results */
    extra = " " ,                /* text to be used in titles */
     debug = 0,
             /** misc specifications; temporary specifications for development **/
     hmean = 0.1, 

     use_loglinear = 1,
     use_logit = 0,
     use_true = 0
           ); 

             /* use optimal treatment formulation in the estimation of psi */   
 
  /***********************************************************
       Need a way to print out what is actually being done:
           1) real or simulated data
             2) what kind of analysis is being done.
                  a) one outcome at the end of followup using an additive model
                     and the EFFICIENT ESTIMATOR.
                  b) one outcome at the end of followup using a multiplicative model
                      and the non-efficient estimator
                  c) repeated measurements using an additive model and non-efficient estimator
                  d) repeated measurements using a multiplicative model and non-efficient estimator
             3) if simulated data is being used what are the parameters that are beig used
                 in the data generation.
             4) what variables are being used in the models for treatment and outcome.
             5) whether or not the optimal regime model is being used.


  ***********************************************************/
     title1  ;
     %local mean_h indiv _rc_overall covA_tmp covC_tmp covC2_tmp covC3_tmp   ;
     %local interA_tmp   interC_tmp interC2_tmp interC3_tmp interEH interVH interDH ;
     %let simulated = 0;
     %let ncens_var = 0 ;
     %let big_bnd = 15000000 ;
     %if &data = %then %do ;
          %let simulated = 1 ; 
          %let outcome = Y;
          %let treat = A ;
          %let id = id;
          %let time = tpoint;

          /* default models for simulated data */

          %if %bquote(&covA) = %then %let covA = tpoint Am1 cumL ;
          %if %bquote(&covEH) = %then %let covEH = cumL cumpl ;
          %if %bquote(&covVH) = %then %let covVH = Am1 L Am1L ;
          %if %bquote(&covDH) = %then %let covDH = L Am1;
          %if %bquote(&cens) = censor %then %do;
               %if %bquote(&covC) = %then %let covC = A L ;
         %end;
     %end; 
 
     /* to save space %*/
     %if %bquote(&cens)= %then %do;
          %let covC = ;
          %let classC = ;
     %end;
     %if %bquote(&cens2)= %then %do;
          %let covC2 = ;
          %let classC2 = ;
     %end;
     %if %bquote(&cens3)= %then %do;
          %let covC3 = ;
          %let classC3 = ;
     %end;
     

     %if %eval(&dr) = 0 %then %do;
          %let covEH = ;
          %let classH = ;
          %let covVH = ;
          %let classVH = ;
          %let covDH = ;
          %let classDH = ;
     %end ;


     %let npsi1 = %numargs(&psivar);  /* need extra for Am */
     /*%put psivar=&psivar,   npsi1=&npsi1;*/
     %let npsi = %eval(&npsi1+&intercept);
     %if %eval(&blip_type) = 1 %then %do;
          %let npsi = 1;
          %let user_npsi = 1;
     %end;
     %else %do;
         /*  %let use_loglinear = 1; */
          %let use_logit = 0;
          %let use_true = 0;
     %end;

     %if %eval(&mult) = 1 OR 
         %eval(&new_parts) = 1 OR 
         %eval(&blip_Q) = 0 OR 
       /*  %eval(&dr) = 0 OR  */
         %eval(&user_rho) = 1 %then  %let analytic = 0; 


     %if %eval(&analytic ) = 1 %then %do;
         %let conf_interval = 0;
         %let check_func = 0;
         %let grid_search = 0;
         %let est_method = 5;
         %let usecov = 0;
      
     %end;

    %if %eval(&useopt) = 0 OR %eval(&repeated) = 1 %then %let use_loop = 0 ;

    %if %eval(&user_rho) = 1 %then %let npsi = %eval(&user_npsi) ;
    
    %if %eval(&user_regimes) = 0 %then %do;
          %let number_models = 1;
          %let user_defined_regimes = ;
    %end;
    %else %do;
          %let number_models = %numargs(&user_defined_regimes) ;
          %let useopt = 0;
          %let user_rho_var = &user_rho_var &user_defined_regimes ; /* add user defined regimes to variables saved */
    %end;

     %let tmp_list = mean_h1;
     %do i = 2 %to %eval(&number_models);
          %let tmp_list = &tmp_list mean_h&i;
     %end;
     %local  &tmp_list  ;
     %if &est_method = 0 %then %do;
          %let bootstrap = 0;
          %let conf_interval = 0;
     %end;
                 
    %if &npsi > 1 %then %let conf_interval = 0;
     %if &bootstrap = 1 %then %let conf_interval = 0;
     %if &bootstrap = 0 %then %let nboot = 0;
     /* if not estimating H, then redifine parameters to use basic H * Q * (A-E(A)) 
                estimating equations and use blip function arguments if blip_Q = 0.
              The reason for this is if blip_Q = 0 then we will need to calculate var(H) which
                needs E(H) to be calculated. 
     */
     %if &dr = 0 %then %do;
     %let new_parts = 0;
          %let blip_Q = 1 ;
          /*%put 'WARNING USING DR = 0 CAUSES THE FOLLOWING CHANGES: NEW_PARTS = 0 , BLIP_Q=1 ';*/
     %end; 
     %else %if %eval(&dr) = 1 %then %do;
          %if %eval(&conf_interval) = 1 %then  %let newparts = 1 ;
     %end;
     %if %eval(&conf_interval) =1 %then %let usecov = 1 ;
     %if &blip_Q = 1 %then %let Q_var = &psivar; 
     %if &grid_search = 0 %then %let plot_graph = 0;
     %if %eval(&est_method) = 3 %then  %sim_ann;;
     %if %bquote(&cens)^= %then %do;
          %let ncens_var = 1;
          %if %bquote(&cens2)^= %then %do;
               %let ncens_var = 2;
               %if %bquote(&cens3)^= %then %let ncens_var = 3 ;
          %end;
     %end; 


    /*%put number of psi is &npsi;*/


     /* use iml for generating information concerning the macro */

/*          proc iml;
             

      %if &useopt=1 %then %do;
         print "Nested mean model results for &outcome based on the treatment &treat, optimal regime",
    
      %end;
      %else %do;
         print "Nested cummulative failure time model results for &outcome based on the treatment &treat",
      %end;
       %if &repeated = 0 %then %do;
          %if &mult = 0 %then %do;
             "using an efficient additive model for H where &outcome is measured once at end of follow up",
           %end;
           %else %if &mult = 1 %then %do;
             "using a multiplicative model for H where &outcome is measured once at end of follow up",
           %end;
       %end;
       %else %if &repeated = 1 %then %do;
          %if &mult = 0 %then %do;
             "using an  additive model for H where &outcome is measured at each time point",
           %end;
           %else %if &mult = 1 %then %do;
             "using a multiplicative model for H where &outcome is measured at each time point.",,
           %end;
           
       %end;
/*  title2 "starting point: &s1  &s2  &s3  &s4"; */
 /*       "Models used in the analysis: &treat: &covA","  H: &covEH "," covVH : &covVH ","covDH : &covDH ",,
       %if &estprob = 1 %then %do;
              "Using estimated probability of &treat using model &covA ",,
        %end;
        %else %do;
             "Using true probability of &treat from gendatam",,
        %end;   
          %if %bquote(&data) =  %then %do;
              "Using simulated data generated as follows: ",
             "Number of subjects=&subjects", "Time points=&tpoints ", "Samples=&mcsamples, dim(psi)=&npsi",
             "True psi used in blip function : psi1=&psi1, psi2=&psi2, psi3=&psi3",
              "A generated using logit(P(A)) = &alpha0 + &alpha1 * tpoint + &alpha2 * Am1 + &alpha3 * cumL + &alpha4 * Log(Y)",
            "L generated using logit(P(L)) = &gamma0 + &gamma1 * U + &gamma2 * tpoint + &gamma3 * cumL + &gamma4 * cumAm1 ",
              "H generated using cumL and cumpL",
             
         %end;
           ;       
          quit;
          */
          run;
        title1  " &extra " ;

     /* for real data we need to use setdata for getting model for A */
     %if %sysevalf(&npsi) = 1 %then %do;
          %let chi_level = 3.84;
     %end;
     %else %if %sysevalf(&npsi) = 2 %then %do;
          %let chi_level = 5.99 ;
     %end;
     %else %if %sysevalf(&npsi) = 3 %then %do;
          %let chi_level = 7.84;
     %end;   

     /** holder for individual results **/
     data      %do i = 1 %to %eval(&number_models);   _snmm&i._    %end; ;
     run;

     %if %bquote(&data)^=  %then %do;
          %let sample = 1 ;
          %do bsample = 0 %to &nboot ;
               %setdatam;
               %do regime = 1 %to %eval(&number_models);
                    %if %eval(&just_basic_models) = 0 %then %snmm ;
               %end;
        %end;    
     %end; /* end of data ^= null */

     %if %bquote(&data)=  %then %do;
            %simulations(n=&subjects,tpoints=&tpoints,dimpsi=3,psi1=&psi1,psi2=&psi2,psi3=&psi3,mu=&mu,std=&std,samples=&mcsamples,seed=&seed);
           
     %end; /* end of if data */

     %if %eval(&save_weights) = 1 & %eval(&failure_time) = 0  %then %do;
          title 'Analysis of weights used in estimating equations ';
          proc univariate data = _weights_ (rename = (col1 = weight));
               var weight;
          run;
          title ;
     %end;

     %if %eval(&just_basic_models) = 0 %then %do;
          %do regime = 1 %to %eval(&number_models) ;
               %if &bootstrap = 1 %then %do;
                    data _boot_results&regime;
                    set _snmm&regime._ ;         
                    if objfunc < &conf_min or _bsample_ = 0 or _rc_overall < 0 ;          
                    run;

                    /* number of bootstrap samples that converged to find an estimate of psi */
                    proc freq data=_boot_results&regime noprint;
                    table sample/ out=_boot_freq;
                    run;           

                    data _boot_freq;
                    set _boot_freq;             
                    if count ne %eval(&nboot + 1) then do;
                              true_count = count - 1;
                              put '***************** WARNING ******************';
                              put 'Sample number ' sample ': only ' true_count " out of &nboot bootstrap samples converge to less than &conf_min" ;
                    end;
                    call symput('bootstraps_that_converged',count);
                    run;
                        
                    %if %eval(&print_boot) = 1  %then %do;
                         proc print data=_boot_results&regime noobs;
                              var %if %bquote(&data)= %then sample;
                                   %do i=1 %to &npsi;
                                        psi&i
                                   %end;
                                   objfunc _bsample_ ;
                         run;                
                    %end;

                    /* results for the original data set */ 
                    proc means data=_boot_results&regime noprint;
                         var  %do i=1 %to &npsi; 
                         psi&i
                         %end; ;
                    where _bsample_ ne  0;
                    by sample;
                    output out = _mean_results 
                         %do i=1 %to &npsi;
                              mean(psi&i)=mean&i std(psi&i) = std&i
                         %end; ;                   
                    run;

                    /* create data set for confidence intervals for original data sets */
                    data  _snmm&regime._;
                         merge _boot_results&regime _mean_results;
                         by sample;
                         if first.sample;
                         %do i = 1 %to &npsi;     
                              format p_value&i PVALUE8.4 ; 
                         lb&i = psi&i - 1.96 * std&i;
                         ub&i = psi&i + 1.96 * std&i;   
                         bbias&i = psi&i - mean&i; 
                             z&i = (psi&i %if %eval(&simulated) = 1 %then - &&psi&i ; ) / std&i;
                         p_value&i = 2*(1-probnorm(abs(z&i))); 
                         %if &bootstrap = 1 %then bbias&i = psi&i - mean&i; ;
                              %if %bquote(&data)= %then %do;
                                   if lb&i < &&psi&i and &&psi&i < ub&i then covered&i = 1; 
                                   else covered&i = 0;
                              %end;
                    %end;
                    run;


                    %if %bquote(&data)=  AND &mcsamples > 1 %then %do;
                         proc univariate data= _snmm&regime._ normal plot ;
                         var %do i=1 %to &npsi;
                                   psi&i mean&i std&i bbias&i
                              %end;;
                    run; 
                         proc freq data= _snmm&regime._;
                              title 'Checking coverage probabilities using bootstrap variance';
                         table %do i=1 %to &npsi ; covered&i %end; /binomial (p=0.05 level=1);
                    run;
                    %end;
               %end; /* bootstrap = 1 (analysis)  */ 

               %if &simulated = 0 %then %do;
                    proc print data = _snmm&regime._ ;
                         var 
                         %do i=1 %to &npsi;
                              psi&i 
                              %if &bootstrap = 1 %then %do;
                              lb&i ub&i mean&i std&i bbias&i  
                                   %if %bquote(&data)= %then covered&i ; 
                         %end;
                              %if &conf_interval = 1 %then psi1_low psi1_high ;
                         %end;
                         objfunc %if &check_func = 1 %then func ; 
                         ;
                    run;
               %end; 
               %output_results;
          %end; /* regime */
     %end;
    
     /* clean up the data sets */
     %if %eval(&debug) = 0 %then %do;
          proc datasets library = work nolist;
               delete _cero_ _eligcount_ _tcount_ _goodcnt_ _st_ 
               %do i = 1 %to %eval(&number_models); all_results&i %end;  
               %if %eval(&zero_one * &blip_type) = 1 %then model_newa ;
               %if %eval(&repeated) = 1 %then %do;
                    for_repeated 
                    %if %eval(&dr) = 1 %then %do;
                         %if %eval(&zero_one) = 1 %then _just_y_ ;
                         _neww_
                    %end;
               %end;
               %if %eval(&bootstrap) = 1 %then _justid _mean_results _boot_freq _cero0_ ;
               %if %eval(&simulated) = 1 %then simdata_all seeds_tmp;
               ;
          quit;
     %end;  

%basic_only:
 
%mend ;

/*******************************************************/
 
proc iml;
 start loglinear(y,x,beta);
      
      z = log(y);
      exp_w = y; 
       
   count = 0;
   diff = 1;
   nbeta = ncol(x);
   beta0 = J(nbeta,1,0);
   tmp = exp_w # x;
   tmp2 = t(x) * tmp;
   free tmp;
   tmp3 = ginv(tmp2);
   z = exp_w # z ; /* reweight the z also */
   tmp = t(x) * z;
   beta0 = tmp3 * tmp;
   free tmp3 tmp tmp2 ;
   count = 1;

   eta = x * beta0;
   tmp = eta  <> -100 /* */ ;
   mu = exp(tmp);
  
   deta = 1.0/mu ;

   z = eta + (y - mu) # deta ;


   
   do while (  count < 20 &  diff > 0.001 );
      
/* weight in regression is mu */

      tmp = mu # x; 
     
      tmp2 = t(x) * tmp;
      free tmp ;
      
      tmp3 = ginv(tmp2);
     
      free tmp2;
      /* need to reweight z */
       z = mu # z ;
      tmp = t(x) * z;
      free  z;   
        
  
   
      beta = tmp3 * tmp ;
      free tmp3 tmp;
      
      diff = max(abs(beta-beta0));
      beta0 = beta;
      
     
   

/** bound eta to be between -100 and 100 for exponential. Hopefully this is not needed after the iteration
    method and the optimization method for psi have settled down to be close to the trueth  **/

      eta = x * beta;
   
   /*   eta = tmp >< 700.0 ;  */
        
       tmp = eta <> -100  /* >< 700  */ ;
        mu = exp(tmp);
      free tmp;
   
   
       
        deta = 1.0/mu ;
       
       z = eta + (y - mu) # deta;
       free eta ;
      count = count + 1;
    /*  forvar = t(beta) || diff || count;
       append from forvar; */
      end;
   /* free y x; */
     itter = count;
 
   
   finish loglinear;

start loglin(y,x,y0,yhat);

  yhat = y0 ;


   count = 0;
   diff = 1;
   do while (count < 100 & diff > 0.01);
     tmp1 = log(yhat) + (y/yhat - 1);
      /* for weighted regression reweight tmp1 by yhat = exp(eta) = exp(W*beta) */
     tmp1 = yhat # tmp1 ;
     tmp1 = t(x) * tmp1;
      tmp2 = t(x) * (yhat  # x );
      tmp2 = inv(tmp2);
      tmp2 = x * tmp2;
      tmp3 = tmp2 * tmp1;
      free tmp1 tmp2;
      tmp3 = tmp3 >< 700 ;
      yhat0 = exp(tmp3);
      diff = max(abs(yhat - yhat0));
      count = count + 1;
     
      yhat = yhat0;


      
   end;
      beta = ginv(x) * tmp3 ;
      
 finish loglin;

 
 start logit(y,x,beta);

/********************************
Algorithm for generalized linear models from McCullagh and Nelder

eta = X * beta 
mu = g(eta) = exp(eta)/(1+exp(eta))
eta = log [ mu / (1 - mu) ] = log(mu) - log(1-mu)
V(mu) = mu * (1-mu) 
Wght^-1 = (d eta / d mu )^2 * V(mu) 
     = 1/(mu * (1-mu))
Wght = mu * (1 - mu)  

 z = eta + (y = mu) * (d eta / d mu )
 beta is the weighted regression of z on x using Wght as the weights

 initialize with mu = 1/2*(y+1/2) and eta = g^-1(mu) = logit(mu)
 this yields:

  



 **************************/

      mu = 0.5 * (y + 0.5) ;
       eta = log( mu /(1.0 - mu )) ;
       Wght = mu # (1-mu) ;
       deta = 1.0 / Wght  ;

       z = eta + (y - mu) # deta ;

      
       
   count = 0;
   diff = 1;
   nbeta = ncol(x);
   beta0 = J(nbeta,1,0);
   tmp = Wght # x;
   tmp2 = t(x) * tmp;
   free tmp;
   tmp3 = ginv(tmp2);
   z = Wght # z ; /* reweight the z also */
   tmp = t(x) * z;
   beta0 = tmp3 * tmp;
   free tmp3 tmp tmp2 ;
   count = 1;



   
   do while (  count < 100 &  diff > 0.001 );
           
        eta = x * beta0;
        tmp = -1 * eta ;

        tmp = tmp <> -30.0 ;
        tmp = tmp >< 30.0 ; 
        tmp = exp(tmp);

        mu = 1.0 / (1.0 + tmp); /* expit transformation */
  
   
        Wght = mu # (1.0 - mu) ; /* weight for regression */

       eta = log(mu) - log(1.0-mu) ;
       deta = 1.0 / Wght ;

       z = eta + (y - mu) # deta ;

/* weighted regression of z on x with weight Wght */

       tmp = Wght # x; 
     
       tmp2 = t(x) * tmp;
       free tmp ;
      
       tmp3 = ginv(tmp2);
     
       free tmp2;
       /* need to reweight z */
        z = Wght # z ;
       tmp = t(x) * z;
       free  z;   
       beta = tmp3 * tmp ;
       free tmp3 tmp;
      
      diff = max(abs(beta-beta0));
      beta0 = beta;
       count = count + 1;
      
     
   
      end;
   
     itter = count;
    
   
   finish logit;

   store module = (loglin loglinear logit );
   
   quit;
   run;
 



%macro setdatam;

/**********************************************************
 
   When using the bootstrap for estimating the variance we need to be
   careful with the names of the datasets being used. In the case where
   we are doing the bootstrap samples one at a time we must keep the original
   dataset for sampling from.

   

   When  boot_one_a_time = 1 setdatam should behave as follows.

    On the first run through for bsample = 0 the final name will be _cero_.
     
    When bsample = 1 we will change the name of _cero_ to _cero0_  and the
    output from %bigboot will be _cero_ so that the names will be preserved
    in the calls to the various snmm macros.

  

************************************************/


%local numskip print_optn;
%let print_optn = noprint ;
/*%if %eval(&just_basic_models) = 1 OR %eval(&bsample) = 0 %then %let print_optn = ;
/*adding to Roger's code!*/
%if %eval(&boosample) = 0 %then %let print_optn = ;
/**/
%if %eval(&simulated) = 1 AND %eval(&sample) > 1 %then %let print_optn = noprint ;
%global individuals tmax _bigdata count;
%local where_statement ;
%let modelA = ;
%let modelC = ;
%let modelY = ;





%if %eval(&bsample) = 0 %then %do ; /* read data once */
     %let where_vars = ;
     %if %nrbquote(&whereC) ^= or %nrbquote(&whereC2)^= or %nrbquote(&whereC3)^= %then %do;
          %let list_tmp = &whereC  &whereC2 &whereC3 ;
          %let where_vars = ;
          %where_list(arg1 = &list_tmp, arg2 = where_vars);
     %end;

     %let covA_tmp = &covA ;
     %let covC_tmp = &covC ;
     %let covC2_tmp = &covC2;
     %let covC3_tmp = &covC3 ;

     %remove_interactions;



     data _cero_ ;  
          set &data (  keep = &id &outcome &psivar &treat &time &covA_tmp &classA &Cens &covC_tmp &classC 
                             &Cens2 &covC2_tmp &classC2 &Cens3 &covC3_tmp &classC3 &weight &user_rho_var &where_vars
                             &covEH &classH &covVH &classVH &covDH &classDH &byA  
                             &eligible  
                              ) ;  
          _bsample = 0 ;
          %if %bquote(&eligible)= %then _eligible_ = 1;;
     run;
     %if %bquote(&eligible) = %then %let eligible = _eligible_ ;
/* only need to sort the data once if bootstrapping samples one at a time */
     proc sort data= _cero_ ; 
     by &id &time; 
     run;
%end;   /*&bsample=0*/

%if %eval(&bootstrap) = 1 and %eval(&bsample) = 0 %then %do;
     data _justid ;
     set _cero_ (keep = &id)  ;     
     by  &id  ;
     if last.&id ;
     run;

     proc means data = _justid  n noprint ;
     var &id ;
     output out = for_size n = count;
     run;

     data for_size;
     set for_size end = _end_;
     if _end_ then do;
         call symput('count',count);
     end;
     run;

      data _cero0_;
      set _cero_;
      run;

      proc datasets library = work nolist;
      delete for_size;
      quit;
%end;




/* dealing with categorical variables in covDelta */


%if %eval(&bootstrap) = 1 AND %eval(&bsample) > 0 %then 
     %bootstrap_sample(  datain  = _cero0_ ,   
                                             _id    = &id,
                                        sortvar = &time, 
                                                  size = &count,
                                        dataout = _cero_ );   
%if  %bquote(&eligible)^= %then %let where_statement =   &eligible = 1 ; 

/*added by Sally: we only want to model treatment probability at times when it is measured.*/
%if &treat1skip ne -1 %then %do;
    %let numskip=%numargs(&treat1skip);
    %let wheretwo= 1=1;
    %do skippy=1 %to %eval(&numskip);
        %let wheretwo&skippy=&time ne %eval(%qscan(&treat1skip,&skippy));
        %let wheretwo=&wheretwo and &&wheretwo&skippy;
        %end;
    %if %bquote(&where_statement)^= %then %do;
        %let where_statement2 = &where_statement and &wheretwo;
        %end;
    %else %do;
        %let where_statement2 = &wheretwo;
        %end;
    %end;
%else %let where_statement2 = &where_statement ;;
/**/

%if &estprob = 1 %then %do;
    %let ceroby = _cero_;
    %if %bquote(&byA)^= %then %do;
        proc sort data=_cero_ out=_ceroby_; by &byA; run;
        %let ceroby = _ceroby_;
        %end;
        
     proc logistic data= &ceroby descending outest = _est_treat(keep = _status_ )  &print_optn    ;
               ods exclude ClassLevelInfo TypeIII Association FitStatistics GlobalTests Oddsratios;
               %if %bquote(&classA)^= %then class &classA /param=ref descending;;
               %if %bquote(&byA)^= %then by &byA;;
                model &treat = &covA;
               %if %bquote(&where_statement2)^= %then where &where_statement2 ;;  /*2 added by sally*/    
               output out=modelA(keep= &id &time &outcome _pA_t) p=_pA_t;       
          run;
         %let modelA = modelA;
         proc sort data=modelA; by &id &time; run;
         /*added by Sally, but then deleted again because it turns out the probability will
                                    cancel itself out at times when treatment not measured!

         %if &treat1skip ne -1 %then %do;
             data idty;
                set _cero_ (keep=&id &time &outcome &treat);
                run;
             proc sort; by &id &time; run;
             proc sort data=modelA; by &id &time; run;
                data &modelA (drop=Az &treat);
                    merge idty &modelA;
                    by &id &time;
                    retain Az;
                    if _pA_t=. then _pA_t=Az;
                    Az=&treat;
                    run;
            proc datasets library=work nolist;
                delete idty;
                quit; 
            %end; /*treatment not always measured*/
        /**/
%end; /* end of  estprob = 1 */



%if &est_pcens = 1 %then %do;

    %if %bquote(&Cens)^= %then %do;
            proc logistic data =  _cero_ outest = _est_cens(keep = _status_ ) &print_optn ; 
                 where &cens ne . ; 
                 ods exclude ClassLevelInfo TypeIII Association FitStatistics GlobalTests Oddsratios; 
                %if %bquote(&classC)^= %then class &classC/param=ref descending;; 
                model &Cens = &covC;
               %if %bquote(&whereC)^= %then where &whereC; ;
               output out =  modelC (keep= &id &time pC_t) p = pC_t;
          run;       
          %let modelC = modelC ;

     %end;

     %if %bquote(&Cens2)^= %then %do;
         proc logistic data =  _cero_  outest = _est_cens2(keep = _status_ ) &print_optn ; 
               where &cens2 ne . ; 
               ods exclude ClassLevelInfo TypeIII Association FitStatistics GlobalTests Oddsratios; 
               %if %bquote(&classC2)^= %then class &classC2/param=ref descending;; 
               model &Cens2 = &covC2;
               %if %bquote(&whereC2)^= %then where &whereC2; ;
               output out =  modelC2 (keep= &id &time pC2_t) p = pC2_t;
         run;        
         %let modelC = &modelC modelc2;
     %end;

     %if %bquote(&Cens3)^= %then %do;
          proc logistic data =  _cero_  outest = _est_cens3(keep = _status_ ) &print_optn ; 
               where &cens3 ne . ; 
               ods exclude ClassLevelInfo TypeIII Association FitStatistics GlobalTests Oddsratios; 
               %if %bquote(&classC3)^= %then class &classC3/param=ref descending;; 
               model &Cens3 = &covC3;
               %if %bquote(&whereC3)^= %then where &whereC3; ;
               output out =  modelC3 (keep= &id &time pC3_t) p = pC3_t;
          run;       
          %let modelC = &modelC modelC3 ;
     %end;

%end;


%let _rc_treat = 0; /* for user defined pA_t when estprob = 0 */
%let _rc_cens = 0 ;
%let _rc_cens2 = 0;
%let _rc_cens3 = . ;

%if &estprob = 1 %then %error_code(_est_treat,_rc_treat);
%if %bquote(&cens)^= %then %error_code(_est_cens,_rc_cens);
%if %bquote(&cens2)^= %then %error_code(_est_cens2,_rc_cens2);
%if %bquote(&cens3)^= %then %error_code(_est_cens3,_rc_cens3);

%let _rc_overall = %sysfunc(min(&_rc_treat, &_rc_cens, &_rc_cens2, &_rc_cens3)) ;





data _cero_ ;
     merge _cero_ %if &estprob = 1 %then  &modelA;             
     %if %bquote(&Cens)^= %then  &modelC ; ;
     by  &id &time;
     %if %bquote(&where_statement)^= %then if not( &where_statement) AND _pA_t = . then _pA_t = &treat ;;
     %if %bquote(&Cens)^=  %then if pC_t = .  then pC_t = 1;;
     %if %bquote(&cens2)^= %then if pC2_t = . then pC2_t = 1;;
     %if %bquote(&cens3)^= %then if pC3_t = . then pC3_t = 1;;     
     in_eeq = 1;    /*  need to keep all the data for calculating the new-parts and identify those that will enter in the EEQ */
     if &outcome = . then in_eeq = 0  ; 

run;
run;



/* find the number of measurments that each subject has and the
   maximum number of measurements that any subject has. This value
   will be used in setting the macro variable tmax that is used 
   in the snmm macros 
 */

    
proc freq data= _cero_ noprint;
table &id / out=_tcount_(drop=percent);
run;
     

data _cero_ ;
merge _cero_ _tcount_;
by &id;
run;
 
proc means data=_tcount_ max n noprint; 
var count;
output out=_tmax max=tmax n = count ;
run;

data _tmax;
set _tmax end = _end_;;
if _end_ then do;
  call symput('tmax',tmax);
  call symput('individuals',count);
end;
run;

/*  calculating # of uncensored subjects this must be the # of individuals that is reported for each bootstrap sample*/

proc freq data= _cero_ noprint;   
     table &id / out=_in_(drop=percent);
     where in_eeq =1;
run;
proc means data=_in_ n noprint; 
     var count;
     output out=_tmax n = count;
run;
data _tmax;
     set _tmax end = _end_;;
     if _end_ then do;
          call symput('indiv',count);
     end;
run;


/*%put number of individuals in bsample  &bsample : total number of individuals =  &individuals , non-censored =  &indiv ; /* #### */




%if %sysevalf(&repeated) = 1 %then %do;

      proc freq data = _cero_ noprint;
       table &eligible  /out = _goodcnt_ (drop=percent);
      by &id;
       run;

       data _eligcount_;
       set _goodcnt_;
       by &id;
       if last.&id;
       if &eligible = 0 then count = 0;
       drop &eligible;
       run;

 


          data _just_y_ ;
          set _cero_ (keep= &id  &outcome count );
          by &id;
          array _newys(%eval(&tmax));
          if first.&id then do;         
               _ind  = 0;
          end;
          _ind = _ind + 1;
          _newys(_ind) = &outcome;
          if last.&id then do;
               _newind = 0;
               do m=1 to count;           
                    do k=m to count;
                         _newind = _newind + 1;
                         newy = _newys(k);
                         in_eeq = 1;
                         if newy = . then in_eeq = 0 ;
                         output;
                    end;
               end;
          end;
          retain _newys _ind;        
          keep &id  _newind newy in_eeq ;
          run;

  /******************************************** 
       create a data set which will repeat each row a certain
       number of times dependent on how many observations (in time)
       a subject has. If a subject has three measurements then the data
       set will look like

        w1
        w1
        w1
        w2
        w2
        w3

    *******************************/  
    %if %eval(&dr) = 1 %then %do;
          data _neww_;
          set _cero_ end = _end_ ;
          by &id;
          %if %bquote(&k_step)^= %then _copyk = %eval(&k_step);;

          if first.&id then  do;
               _newind = 0; 
               _copy = count;          
          end;
          else do;
               _copy = _copy - 1;         
          end;
          %if %bquote(&eligible)^= %then  if &eligible = 1 then do;;
          %if %bquote(&k_step)^= %then %do;
               if _copyk > _copy then _copy_count = _copy;
               else _copy_count = _copyk;
          %end;
          %else %do;
               _copy_count = _copy;
          %end;
          do _i = 1 to _copy_count;
             _newind = _newind + 1;
              output;
          end;
          %if %bquote(&eligible)^= %then end;;
          retain _copy _newind ;
          if _end_ then  do; 
               call symput('_LengthW_',_N_); 
          end;
          run;

   
          data for_repeated;
          merge _neww_ (drop = _copy _i  in = aaa) _just_y_ ;
          by &id _newind;
          if aaa;
          run;
     %end;
     %else %do;
    /* need to add in a portion if there are external weights defined */
          data for_repeated;
          set _just_y_;
          by &id _newind;
          run;
     %end;
    

     %if  %eval(&zero_one * &mult * &dr ) = 1  %then %do;


       /*** need to add in censoring weights in model of Y */

       
          %if %bquote(&Cens)^= %then %do;
               proc iml ;

          
               use _tcount_;
               read all var {count} into tcount;
               close _tcount_;

               use _cero_;
                  
               read all var { &time} into time ;
               p_time = nrow(time) ;
           
               %if %bquote(&eligible)^= %then %do;
                    read all var {&eligible} into eligible;
               %end;
               %else %do;
                    eligible = J(p_time,1,1);  /* every observation is eligible */
               %end;
           
          

               read all var {pC_t} into pC1 ;
               %if %bquote(&cens2)^= %then read all var {pC2_t} into pC2 ;;
               %if %bquote(&cens3)^= %then read all var {pC3_t} into pC3 ;;

               %if %bquote(&weight)^= %then read all var {&weight } ;;

        
               pC_t = J(nrow(pC1),%eval(&ncens_var),0);
               %do i = 1 %to %eval(&ncens_var) ;
                    pC_t[,&i] = pC&i ;
               %end;


               close _cero_;
               use for_repeated ;
               read all var { in_eeq } ;
               close for_repeated ;

               nw = nrow(in_eeq) ;

               last = 0;
               row = 0;
               e_row = 0;
       
               ipcw_all = J(nw,%eval(&ncens_var),0);
               m_index = J(nw,1,0);
               k_index = J(nw,1,0);
               do id = 1 to %sysevalf(&individuals);

                    count = tcount[id];
                    start = last + 1;
                    last = start + count -1;

                    pC_id = pC_t[start:last,];
                    do m=1 to count;
                    
                         tmp = 1.0;
                         e_row = e_row + 1;
                         if eligible[e_row] = 1 then do;
                              c_tmp = count;
                              %if % bquote(&k_step)^= %then c_tmp = m + min(%eval(&k_step),count-m + 1) - 1 ; ;              
                              do k = m  to  c_tmp /* count */;
                                   tmp = tmp # pC_id[k,];
                                   row = row + 1;
                                   ipcw_all[row,] = 1.0/tmp;
                                   m_index[row] = m;
                                   k_index[row] = k ;
                              end;
                         end; /* end of eligible = 1 */
                    end; /* m */                
               end;

               ipcw = ipcw_all[,1];
               %do j = 2 %to %eval(&ncens_var);
                    ipcw = ipcw # ipcw_all[,&j] ;
               %end;
          %end;
          %else %do;
               ipcw = J(nw,1,1.0);  /** if no censoring then set weights to be 1 */
          %end;

          %if %bquote(&weight)^= %then ipcw = ipcw # &weight ;;

          ipcw = in_eeq # ipcw ; /* only want those who have non-missing Y_{k+1} */
          for_data = ipcw || m_index || k_index ;
          create _ipcw_ from for_data;
          append from for_data  ;
          quit;

          data for_repeated ;
          merge for_repeated _ipcw_ ;
          rename col1 = ipcw col2 = m_index col3 = k_index; 
          run;

          proc logistic data= for_repeated descending  &print_optn  ;
          ods exclude ClassLevelInfo TypeIII Association FitStatistics GlobalTests Oddsratios;
          weight ipcw ;
          model newy = &covEH;
          %if %bquote(&eligible)^= %then %do;
               where &eligible = 1 ;
          %end;
          output out=modelY(keep= &id  _pY_t &time &covEH _newind) p=_pY_t;
          run;
          %let modelY = modelY;

          data for_repeated;
          merge for_repeated modelY;
          by &id _newind;
          run;

     %end;  /*zero-one, mult, dr*/





/** need to recheck the number of subjects. it is possible that
    a subject has only one measurement which is not eligible and will
    be removed. can compare the number of rows in _tcount and _elgcount

**/
/*
%if %bquote(&eligible)^= %then %do;
proc iml;


use  _eligcount_;
read all var {count} into good_count;

ngood = nrow(good_count);
ngc = char(ngood);
call symput('individuals',ngc);

quit;

%end;
 */ 

%let _numH = %numargs( &covEH) ;
%if %eval(&dr) = 1 %then %do;
   %let _sizeW = %eval(&_LengthW_ * &_numH) ;
   %if &_sizeW > &big_bnd  %then %let _bigdata = 1;
%end;
%else %let _bigdata = 0;




%end;  /* end of repeated  */
%else %do;
%if &itt = 1 & &npsi = 1 %then %do;
/* define a new variable _a_new which is the constant value of A_1  */

   data _cero_;
   set _cero_;
   by _bsample_ &id &time ;
   if first.&id then _a_new = &treat;
   retain _a_new;
   run;
 %end; /* itt = 1 and npsi = 1 */
%end;



 proc datasets library=work nolist;
 delete _tmax /*&modelA*/ &modelC &modelY _in_ %if %eval(&estprob) = 1 %then _est_treat ;
    %if %bquote(&byA) ^= %then _ceroby_;
 %if %eval(&est_pcens) = 1 %then %do;
     %if %bquote(&cens) ^= %then _est_cens ;
     %if %bquote(&cens2)^= %then _est_cens2 ;
     %if %bquote(&cens3)^= %then _est_cens3 ;
 %end;
         ;
 run;
quit;
          
 
%mend setdatam;





%macro simulations(n,tpoints,dimpsi,psi1,psi2,psi3,mu,std,samples,seed);

%global  plotgraph sU ;
%let data = _cero_;


     data seeds ; /* data for seeds */
      run;
     %let sU = %sysevalf(&seed); /* initialize seed for U and streaminit */
     
     %if &use_stream ne 1 %then %do; /* initialize the seeds for using call ranX */
          %global sL sA sE sCense sCens sY;  

         %if %sysevalf(&use_new_stream) = 1 %then %do;
               %let sU = 123;
               %let sL =  1217607105;
               %let sA =  1920525422;
               %let sE =  1438271513;
               %let sCens2 =185685284;
               %let sCens1 = 741;
               %let sY = 2;
               %let sElig =  98765432;
               %let sT0 = 123 ;
          %end;
          %else %do;
               %let sL = %sysevalf(100 + &seed);
               %let sA = %sysevalf(1000 + &seed);
               %let sE = %sysevalf(10000 + &seed);
               %let sCens2 = %sysevalf(50 + &seed);
               %let sCens1  = %sysevalf(500 + &seed);           
               %let    sY = %sysevalf(5000 + &seed );
               %let sElig = %sysevalf(50000 + &seed);
               %let sT0 = &sY ;
         %end;    


   %end;
   


   %if &usebig = 1 %then %do;
     %gendatam(simdata_all,&samples,&n,&tpoints,&psi1,&psi2,&psi3,&mu,&std);
   %end;
   

   %do sample=1 %to &samples; 
      %put evaluating simulation &sample ;
      %if &bootstrap = 1 %then %let id = id ;
      %if &usebig = 0 %then %do;
          %if &use_stream = 1 %then %do;
               %let sU = %sysevalf(&seed + &sample);
         %end;
       
          data seeds_tmp;
          sample = &sample;
          sU=%sysevalf(&sU);
          sL = %sysevalf(&sL) ;
          sA = %sysevalf(&sA) ;
          sE = %sysevalf(&sE) ;
          sCens2 = %sysevalf(&sCens2) ;
          sCens1  =  %sysevalf(&sCens1);
          sY = %sysevalf(&sY);
          sElig = %sysevalf(&sElig);
          sT0 = %sysevalf(&sT0) ;   
          run;
            
          %if %eval(&zero_one ) = 0 %then %gendatam(simdata_all,1,&n,&tpoints,&psi1,&psi2,&psi3,&mu,&std); 
          %if &zero_one = 1 and &failure_time = 1   %then %gendatam_cumft(simdata_all,1,&subjects , &tpoints,&psi1); 
       
          %if &use_stream = 0 %then %do;
               data seeds ;
               set seeds seeds_tmp;
               if sample ^= .;
               run;
          %end;

      %end;
      data _cero_;
       set simdata_all;
       %if &usebig = 1 %then if sample = &sample ;  ;
       %if &estprob = 1 %then rename _pA_t = pArorig ;;
       %if &est_pcens = 1 and %bquote(&cens) ^= %then  rename pC_t = pc1orig pC2_t = pc2orig;;    /* #### */

       run;

       %let eligible_tmp = &eligible ;
       %do bsample = 0 %to &nboot ;
               %if &bootstrap = 1 %then %put "sample  &sample and bootstrap &bsample ";
               %setdatam;
               %do regime = 1 %to %eval(&number_models);
                    %if %eval(&just_basic_models) = 0 %then %snmm ;
               %end;
                
               %let eligible = &eligible_tmp ;
       %end;  
  %end;
      
  
%if %eval(&just_basic_models) = 0 %then %do;
     %do regime = 1 %to %eval(&number_models);
          %if &print_sim = 1 %then %do;
               proc print data =_snmm&regime._ noobs;
               var  sample 
                    %do i = 1 %to &npsi ; psi&i %end;
                    %if %eval(&analytic) = 0 %then %do;
                         %if &est_method ^= 0 | &grid_search = 1 %then objfunc ;
                         cp func_psi0
                         %do i = 1 %to &npsi; eq&i z&i %end;
                    %end;
                    %if &bootstrap = 1 %then _bsample_ ;
                    %if &conf_interval = 1 %then psi1_low psi1_high ;
                    ;                
               run;
          %end;
          
          %if &analysis = 1 & %eval(&mcsamples) ^= 1  %then %do;
               %if &usecov = 1 %then %do;
                    proc freq data=_snmm&regime._;
                    table cp /binomial ( p=0.05 level=1);
                    run;
               %end;
               
               %if %eval(&analytic) = 0 %then %do;
                    proc univariate data=_snmm&regime._;
                    var %do i = 1 %to &npsi; eq&i  z&i%end ;;
                    run;
               %end;

               %if %eval(&est_method) > 0  OR %eval(&analytic) = 1 %then %do;
               
                    proc univariate data=_snmm&regime._ mu0= %do i = 1 %to &npsi; &&psi&i %end;  ;
                    var  %do i = 1 %to &npsi; psi&i %end ;   %if %eval(&analytic) = 0 %then objfunc ; ;
                    run;
                 

               %end;   
               title3; title4;
          %end;
     %end;
%end;
 %let data = ;
    run;
%mend simulations;





 
%macro bootstrap_sample(datain=,_id=,size=,sortvar=,dataout=);

     

        %let stmp = %sysevalf(&bseed * &bsample) ;

         proc surveyselect data=_justid noprint method = urs
             n= &size 
           seed = &stmp  
           out = _idsamp;
         run;


         data _tmp;
             merge _idsamp (in = _a) &datain ;
             by &_id;
             if _a;
             _bsample_ = %eval(&bsample ) ;
         run;

         data _tmp;
            set _tmp ;
            previd = lag(&_id);
            prevhits = lag(numberhits);
            if _n_ = 1 then _tmpid = 0;
            else do;
               if previd ^= &_id  then _tmpid = _tmpid + prevhits;
            end; 
            do i = 1 to numberhits;
              _newid_ = _tmpid + i;
              output;
            end;

            retain _tmpid ;
          *  drop _tmpid previd prevhits i numberhits;
         run;

          proc sort data= _tmp;
               by _newid_ &sortvar;
          run;

        
              data &dataout;
               set _tmp ( drop = &_id);
                    rename  _newid_ = &_id  ;
              run;
          

     
     %if %eval(&debug) = 0 %then %do;
           proc datasets library = work nolist;
             delete _idsamp _tmp
            
             ;
             run; quit;
      %end;

 %mend;      


%macro snmm;


     
      
       %if %eval(&analytic) = 1 %then %analytic_sol;
       %else %do;
          %if &failure_time = 0 %then %do;
               %if       &repeated = 1  %then  %snmm_rep_define ; /* included both models for  repeated measurements  */
               %else %if &repeated = 0   %then  %snmm_no_rep_define ; /* includes both models for non-repeated measurements */
          %end;
          %else %sncftm ;

          %find_estimate ;
       %end;

       data _snmm&regime._;
       set _snmm&regime._ _kk_ ; /*appending to dataset _snmm&regime._ if it already exists*/
       if psi1 ne . ;    
       if sample=. then sample= &sample;
       if _bsample_ = . then _bsample_ = &bsample ; 
       if _rc_overall = . then _rc_overall = &_rc_overall ;
       run;

       proc datasets library=work nolist;
       delete _kk_;
       quit;
       run;


       %if &plot_graph = 1 %then %plot_graph ;

     
       
          

 %mend snmm;

 

/****/ 
%macro sncftm  ;

     data tmp0 ;
     set _cero_ (keep = &id &outcome) ;
     by &id ;
     if last.&id and &outcome = 1 ;
     run;



     data _cero_ ;
     merge _cero_ tmp0 (keep = &id in = __a );
     by &id ;
     if __a ;
     run ;


     proc means data = _tcount_ noprint ;
     var count ;
     output out = max_followup(keep = max_followup) max = max_followup;
     run;

     data _tcount_ ;
     merge _tcount_ tmp0 (keep = &id in = __a ) ;
     by &id ;
     if __a ;
     run;

     proc means data = tmp0 n noprint ;
     var &id ;
     output out = idcount (keep = n) n = n ;
     run;

     data _null_;
     set idcount ;
     call symput('individuals',trim(left(n)));
     run;



     data tmp10;
     set _cero_ (keep = &id &treat );
     by &id ;
     retain evertreat ;
     if first.&id then evertreat = 0;
     if evertreat = 0 and &treat = 1 then evertreat = 1;
     if last.&id ;
     run;

     data _cero_untreated _cero_treated;
     merge _cero_ tmp10 (keep = &id evertreat ) ;
     by &id;
     if evertreat = 1 then output _cero_treated;
     else if evertreat = 0 then output _cero_untreated ;
     run;


  
     data _null_ ;
     set _cero_treated(keep = &id ) end = _end_ ;
     by &id;
     retain newn ;
     if _n_ = 1 then newn = 0 ;
     if first.&id then newn = newn + 1 ;
     if _end_ then call symput('newn',trim(left(newn)));
     run;


     %let treated = &newn ;
     %let non_treated = %eval(&individuals - &treated ) ;

     %let nevents = &individuals ;
     /*%put treated = &treated non-treated = &non_treated ;*/
     %let cross_sectional = 0 ;

     proc iml %if &_bigdata = 1 %then worksize = 100000000 ; ;






     start mainf(psi) ;



          load Y_treated A_treated pC_t_treated pC_t_untreated
                /**/%if &blipfunction=3 %then %do; Al1_treated %end;/**/  
                /*%if &blipfunction=4 %then %do; ca_Al1_treated %end;/**/  
                /**/%if &blipfunction=5 or (&blipfunction=3 and &lagsetto0at0=0) %then %do; 
                        Apreb_treated Apreb_untreated %end;/**/
                tcount_treated  hd_time_treated p_time_treated max_followup ;

          /* calculate H_{m,k} and sum H_{m,k} over k for each subject */


          last = 0;                        
          Hm = J(p_time_treated,1,0) ;
          do id=1 to %sysevalf(&treated);   
               count = tcount_treated[id];/*num of tpoints for person id */
               start = last + 1;
               last = start + count - 1;                                          

               Ctmp= pC_t_treated[start:last];
               Atmp = A_treated[start:last] ;
               %if &blipfunction=3 %then %do; Al1tmp = Al1_treated[start:last]; %end;
               /*%if &blipfunction=4 %then %do; ca_Al1tmp = ca_Al1_treated[start:last]; %end;*/
               %if &blipfunction=5 or (&blipfunction=3 and &lagsetto0at0=0) %then %do; 
                    Aprebtmp = Apreb_treated[start:last]; %end;
               Ytemp = Y_treated[start:last];  
               /* in the estimating equation the subject will only contribute as long as they are alive.
               The new data set only contains these observations so we will not need any of the indicator
               or extra observations for when Y = 1. When Y = 1 we know that A = 0 and know what the blip 
               function should be. We will take that into account here to fill in the Hmk values

               The code below will create the sum of the Hmk for k > m 

               Changing the index in the H_mk formulation to match the data structure we have

               H_{m,k+1} = Y_{k+1} * product(j=m to j = k) (k + 1 - j)/( k - j + exp(A_{j} * psi ))

               for matrix structure this becomes Y_{k+1}= Y[k] and m <= k

               */
               Hm_id = J(count,1,0) ;
               do m = 1 to count ;                            
                    do k = m to max_followup;
                         sumblip = 0;
                         do j = m  to   k ; /* After last obs A = 0 since the subject has Y[count] = 1*/

                              if j <= count then do;

                                /*create the weight W_{m,k}*/
                                if j=m then do;
                                    tmpw=Ctmp[j,];
                                    end;
                                else do;
                                    tmpw=tmpw#Ctmp[j,];
                                    end;
                                ipcwsep=1.0/tmpw;
                                ipcwnew=ipcwsep[,#];

                                /*create the H_{m,k}*/
                                %if &blipfunction=1 %then %do;/*in original program, this is the only option; options added by Sally*/
                                   numtmp = (k + 1)  - j  ;
                                   denomtmp = k - j  + exp(psi[1] * Atmp[j]) ; 
                                %end;
                                %if &blipfunction=2 %then %do;
                                    numtmp = 1 ;
                                    denomtmp = exp(psi[1] * Atmp[j]) ; 
                                %end; 
                                %if &blipfunction=3 %then %do;
                                   numtmp = (k + 1)  - j  ;
                                   Ajjm1 = J(2,1,1);
                                   Ajjm1[2] = Al1tmp[j];/* creating column vector [1 A_l1] */
                                   Ajjm1 = Atmp[j]#Ajjm1;/* creating column vector [A A*A_l1] */
                                   
                                   denomtmp = k - j  + exp(psi*Ajjm1) ; /*row vector times column vector = scalar*/
                                   /*denomtmp = k - j  + exp(psi[1] * Atmp[j] + psi[2] * Atmp[j] * Al1tmp[j]) ; */
                                %end;
                                /*%if &blipfunction=4 %then %do;
                                   numtmp = (k + 1)  - j  ;
                                   Ajjm1 = J(2,1,1);
                                   Ajjm1[2] = ca_Al1tmp[j];/* creating column vector [1 ca_A_l1] */
                                  /* Ajjm1 = Atmp[j]#Ajjm1;/* creating column vector [A A*ca_A_l1] */
                                   
                                   /*denomtmp = k - j  + exp(psi*Ajjm1) ; /*row vector times column vector = scalar*/
                                /*%end;*/
                                %if &blipfunction=5 %then %do;
                                    numtmp = (k + 1) - j ;
                                    AApreb = J(2,1,1);
                                    AApreb[2] = Aprebtmp[j];/*creating column vector [1, Apreb]*/
                                    AApreb = Atmp[j]#AApreb;/*creating column vector [A, A*Apreb]*/
                                    denomtmp = k - j + exp(psi*AApreb) ; 
                                %end;
                                blip_tmp2 = log(numtmp) - log(denomtmp);                                                                                   
                                sumblip = sumblip + blip_tmp2 ;
                              end;/*if j*/
                         end;     /*j loop*/                       
                         exp_blip = exp(sumblip);  
                         /* for m < k + 1 < count subject is still alive so H_{m,k+1} = 0. Only non-zero terms are for
                         k + 1 >= count and here Y = 1 so H_{m,k+1} = exp_blip  */

                         if k >= count then Hm_id[m] = Hm_id[m] + ipcwnew # exp_blip ; /*weights should happen here!*/
                    end;/*k loop*/                                                                                                                                       
               end ; /* m loop */                   
               Hm[start:last] = Hm_id ;                             
          end; /* end of id loop */         

          load residA_treated ;

          /*load ipcw_treated ;*/
          U0i_treated =  /*ipcw_treated #*/  Hm # residA_treated ;


          newUi = U0i_treated;





/*          usum = J(1,&npsi,0);
          zero = J(1,&npsi,0);*/



          last = 0;
          /*zero_row = J(1,&npsi,0);*/

          newU1_treated = J(&treated,&npsi,0) ;
          do id=1 to %sysevalf(&treated);
               count = tcount_treated[id];            
               start = last + 1;
               last = start + count - 1;               
               uu_id = U0i_treated[start:last,];/*start to last contains I(Ym=0){Am-EAm}*sum over k(Hm,k) for each m*/
               usum = uu_id[+,]; /*usum sums over m*/                     
               newU1_treated[id,] = usum; /*person id's contribution to estimating equation*/
          end;

          /*newU1_treated_sum = sum(newU1_treated);
          store newU1_treated_sum ;*/

          %if &usecov=1 %then %do;
               newcov_treated = t(newU1_treated) * newU1_treated; /*since E(U1[id])=0*/
               load newcov_non_treated newU1_non_treated_sum  ;
               newcov = newcov_treated + newcov_non_treated ;

               store newcov;

               newU = newU1_treated[+,] + newU1_non_treated_sum ;/*value of estimating equation*/

               store newU;

                /*Next part does not work for npsi=2, so has been removed.*/
              /* z=newU/sqrt(newcov);/*this is the standardized estimating equation,
                    standardized by a conservative estimate of its variance - 
                    only conservative when nuisance parameters estimated*/
               /*store z;*/

               newchi = newU * ginv(newcov) * t(newU);/*this is what we are trying to minimize in optimization*/

          %end;
          %else %do;
               newU = newU1_treated[+,]+ newU1_non_treated_sum ;
               newchi = newU * t(newU); /*this is what we are trying to minimize in optimization*/
          %end;                         

          store newU /*z*/ newcov;



          %if %eval(&est_method) = 4 %then return(newU);
          %else return (newchi);
          ;

     finish mainf;

     store module=mainf;



     use _tcount_;
     read all var { count } into tcount ;
     close _tcount_ ; 


     use max_followup;
     read all var { max_followup } into max_followup;
     close max_followup ;
     store max_followup;

     use _cero_treated;                
     read all var { &treat} into A_treated ;


    
     %if &blipfunction=3 %then read all var { &treat._l1 } into Al1_treated ;;
     /*%if &blipfunction=4 %then read all var { &blem } into ca_Al1_treated ;;*/
     %if &blipfunction=5 or (&blipfunction=3 and &lagsetto0at0=0) 
        %then read all var { &treat.preb } into Apreb_treated ;;

     read all var { &outcome } into Y_treated ;                                          
     read all  var { _pA_t } into _pA_t_treated  ; 
     read all var { count } into tcount_treated where(&outcome=1) ; 
     %if %bquote(&weight)^= %then read all var { &weight } into weight_treated ;; 
     close _cero_treated;   
     p_time_treated = nrow(Y_treated);
     p_time_non_treated = 0 ;   /* holder */



     %if %eval(&non_treated ) > 0 %then %do;

          use _cero_untreated;                
          read all var { &outcome } into Y_non_treated ;                   


            %if &blipfunction=5 or (&blipfunction=3 and &lagsetto0at0=0) 
                %then read all var { &treat.preb } into Apreb_untreated ;;

          read all  var { _pA_t } into _pA_t_non_treated  ; 
          read all var { count } into tcount_non_treated where(&outcome=1) ; 
          %if %bquote(&weight) ^= %then read all var { &weight } into weight_untreated ;;
          close _cero_untreated;
          p_time_non_treated = nrow(Y_non_treated) ;    

 
     %end;

     p_time_all = p_time_treated + p_time_non_treated ;
     %if %bquote(&weight)^= %then %do;
           %if %eval(&non_treated ) > 0 %then weight = weight_treated // weight_untreated ;
           %else weight = weight_treated;
           ;
     %end;

     %if %bquote(&cens)^= %then %do;
          use _cero_treated;
          read all var {pC_t} into pC1_treated ;
          %if %bquote(&cens2)^= %then read all var {pC2_t} into pC2_treated ;;
          %if %bquote(&cens3)^= %then read all var {pC3_t} into pC3_treated ;;

          %if %eval(&non_treated ) > 0 %then %do;
               use _cero_untreated ;
               read all var {pC_t} into pC1_untreated ;
               %if %bquote(&cens2)^= %then read all var {pC2_t} into pC2_untreated ;;
               %if %bquote(&cens3)^= %then read all var {pC3_t} into pC3_untreated ;;


               /* stack the two treated and untreated data sets */
               pC1 = pC1_treated // pC1_untreated ;
               %if %bquote(&cens2)^= %then pC2 = pC2_treated // pC2_untreated ;;
               %if %bquote(&cens3)^= %then pC3 = pC3_treated // pC3_untreated ;;
               tcount_all = tcount_treated // tcount_non_treated ;
     
            pC_t_untreated=J(p_time_non_treated,%eval(&ncens_var),0);
            %do i=1 %to %eval(&ncens_var);
                pC_t_untreated[,&i]=pC&i._untreated;
                %end;

          %end;
          %else %do;
               pC1 = pC1_treated ;
               %if %bquote(&cens2)^= %then pC2 =  pC2_treated ;;
               %if %bquote(&cens3)^= %then pC3 = pC3_treated ;;
               tcount_all = tcount_treated ;
          %end;

            pC_t_treated = J(p_time_treated,%eval(&ncens_var),0);
            %do i = 1 %to %eval(&ncens_var) ;
                pC_t_treated[,&i] = pC&i._treated ;
                %end;
       

          pC_t = J(p_time_all,%eval(&ncens_var),0);
          %do i = 1 %to %eval(&ncens_var) ;
               pC_t[,&i] = pC&i ;
          %end;


     %end;


     /* load data for non-treated subjects */
     %if %eval(&non_treated) > 0 %then %do;

          /* calculate contribution to estimating equations for those subjects who are not treated (no psi dependence ) */
          /* for thos subjects who are never treated H_{m,k} = I(alive at time m) * Y_k for m < k. Changing the index to
               match the data structure this becomes
               H_{m,k+1} = I(alive at time m) Y_{k+1} = I(alive at time m ) * Y[k] = Y[k] since only using m when alive in
               data set .
               We can simplify the calculation of sum(k =m to k = max_followup ) H_{m,k+1} = max_followup -count  + 1
               since Y[k] = 0 for k < count and should be 1 for k >= count (including last observation) until the end
               of the study, which will be taken to be the longest follow-up in the data set for all subjects  
                EXCEPT THAT WE HAVE TO MAKE THAT SUM A WEIGHTED SUM!

            IN THE UNTREATED, EXP[BLIP] IS ALWAYS 1 for the 1-parameter models!. 
                For the 2 parameter models it'll be multiplied by 
                another component--residA is multiplied by a vector function.
          */
          last = 0;                        

          Hm_non_treated = J(p_time_non_treated,1,0) ;
          do id=1 to %sysevalf(&non_treated);   
               count = tcount_non_treated[id];
               start = last + 1;
               last = start + count - 1;         
        /*CORRECTED WEIGHTS!*/      
                Ytmp=Y_non_treated[start:last];
                Ytmp=shape(Ytmp,max_followup,1,1); /*fill in 1s after failure*/
                Ctmp=pC_t_untreated[start:last];
               ipcwnew=J(count,max_followup,0);
               do m = 1 to count ;                            
                    do k = m to max_followup;
                         do j = m  to   k ; 
                            if j<=count then do;
                                /*create the weight W_{m,k}=product j from m to k*/
                                if j=m then do;
                                    tmpw=Ctmp[j,];
                                    end;
                                else do;
                                    tmpw=tmpw#Ctmp[j,];
                                    end;
                                end;/*if j*/
                            end;/*j*/
                            ipcwsep=1.0/tmpw;
                            ipcwnew[m,k]=ipcwsep[,#]; /*product of weights for different kinds of censoring*/
                        end;/*k*/

            /*WEIGHTED Hmk HERE*/
               /*time_dead= max_followup - count + 1 ; 
                Hm_id=J(count,1,time_dead);*/ /*Each H_mk is 1, but we need a weighted sum, so sum the weights over k:*/
                    Hm_id = ipcwnew*Ytmp;/*this picks out the correct ones since when Ym=0, Hmk=Yk for the untreated.*/
                end;/*m*/ 

               Hm_non_treated[start:last] = Hm_id ;
          end; /* end of id loop */         

            residA_non_treated=(- _pA_t_non_treated );
            %if &blipfunction=3 and &lagsetto0at0=1 %then %do;
                extracol = 0 # (- _pA_t_non_treated );
                residA_non_treated = (- _pA_t_non_treated )||extracol;
                %end;
            %if &blipfunction=3 and &lagsetto0at0=0 %then %do;
                extracol=Al1_non_treated # residA_non_treated; /*because at first time point the A_l1 might not be 0*/
                residA_non_treated = (- _pA_t_non_treated)||extracol;
                %end;
            %if &blipfunction=5 %then %do;
                extracol = Apreb_untreated # (- _pA_t_non_treated );
                residA_non_treated = (- _pA_t_non_treated )||extracol;
                %end;
     

          U0i_non_treated = /*ipcw_untreated #*/  Hm_non_treated # (residA_non_treated ) ; 

          newU1_non_treated = J(&non_treated,&npsi,0) ;
          last = 0 ;
          do id=1 to %sysevalf(&non_treated);
               count = tcount_non_treated[id];            
               start = last + 1;
               last = start + count - 1;

               uu_id = U0i_non_treated[start:last,]; 
               usum = uu_id[+,]; /*usum sums over m*/                     
               newU1_non_treated[id,] = usum; /*person id's contribution to estimating equation*/
          end;
          newcov_non_treated = t(newU1_non_treated) * newU1_non_treated ;
          newU1_non_treated_sum = newU1_non_treated[+,] ;
     %end;



     residA_treated = A_treated - _pA_t_treated;   
     %if &blipfunction=3 %then %do;
        extracol = Al1_treated # residA_treated;
        residA_treated = residA_treated||extracol;
        %end;
     /*%if &blipfunction=4 %then %do;
        extracol = ca_Al1_treated # residA_treated;
        residA_treated = residA_treated||extracol;
        %end;*/
     %if &blipfunction=5 %then %do;
        extracol = Apreb_treated # residA_treated;
        residA_treated = residA_treated||extracol;
        %end;

     p_time_treated = nrow(A_treated);              
     hd_time_treated = nrow(Y_treated);


     store  hd_time_treated p_time_treated     
     Y_treated  A_treated /**/%if &blipfunction=3 %then %do; Al1_treated %end;/**/  
                          /*%if &blipfunction=4 %then %do; ca_Al1_treated %end;/**/ 
                          /**/%if &blipfunction=5 %then %do; Apreb_treated Apreb_untreated %end;/**/
     residA_treated   
     pC_t_treated 
     tcount_treated  
     %if %eval(&non_treated) > 0 %then %do;
            U0i_non_treated newU1_non_treated_sum newcov_non_treated tcount_non_treated pC_t_untreated;
            %end;
     ;


     psi1temp=J(1,&npsi,0);
     tempest=mainf(psi1temp);

    load newU newcov;


     create est_truth var {newU newcov tempest};

     append;


     %if &cross_sectional = 1 %then %do;

     tmp    = sum(residA_treated ) / sum(_pA_t_non_treated ) ;
     newpsi = log(tmp);
     test = mainf(newpsi) ;

     create newpsi0 var {sample newpsi test } ;
     append ;

     %end;



     %if &save_weights = 1 and &boosample=0 %then %do;
      
            use _tcount_ ;
            read all var {count} into tcount ;
            individuals = nrow(tcount);
            

          
            use _cero_;
            read all var {pC_t} into pC1 ;
            %if %bquote(&cens2)^= %then read all var {pC2_t} into pC2 ;;
            %if %bquote(&cens3)^= %then read all var {pC3_t} into pC3 ;;
         
    
            p_time = nrow(pC1) ;

            pC_t = J(p_time,%eval(&ncens_var),0);
            %do i = 1 %to %eval(&ncens_var) ;
                pC_t[,&i] = pC&i ;
            %end;


          last = 0;                        

          ipcwnew=J(p_time,max_followup,1);
          do id=1 to %sysevalf(&individuals);   
               count = tcount[id];
               start = last + 1;
               last = start + count - 1;         
        /*CORRECTED WEIGHTS!*/      

                Ctmp=pC_t[start:last];
               ipcwnew_id=J(count,max_followup,0);
               do m = 1 to count ;                            
                    do k = m to max_followup;
                         do j = m  to   k ; 
                            if j<=count then do;
                                /*create the weight W_{m,k}=product j from m to k*/
                                if j=m then do;
                                    tmpw=Ctmp[j,];
                                    end;
                                else do;
                                    tmpw=tmpw#Ctmp[j,];
                                    end;
                                end;/*if j*/
                            end;/*j*/
                            ipcwsep=1.0/tmpw;
                            ipcwnew_id[m,k]=ipcwsep[,#]; /*product of weights for different kinds of censoring*/
                        end;/*k*/

                    /*Wmk_id = ipcwnew;*/
                end;/*m*/ 

               ipcwnew[start:last,] = ipcwnew_id ;
          end; /* end of id loop */        
            
        %if %bquote(&weight)^= %then ipcwnew = ipcwnew # &weight ;;
        varnames={%do k=1 %to &timepoints; W_m&k %end;};
        create _ipcw_ft from ipcwnew[colname=varnames];
        append from ipcwnew ;
quit;
          data _weights_ft ;
          merge _cero_ (keep = &id &time pC_t %if %bquote(&cens2)^=  %then pC2_t ;
                                             %if %bquote(&cens3)^=  %then pC3_t ; 
                                             in_eeq
                                 )  _ipcw_ft  /*(rename = (col1=ipcw))*/;
        
          run;

          proc datasets library = work nolist ;
          delete _ipcw_ft ;
          quit;

   %end; 
%else %do;
    quit;
    %end;

%mend sncftm;





/*****************/
%macro gendatam_cumft (dataout,simulations,n,tpoints,psi1);

  

     data &dataout;
       length  A Am1 L  Y eligible  3;
       keep id A Am1 L Lm1 Y eligible T0 tpoint chk _pA_t 
              %if &useCens1 = 1 %then cens pC_t pC2_t ;
                 ;
         

           sT0 = %sysevalf(&sT0) ;
           sL = %sysevalf(&sL) ;
           sA = %sysevalf(&sA) ;
           sCens1 = %sysevalf(&sCens1);
                                         
        eligible = 1; 
        %if &useCens1 = 1 %then pC2_t = 1 ;; 
        do sim=1 to %sysevalf(&simulations);                       
               do id=1 to %sysevalf(&n); 
                                              
               
               
                    call ranexp(sT0,expvar); 
                    T0=(&t0_mean*expvar);   

                    if T0<30 then I30=1;
                    else I30=0;       

                

                    A = 0;
                    L = 0;
                    Y = 0;
                    tpoint = -1;
                    chk = 0 ;
                    do while ( tpoint <= %eval(&tpoints - 2) & Y = 0); /* changed to tpoints - 2 to have Y1,Y2,Y3,...,Y(tpoint) 
                                         at times tpoint = 0,1,...,&tpoints - 1 */
                          Am1 = A;
                          Lm1 = L ;
                          tpoint = tpoint + 1 ;
                        
                                                                                                     
                              
                           logitL = &gamma0 + &gamma1 * I30 + &gamma2 * Am1 + &gamma3 * Lm1;
                           pL=exp(logitL)/(1+exp(logitL));
                                                   
                           if pL = 1 then do; /*need this because sometimes if pL is 1 will give L1 missing*/
                                  L = 1;
                           end;
                           else do;                          
                                call ranbin(sL,1,pL,L);                            
                           end;
                        
                           logitA = &alpha0 + &alpha1 * L + &alpha2 * Lm1 + &alpha3 * Am1 + &alpha4 * tpoint;
                        
                           _pA_t=exp(logitA)/(1+exp(logitA));                                                                                                                          
                           call ranbin(sA,1,_pA_t,A);                                 
                           
                           /************* add in portion for censoring ******************/
                       
                        
                                  
                        
                            %if %eval(&useCens1) = 1 %then %do;
                                        logitCens = &theta0 + &theta1*A + &theta2*L;
                                        pCens = 1/(1+exp(-logitCens));
                                        pC_t = pcens ;
                                        call ranbin(sCens1,1,pCens,cens);                              
                                        if cens = 1 then do ;
                                              Y = . ;
                                              output ;
                                               
                                        end;
                                        else do ;
                                           chk =chk + exp(&psi1 * A );
                                           if T0 < chk then Y = 1 ;
                                           output ;
                                        end;
                                        if cens = 1 then leave ;
                          %end;                              
                          %else %do;

                              chk =chk + exp(&psi1 * A );
                              if T0 < chk then Y = 1 ;                                                                                              
                              output ;      
                          %end;
                        
                    end; 
              
           

  
                         


               end;  /* end of id loop */
          end;  /* end of sim loop */

             %if &use_stream ne 1 and &usebig = 0 %then %do; 
              
               call symput('sA',sA);
               call symput('sL',sL);
                
                call symput('sT0',sT0);
               
          %end;  
           run;
%mend gendatam_cumft;
%macro numargs(arg);
  %local n arg word; 
   %let n=1;
   %if %bquote(&arg)^=  %then %do;
      %do %until (%qscan(&arg,%eval(&n),%str( ))=%str()); 
        %let word = %qscan(&arg,&n);
          
        %let n=%eval(&n+1);
      %end; 
     %end;
   %eval(&n-1)
%mend numargs;


%macro remove_class;
/* want to remove entries in list2 from list1 */
 %local i j  test word ;
%let i = 1;
%let _nonclass= ;
%do %until(%scan(&_allcov,&i,%str( )) = %str() );
  %let test = %scan(&_allcov,&i);
  %let j = 1;
  %let add = 1;
  %do %until(%scan(&_class,&j,%str( ))=%str());
     %let word = %scan(&_class,&j);
      %if &test = &word %then %let add = 0 ;
      %let j = %eval(&j + 1);
  %end;
  %if %eval(&add) = 1 %then %let _nonclass = &_nonclass &test ;;
  %let i = %eval(&i + 1);
%end;



 %mend remove_class;


%macro newclass ;
     %local old_options ;
     %let old_options = %sysfunc(getoption(mprint)) %sysfunc(getoption(SYMBOLGEN)) %sysfunc(getoption(MLOGIC)) %sysfunc(getoption(NOTES)) ;
     options nomprint nomlogic nosymbolgen nonotes ;
     %local cnt uniquecnt newlist _nonclass _cls_cnt ;

     %let class_var = ;
     %let cov_var = ;
 
     %let _cls_cnt = 0; 
     %if %bquote(&classH)^= %then %do;
          %let class_var = %nrstr(&classH);
          %let cov_var = %nrstr(&covEH);
          %let _cls_cnt = %eval(&_cls_cnt + 1);
     %end;
 
     %if %bquote(&classVH)^= %then %do;
          %let class_var = &class_var %nrstr(&classVH);
          %let cov_var = &cov_var %nrstr(&covVH);
          %let _cls_cnt = %eval(&_cls_cnt + 1);
     %end;
 
     %if %bquote(&classDH)^= %then %do;
          %let class_var = &class_var %nrstr(&classDH);
          %let cov_var = &cov_var %nrstr(&covDH);
          %let _cls_cnt = %eval(&_cls_cnt + 1);
     %end;
 
     %let keyfld = %upcase(%unquote(&class_var));

     /* create a list of the unique elements of all the class variables */

     %let i = 1;
     %do %until(%scan(&keyfld,&i,%str( ) ) = %str() );
          %let var&i = %scan(&keyfld,&i,%str( ));
          %let i = %eval(&i + 1);
     %end;
     %let cnt = %eval(&i - 1);
 
     %let newlist = &var1 ;
     %let in = 1;
  
     %let test = 2;
 
     %do %while (%eval( &test) <=  %eval(&cnt));
          
          %let testword = &&var&test;
          %let add = 1;
          %do i = 1 %to &in;
               %let tmp = %scan(&newlist,&i,%str( ));
               %if &testword = &tmp  %then %do;
                    %let add = 0;
                    %let i = %eval(&in + 1);
               %end;
          %end;
          %if &add = 1 %then %do;
               %let newlist = &newlist &testword ;
               %let in = %eval(&in + 1);
          %end;
    
          %let test = %eval(&test + 1);
     %end;
 
     %let uniquecnt = &in;


     /* find the number of levels for each class variable.
          store the results in a macro variable */

     %do varnum = 1 %to &uniquecnt ;

          %let vartmp = %scan(&newlist,&varnum);
          proc freq data= _cero_ noprint;
          tables &vartmp / out = &vartmp;
          run; 

          data &vartmp;
          set &vartmp (drop = count percent) end = _end_;
          by &vartmp;
          if first.&vartmp then _levels_ + 1;
          call symput("_&vartmp"||left(put(_levels_,2.)),&vartmp);
          if _end_ then do;
              call symput("_&vartmp"||'levels',_levels_);
          end;
          retain _levels_;   
          run;
      

               
          data _class_levels;
          set _class_levels ;
          &vartmp = &&_&vartmp.levels ;
          run;

          proc datasets library = work nolist;
          delete &vartmp;
          run; quit;
     %end;
     /* create the new class variables in the original data set */     
     data _cero_;
     set _cero_;
     %do varnum = 1 %to &uniquecnt;

          %let vartmp = %scan(&newlist, &varnum);
          %let new_levels = %eval(&&_&vartmp.levels - 1);

          %do i = 1 %to &new_levels;
               _&&vartmp.&i = 0;
               %let holder = &&_&vartmp.&i;

               if &vartmp = "&holder" then _&&vartmp.&i = 1;
          %end;

     %end; /* varnum */
     run;

     /* create the new lists of covariates which will include the re-coded
     class variables */


     %do i = 1 %to  &_cls_cnt ;             
          %let var = %qscan(&class_var, &i,%str( ));
          %let cvar = %qscan(&cov_var,&i,%str( ));
          %let _class = %unquote(&var);
          %let _allcov = %unquote(&cvar);
          %remove_class;
  
          %let j = 1;
          %let in = 1;
          %let tmp = ;
          %let ctmp= ;
          %do %until(%scan(%unquote(&var),&j,%str( ) ) = %str() );
               %let vartmp = %scan(%unquote(&var),&j,%str( ));
               %let j = %eval(&j + 1);
               %let new_levels = %eval(&&_&vartmp.levels - 1);
               %do k = 1 %to &new_levels;
                    %let tmp = &tmp _&&vartmp.&k ;
               %end;
          %end;

          %let %substr(&cvar,2) = &_nonclass &tmp;
      %end;
     options &old_options ;
%mend newclass;
%macro plot_graph;
            %if %eval(&save_graph) ^= 0 and %bquote(&graph_path)^= %then 
                                            filename grafout "&graph_path..&ftype";;
               goptions reset=global 
                      %if %eval(&save_graph) ^= 0 %then %do;
                        device = &gdevice
                          gsfname = grafout
                          gsfmode = replace
                         %end;
                     gunit=pct 
                     border 
                     cback=white
                          colors=(black blue green red)
                          ftitle=swissb ftext=swiss htitle=4 htext=2;

               symbol1 color=black
                    interpol= spline /* none step needle join r */
                         value= none
                         line=1


                       ;

 
 
               %if &npsi = 1 %then %do;

                    proc gplot data = _graph_;
                         plot test*psi1/ 
                             /*haxis = &Psi1lower to &Psi1upper 
                              %if %eval(&pts1) < 20 %then by &Psi1step; */
                              hminor = 9 
                                   vminor = 4 grid
                                   %if %eval(&est_method) ^= 3 %then vref= &chi_level ;
                                   %else vref = 0 ;
                              cvref= red;
                        run;
          
               %end;

               %else %if &npsi = 2 %then %do;

                    proc g3d data= _graph_;
                         plot psi2*psi1= test /   grid tilt= 70 rotate=-30 0 30 60 90
                         /*
                                        xticknum=%sysevalf(((&Psi1upper- &Psi1lower)/&Psi1step)+1) 
                                        yticknum=%sysevalf(((&Psi2upper- &Psi2lower)/&Psi2step)+1) 
                                        zticknum=6
                         */
                                        zmin=0 zmax= &chi_level
                                        ctop=green cbottom=red;
                        run;

     /*added by Sally*/ %if %eval(&save_graph) ^= 0 and %bquote(&graph_path2)^= %then 
                                            filename grafout "&graph_path2..&ftype";;
               goptions reset=global 
                      %if %eval(&save_graph) ^= 0 %then %do;
                        device = &gdevice
                          gsfname = grafout
                          gsfmode = replace
                         %end;
                     gunit=pct 
                     border 
                     cback=white
                          colors=(black blue green red)
                          ftitle=swissb ftext=swiss htitle=4 htext=2;

               symbol1 color=black
                    interpol= spline /* none step needle join r */
                         value= none;

                       
                /*if there are 2 parameters then &chilevel=5.99*/;
                    proc gcontour data= _graph_;
                         plot psi2*psi1= test 
                            /   grid 
                                levels=0 to 6 by 1
                                llevels=1 2 3 4 8 14 1
                                clevels=black blue green red brown magenta black;
                        run; 

               %end;     


%mend plot_graph;

%macro find_root;
     
     /* for one dimensional psi */
    start mainfeq(psi);
      tmp = mainf(psi);
      load newU;
      return (newU);
     finish mainfeq;
     leftbnd = %sysevalf(&s1-&srchbnd);
     rightbnd = %sysevalf(&s1 + &srchbnd);
     lf = mainfeq(leftbnd);
     rf = mainfeq(rightbnd);
     chk = lf * rf;
    delta = %sysevalf(2*&srchbnd/&pts1 );
     if chk > 0 then do;

          /* print leftbnd lf rightbnd rf chk;*/
          newleftbnd = leftbnd;
          newrightbnd = newleftbnd + delta;
          rf = mainfeq(newrightbnd);
          chk = lf * rf;

          do while ( chk > 0 & newleftbnd < rightbnd );
               newleftbnd = newrightbnd;
               lf = rf;
               newrightbnd  = newrightbnd +  delta;
               rf = mainfeq(newrightbnd);
               chk = lf * rf;
               /* print newleftbnd lf newrightbnd rf chk; */
          end;

          if newrightbnd <= rightbnd & chk < 0 then do;
          leftbnd = newleftbnd;
               rightbnd = newrightbnd;
               delta = rightbnd - leftbnd;
          end;

        else print "Bad starting points. Can not find end points in [%sysevalf(&s1-&srchbnd) , %sysevalf(&s1+&srchbnd)] having opposite signs ";

     end;

     delta = rightbnd - leftbnd;
     eval = 0;
     if chk < 0 then do; /* the end points have different signs */

        middle = (leftbnd + rightbnd)/2.0;
        mf = mainfeq(middle);

        do while ( abs(mf) > 0.0001 & delta >  1.e-6 );
            delta = rightbnd - leftbnd;
            middle = (leftbnd + rightbnd)/2.0;
            mf = mainfeq(middle);
               /* print eval leftbnd rightbnd lf rf mf; */
            if lf*mf < 0 then do;
                rightbnd = middle;
                  rf = mf;
             end;
             else do;
               leftbnd = middle;
                lf = mf;
             end;
             eval = eval + 1;
         end;

         xr = middle;
         objfunc = mf;
         
        
      
     end; /* chk < 0 */

     else do;

         xr = .;
         objfunc = .;
        
      

     end; /* could not find any good starting points */

%mend find_root;


%macro sim_ann;
           
     proc iml;
        
          start find_t0(usrfunc,dimen,miter,t0,lbd,ubd,xmin,ymin);

               sdof = 0.0;
               range = ubd - lbd;
               newrho = &rho;
               init = (lbd + ubd)/2.0;
               x = J(1,dimen,init);
               xnew = J(1,dimen,0);
               call execute("y0 =" + usrfunc + "(x);");
               ymin = y0;
               yold = y0;
               xmin = x;
               inc = 0;

               do i=1 to miter;

                    do j=1 to dimen;
                         t1 = ranuni(300);
                    /*   theta = 2.0 * range*theta - range; */
                         theta = lbd[j] + range[j] * t1;                            
                         xnew[j] =   theta;
                    end;

                    call execute("ynew =" +  usrfunc + "( xnew );");

                    if ynew < ymin then do;
                    ymin = ynew;
                         xmin = xnew;
                    end;

                    if ynew >  yold  then do;
                         inc = inc + 1;
                         dof = ynew - yold;
                         sdof = sdof + dof;
                    end;

                    x = xnew ;
                    yold = ynew;

               end;

               dfave = sdof/inc;
               t0 = 0.22314*dfave;

               return;

          finish find_t0;

          store module=find_t0;

          /* iterate a few times at the present temperature to reach thermal equilibrium */
          start equilibrate( usrfunc,dimen,t, n,x,xnew,xbest,y,ynew,ybest,lbd,ubd);

               range = &rrange;
               newrho = &rho;
               equil = 0;
               
               xtmp = J(1,dimen,0);
               delta = 1.0;

               do  i = 1 to  n ; 
                    do while (equil < 10);
                         do j = 1 to dimen;

                         theta = ranuni(300);
                         theta = 2.0 * range*theta - range;
                              xc = newrho * t * tan ( theta ); 
                              xnew[j] = x[j] + xc;
                              if xnew[j] > ubd[j] then xnew[j] = ubd[j];
                              if xnew[j] < lbd[j] then xnew[j] = lbd[j];
                          
                         end;
                         /* "energy" */
                         call execute("ynew =" +  usrfunc + "( xnew );");
                    
                         if dimen = 1 then do;
                              p1 = xnew[1];
                         end;

                         else if dimen = 2 then do;
                              p1 = xnew[1];
                              p2 = xnew[2];
                         end;
                    
                         c = ynew - y;
                         if c < 0.0 then do; /* keep xnew if energy is reduced */

                              xtmp = x;
                              x = xnew;
                              xnew = xtmp;
                              y = ynew;
                              if y < ybest  then do;   
                                   do j = 1 to  dimen; 
                                        xbest[j] = x[j];
                                   end;
                                   ybest = y;
                              end;
                      
                              delta = abs( c );
                              if y ^= 0.0 then delta = delta/y; ;
                              else if ynew ^= 0.0 then delta = delta/ynew; ;
                          
                              /* equilibrium is defined as a 10% or smaller change
                              in 10 iterations */
                              if  delta < 0.10 then equil = equil + 1; ;
                              else equil = 0; ;

                         end;

                         else do;
                         equil= equil + 1;
                         end; /* end of c < 0 */
                                        
                    end;
                     
                    if equil > 9 then return (i);

               end;
           
               return (i);

          finish equilibrate;

          store module = equilibrate;

          /* cool the system with annealing */
          start anneal(usrModule,dimen, iter,x,t0,lbd,ubd);

              xtmp = J(1,dimen,0);            
               xnew = J(1,dimen,0);
               y = J(1,1,0);
               ynew = J(1,1,0);
               ybest = J(1,1,0);
               call execute("y =" + usrModule + "( x );");
               xbest = x;
              ynew = y;
               ybest = y;
               newrho = &rho;
               range = &rrange;
               newtscale = &tscale;
              n = iter;
               
              ntmp = %eval(10*&ddwell);
               tmp =  equilibrate( usrModule,dimen,t0, ntmp ,x,xnew,xbest,y,ynew,ybest,lbd,ubd);
               told = t0;
           
               /* do i = 1 to n;  */
                t = t0;
                i = 0;
                do while ( ( i <= n | t > 1.0 ) & ybest > 0.0001);
                   i = i + 1;
                    t = t0 /(1.0 + i * newtscale ); 
                    
                    told = t;
                    tmp = equilibrate(usrModule,dimen,t, &ddwell,x,xnew,xbest,y,ynew,ybest,lbd,ubd); 
               /*   print i t xbest ybest ; */
                    do j = 1 to  dimen;

                       theta1 = ranuni(100);
                         theta = 2.0*range*theta1 - range; 
                         xc =  (newrho * t * tan(theta)); 
                         xnew[j] = x[j] + xc;
                         if xnew[j] > ubd[j] then xnew[j] = ubd[j];
                         if xnew[j] < lbd[j] then xnew[j] = lbd[j];

                     end;
                    /* "energy" */
                    call execute("ynew =" + usrModule + "( xnew );");
                    
                    if dimen = 1 then do;
                         p1 = xnew[1];
                         
                    end;

                    else if dimen = 2 then do;
                         p1 = xnew[1];
                         p2 = xnew[2];
                         
                    end;
                    
                    c = ynew - y;
                     
                    if ynew<=y then do; /* keep xnew if energy is reduced */
               
                         xtmp = x;
                         x = xnew;
                         xnew = xtmp;
                         y = ynew;

                         if y<ybest then do;
                         
                              do  j = 1 to dimen; 
                                   xbest[j] = x[j];
                              end;
                              ybest = y;
                         
                         end;

                    end;

                    else do;
                    
                    /* keep xnew with probability, p, if ynew is increased */
                         p = exp( - (ynew - y) / (&K * t) );
                      theta = ranuni(200);
                         if ( p >  theta ) then do;
                         
                              xtmp = x;
                              x = xnew;
                              xnew = xtmp;
                              y = ynew;

                          end;

                     end;
 
               end;

              ntmp = %eval(10 * &ddwell);
               tmp = equilibrate(usrModule,dimen,t, ntmp ,x,xnew,xbest,y,ynew,ybest,lbd,ubd);
               x = xbest;
               
               return (t);

          finish anneal;

          store module=anneal;
      
     quit;

%mend sim_ann;
%macro gridsearch;


index_chk = J(1,&npsi,0);
final_index = J(1,&npsi,%eval(&pts1 ) );

index_holder = &npsi;
last_index = &npsi;
ok = 1;

%if %bquote(&starting_point)^= %then %do;
   %if %bquote(&user_range)^= %then %do;
      start = {&starting_point } - half_range ;  /* defined from the individual snmm macro */
   %end;
   %else %do;
      start = { &starting_point } - %sysevalf(&srchbnd / 2.0 ) ;
   %end;
%end;
%else %do ;
  start = J(1,&npsi,0) - %sysevalf(&srchbnd / 2.0 ) ;
%end;

xmin = start;
ymin = mainf(xmin);

%if %bquote(&user_range)^= %then %do;
   delta = { &user_range } / %eval(&pts1) ;
%end;
%else %do;
 delta = %sysevalf(&srchbnd / &pts1 );
%end;

/* delta = %sysevalf( 2.0* &srchbnd / &pts1 ); */

create _graph_ var {  %do i = 1 %to &npsi; psi&i %end; test };
do while (ok);

  do i=0 to &pts1;
      index_chk[last_index] = i;
      psi = start + delta # index_chk ;
       
      test = mainf(psi);
      
       if test < ymin then do;
         ymin = test;
         xmin = psi ;
       end;
       __gtmp_ = psi || test;
       append from  __gtmp_ ;
  end; /* i */
  good = 1;
  
  index_holder = last_index - 1;
  
  if index_holder = 0 then do;
     good = 0;
      ok = 0;
  end;
   
  do while (good = 1  );
        
        if  (index_chk[index_holder] < final_index[index_holder]) then do; 
           index_chk[index_holder] = index_chk[index_holder] + 1;
           do i = index_holder + 1 to last_index;
             index_chk[i] = 0;
           end;
           good = 0;
         end;
         else do;
           index_holder = index_holder - 1;  
           if index_holder = 0 then do; /* done with grid */
              good = 0;
              ok = 0;
           end;     
         end;
          
  end; /* while good */
end; /* while ok */


%mend gridsearch;



 



  

%macro output_results;

%let get_cov = 0;

%if %eval(&zero_one) = 1 %then %do;
   proc freq data = _cero_ noprint;
   table &outcome /out = ycount (drop = percent);
   run;

   data ycount;
   set ycount;
   if &outcome = 1 then call symput('num_events',count);
   run;

   proc datasets library = work nolist;
   delete ycount;
   quit;

%end;


%if %eval(&npsi)^= 1 & %eval(&bootstrap) = 0 %then %do;
  proc transpose data = _snmm&regime._ out = _st_ (rename = ( col1 = estimate _name_ = name ));
  var %do i = 1 %to &npsi ; psi&i %end; ;
  by sample _bsample_ objfunc
  %if %eval(&grid_search) = 1 %then %do;
     grid_min
      %do i = 1 %to &npsi;
        grid_psi&i
      %end;

  %end;
  ;
  run;

%end;
%if %eval(&bootstrap) = 1 & %eval(&npsi)^= 1 %then %do;
   data _st_;
   set _snmm&regime._;
   format p_value PVALUE8.4 ;
   %do i = 1 %to &npsi;
     name = "psi&i";
     estimate = psi&i;
     mean = mean&i;
     std = std&i;
     bbias = bbias&i;
     z = z&i;               /*isn't z scalar? there shouldn't be a separate one for each psi.*/
     lb = lb&i;
     ub = ub&i;
      p_value = p_value&i ;
     output ;
   %end;
   keep sample name estimate mean std bbias z p_value lb ub objfunc 
     %if %eval(&grid_search) = 1 %then %do;
          grid_min
           %do i = 1 %to &npsi;
             grid_psi&i
           %end;
      %end;
      ;
%end;


    %if %eval(&blip_type) = 0 %then %do;
        %if %eval(&user_rho) = 1 %then %do;
        
       %end;
       %else %do;
         %let cov_psi = &treat;
           %do i = 2 %to &npsi ;
               %let j = %eval(&i - 1) ;
                %let vartmp = %scan(&psivar,%eval(&j));
               %let cov_psi = &cov_psi  &treat*&vartmp ;
           %end;
       %end; 
    %end;
     %else %let cov_psi = &treat ;


   data _st_;
   set %if %eval(&npsi) = 1 %then  %do;
           _snmm&regime._ (rename = (psi1 = estimate 
           %if %eval(&conf_interval) = 1 %then psi1_low = lb psi1_high = ub ;
                      )) 
          %end;
        %else _st_ ;
        ;
   by sample ;
   if first.sample then do;
    start_row = _n_;
   end;
   _psi = _n_ - start_row + 1;
   
   /*SP modified next chunk in order to print actual variable name used 
   for effect modification, which has been sort of hacked in under the
   name of the first effect modifier programmed but this should not be
   apparent to the user.*/
   /*%do i = 1 %to &npsi;
       /*%let vartmp = %qscan(&cov_psi,%eval(&i),' ');
        if _psi = %eval(&i) then name = "&vartmp" ;*/
   /*%end;*/
    if _psi = 1 then name = "&treat1";
    %if &npsi=2 %then %do;
        %if &blipfunction=3 %then %do;
            %let vartmp = %qscan(&cov_psi,2,' ');
            if _psi = 2 then name = "&vartmp" ;
            %end;
        %if &blipfunction=5 %then %do;
            %if %bquote(&blem)^= %then %do;
                if _psi = 2 then name = "&treat1*&blem";
                %end;
            %else %do;
                %if %bquote(&tvem)^= %then %do;
                    if _psi = 2 then name = "&treat1*&tvem";
                    %end;
                %else %do;
                    %let vartmp = %qscan(&cov_psi,2,' ');
                    if _psi = 2 then name = "&vartmp" ;
                    %end;
                %end;
            %end;
        %end;

   retain start_row;
   run;


     %local for_crr for_crr2 ;
     %if &get_cov = 1 %then %do;

          data _cov ;
               merge _cov name_order;
               by name;
          proc sort data = _cov out = _cov (drop = order );
               by  order;
          run;

          data _cov;
               set _cov;
               if sample = . then sample = 1;
          run;

     %end; 

     data all_results&regime;
          set _st_   %if &get_cov = 1 %then  _cov;;
          if sample = 1;
           
            est = left(estimate);
            %do i = 1 %to &npsi;
               if _n_ = &i  then call symput("psiopt&i",estimate);
            %end;
           
     run;
    
     
     %if %eval(&grid_search) = 1   %then %do;

        %let linegrid0 =Grid search: ;
          %let forgrid = grid_psi1 8.4;
          %if %eval(&npsi)^= 1 %then %do;
               %let forgrid = '(' grid_psi1 8.4  ; 
               %do i = 2 %to %eval(&npsi);
                    %let forgrid = &forgrid   ',' grid_psi&i 8.4 ;
                    %end;
                    %let forgrid = &forgrid ')';
         %end;
        %if %eval(&est_method) = 0 %then %let indent = @32 ;
     %end;
   
     %if  %eval(&est_method)^= 0 %then %do;
        %let indent = @35 ;
          %if %eval(&est_method) = 1       %then %let method = Nelder-Mead Simplex: ;
          %else %if %eval(&est_method) = 2 %then %let method = Newton-Raphson method: ;
          %else %if %eval(&est_method) = 3 %then %let method = Simulated Annealing ;
          %else %if %eval(&est_method) = 4 %then %let method = Bisection method ;  
          %else %if %eval(&est_method) = 5 %then %let method = Analytic Solution ; 
     %end;

     
       %let firstline = @16;

     
        %let line0 = Structural Nested Cumulative Failure Time Model;
          %let line01 = ;
          %let line02b = ;
          
          %if %eval(&dr) = 1 %then %do;
            %if %eval(&new_parts) = 1 %then %do;
               %if %eval(&blip_q) = 0 %then %do; 
                     %let line01a = Using doubly-robust efficient estimator with S_eff  and pieces for estimation;
                     %let line01b =  of parameters from the model for E[&treat | covA] and E[H | covH ]  ;
                %end;
                %if %eval(&blip_q) = 1 %then %do;
                      %let line01a = Using doubly-robust estimator without S_eff, while including the extra pieces for estimation;
                      %let line01b = of parameters from the models for  E[&treat | covA] and E[H | covH ] ;
                 %end;
             %end;
           %if %eval(&new_parts) = 0 %then %do;
                    %if %eval( &blip_q) = 0 %then %do;
                       %let line01a = Using doubly-robust estimator using the effinceint S_eff, but not including the extra ;
                       %let line01b = pieces for estimation of parameters from the models for E[&treat | covA] and E[H | covH ] ;
                      %end;
                  %if %eval(&blip_q) = 1 %then %do;
                       %let line01a = Using doubly-robust estimator without  S_eff and the  pieces for estimation of parameters;
                       %let line01b = from models for E[&treat | covA] and E[H | covH ]   ;
                      %end;
            %end;
          %end;
          %else %do;
           %let line01a = Using non-doubly robust estimator and non-efficient S_eff ;
             %let line01b = ;
          %end;
          %if %eval(&user_rho) = 0 %then %do;
            %if %eval(&blip_type) = 0 %then %let line02 = Using linear blip function gamma = &treat * psi * (1  &psivar) ; 
               %if %eval(&blip_type) = 1 %then %let line02 = Using linear blip function &treat * psi ;
          %end;
          %else %if %eval(&user_rho) = 1 %then %let line02 = Using user defined blip function ;
     

     %if &npsi =1 %then  %let vartype = v;
     %else %let vartype = cov ;

    %if %eval(&npsi) = 1 %then %let for_est_line = _psi1 8.4 ;
    %else %do;
          %let for_est_line = '(' _psi1 8.4 ;
          %do i = 2 %to &npsi;
               %let for_est_line = &for_est_line ',' _psi&i 8.4 ; 
          %end;
            %let for_est_line = &for_est_line ')' ;
    %end;
     %if &bootstrap = 0 %then %do;         
               %if &conf_interval = 1 & &get_cov = 0 %then %do;                               
                %let line01 = ;
                    %let line02 = Inverting test to find confidence interval ;
                  %let line1 = @35 '   95% Confidence' ;
                  %let line2 = @10 'Parameter'  @25 'Estimate'   @35 '       Limits'  ;
                  %let line3 = @10 name @25 estimate 8.4   @35 lb 8.4 @45 ub 8.4     ;
               %end; /* conf_interval = 1 */
              
              %else %if &get_cov = 1 %then %do; 
                    %let line01 =  empirical estimates of mean and variance using &mcsamples samples.;
                   %let line1 =   @40 'Standard' @50 '   95% Confidence' ; ;
                  %let line2 = @5 'Parameter'  @20 'Estimate'  @30 '   Mean' @40 '   Error' @50 '       Limits' @70 '    Z' @80 'Pr > |Z|' ;     ;
                  %let line3 = @5 name @20 estimate 8.4 @30 mean1 8.4 @40 std1 8.4 @50 lb1 8.4 @60 ub1 8.4   @70 Z1 8.4 @80 p_value1 ;      ;
               %end;

              %else %do;
                    %let line1 = ;
                  %let line2 = @10 'Parameter'  @25 'Estimate'    ;
                  %let line3 = @10 name @25 estimate 8.4   ;
              %end;            
       
     %end;
     %else %do;
          %let line01 = Bootstrap Mean and &vartype.ariance using &nboot bootstrap samples ; 
          %let line02b = &bootstraps_that_converged samples, includeing original sample, converged to less than &conf_min  ;
          %let line1 = @35 'Standard' @45 '   95% Confidence' ;
          %let line2 =  @1 'Parameter'  @15 'Estimate' @25 '   Mean' @35 '   Error' @45 '       Limits' @65 '    Z' @75 'Pr > |Z|' ;
          %let line3 =  @1 name @15 estimate 8.4 @25 mean1 8.4 @35 std1 8.4 @45 lb1 8.4 @55 ub1 8.4   @65 Z1 8.4 @75 p_value1 ;     
     %end;

    

     data _null_;
          set all_results&regime end = _end_ ;
          file print notitles;
          if _n_ = 1 then put &firstline "&line0" //  @5 "&line01a" %if %eval(&zero_one) = 1 %then  @65 "Number of events : &num_events" / ;/*%if %eval(&dr) = 1 %then / @5 "&line01b"*/ ;  
                                                   
          %if %bquote(&data)= %then // @5 'Sample = 1 ' /;
          ;
          %if %eval(&dr) = 1 %then %do;
                              
          %end;
          %if %eval(&user_regimes) = 1 %then %do;
               %let regime_tmp = %scan(&user_defined_regimes,%eval(&regime),%str(' '));
               if _n_ = 1 then put @5 "Results are for user defined regime : &regime_tmp " / ;
          %end;
          %if   ( 0 <  %eval(&est_method) < 4 )     %then %do;
             if _n_ = 1 then put // @5 'Minimum value of estimating function' ;
          %end;
          %if %eval(&grid_search) = 1 & %eval(&est_method) ne 4 %then %do;
              if _n_ = 1 then put @10 'Grid search: ' &indent grid_min 8.4  ' at psi = ' &forgrid ;
          %end;
          %if %eval(&est_method)^= 0  %then %do;            
               
               %if %eval(&est_method)^= 4 %then %do;
                       %do i = 1 %to &npsi;
                               _psi&i = &&psiopt&i;
                       %end;                       
                       if _n_ = 1 then  put @10  "&method" &indent objfunc  e8.4 ' at psi = ' &for_est_line     //                  
                           %if %eval(&conf_interval) = 1 %then  @5 "&line02" // ;;
                %end;
                %else %do;
                   if _n_ = 1 then put @10 'Bisection method yields estimating equation = ' objfunc e8.4 ' at psi = ' estimate 8.4; 
                      if _n_ = 1 then put @10 'Confidence intervals can not be calculated using this method.';
                %end;
          %end;
      
          

          %if %bquote(&line01)^= %then %do;
             if _n_ = 1 then put  @10 "&line01" / %if %bquote(&line02b)^= %then  @10 "&line02b" / ; ;
          %end;
          
           %if %eval(&mean_of_h) = 1 %then 
               if _n_ = 1 then put  @5 "Average value of H is " mean_h // ;;

         /* headings for main results */

          if _n_ = 1 then put // &line1 /  &line2 /;
          if _n_ <= %eval(&npsi) then put &line3;

          
          if _n_ > %eval(&npsi) then do;
          %if %eval(&npsi) < 5  & &get_cov = 1  
         %then %do;
               if _n_ = %eval(&npsi+1) then put // @40 "&vartype.ariance matrix " // @10 'Parameter' 
             %let start  =  30;
             %do i = 1 %to &npsi;
         
                    %let vartmp = %scan(&cov_psi,%eval(&i)); 
                    %let varlength&i = %length(&vartmp);
          
                    %if %eval(&&varlength&i) < 11 %then %do ;
                         %let spaces = %eval(10 - &&varlength&i);
                         %if %eval(&spaces) > 1 %then %let spaces = %eval(&spaces / 2) ; 
                    %let position  = %eval(&start + &spaces) ;   
               %end;

                    %else %do;
                    %let position  = %eval(&start) ;
                    %let vartmp = %substr(&vartmp,1,9);
                    %end;
               %let start = %eval(&start + 10) ;
                    @%eval(&position) "&vartmp"  
               %end;

               // ;
               if _n_ >= %eval(&npsi + 1) then put @10 name 
             %let start = 30;

             %do i = 1 %to &npsi;
                    %let vartmp = %scan(&cov_psi,%eval(&i));
                    @%eval(&start) &vartmp 8.4 
                    %let start = %eval(&start + 10);
               %end;;  /* put */

          %end; /* n_betas < 5 */

          %else %do;
               if _n_ = %eval(&npsi + 2) then do;
                    
                    %if &get_cov =1   %then put // @10 'Covariance matrix is contained in data set cov_betas' ;;
               end;
          %end;
 
          end;
 
     run;

     %if %eval(&bootstrap) = 1 %then %do;
      %if %eval(&bootstraps_that_converged - 1) < %eval(&nboot)  %then %do;
         data _null_;
           
          file print notitles;
          if _n_ = 1 then put @10 "Warning: Since fewer than &nboot + 1 samples converged to at least &conf_min check " /
                              @10 "the bounds used in the optimization routines and the log file for any warnings or " /
                                   @10 "any error statements concerning the convergence of the optimization method or the calculation" /
                                   @10 "of any weights or probabilities used in the estimating equations " ;
        run;


      %end;
  %end ;
%mend output_results;    
%macro find_estimate ;
proc iml %if &_bigdata = 1 %then worksize = 100000000 ; ;  

%if %eval(&mean_of_H) = 1 %then meanH = 0 ;;

load module = (%if &failure_time = 0 %then rev_cumsum ;
                  mainf) ;
%if %eval(&repeated) = 1 %then load module = mylog ;;
optn = {0 &optn_level};
 
 /* options: minimization, print (all: 5, nothing: 0 */

      
        
          con = J(2,&npsi,0);

%if %bquote(&starting_point)^= %then start = { &starting_point } ;
%else start = J(1,&npsi,0); ;

%if %bquote(&user_range)^= %then %do;
        half_range = { &user_range} / 2.0;
        con[1,] = start - half_range;
        con[2,] = start + half_range;

%end;
  %else %do;
        %do i=1 %to &npsi;
             con[1,&i] =  start[&i] - &srchbnd ;
             con[2,&i] =  start[&i] + &srchbnd ;
          %end;
  %end;   

%if &check_func = 1 %then %do;
      func = mainf(start);
      load newU;
%end;
  
 


%if &grid_search = 1 %then %do;

   %gridsearch ;
   %if &est_method^= 0 %then start = xmin;;
   %if &est_method = 0 %then %do;
     %do i = 1 %to &npsi;
       psi&i = xmin[&i];
     %end;    
      objfunc = ymin ;
   %end;

   grid_min = ymin;
   %do i = 1 %to &npsi;
         grid_psi&i = xmin[&i] ;
   %end;
   %if %eval(&optn_level)^= 0 %then print xmin ymin;;
%end;
   
%if &est_method = 0 & &grid_search = 0 %then %do;

   
    %if &check_func = 0   %then %do;
       psi1 = . ;
       objfunc = . ;
    %end;
    %else %do ;
       %do i =1 %to &npsi ;
           psi&i = start[&i] ;
         %end;
       objfunc = . ;       
    %end;
%end; 
 
%else %if &est_method = 1 %then %do; 

    %if %eval(&repeated) = 0 AND %eval(&use_loop) = 1 %then %smooth_indicator ; /* smooth indicator in optimal regime */
    
    optn = {0 &optn_level};  
     _done_ = 0;
     _called_ = 0 ;
     do while ( _done_ = 0) ;
       call nlpnms(rc,xr,"mainf",start,optn,con); 
        _called_ = _called_ + 1 ;
       if rc > 0 then  _done_ = 1 ;
        else do;
           if rc = -5 | rc = -8 then do ;          
                if rc = -8 & _called_ = 2 then do;
                 print "Warning: Optimization needed more than maximum number of function evaluations for two calls to nlpnms";
                    _done_ = 1 ;
               end;
              if rc = -5 & _called_ = 5 then do ;
                 print "Warning: Optimization routine (nlpnms) was called 5 times and convergence was not obtained. Please ";
                  print "check bounds and starting point used and the log file for any warnings or errors";
                  _done_ = 1 ;
              end;

              start = xr ;
            %if %bquote(&user_range)^= %then %do;             
                con[1,] = start - half_range;
                con[2,] = start + half_range;
            %end;
            %else %do;               
                  con[1,] =  start - &srchbnd ;
                  con[2,] =  start + &srchbnd ;
             %end;
         end;
         else do;
            print "Problem with optimization routine (nlpms). Return code rc = " rc ;
             _done_ = 1 ;
         end;
       end;
   end;

    store xr; 
    objfunc = mainf(xr);
    %do i = 1 %to &npsi;
      psi&i = xr[&i];
    %end; 
%end ; /* end of est_method = 1 */
%else %if &est_method = 2 %then %do;
    optn = {0 &optn_level};  
   
     _done_ = 0;
     _called_ = 0 ;
     _printed_warning = 0 ;
     do while ( _done_ = 0) ;
       call nlpqn(rc,xr,"mainf",start,optn,con); 
        _called_ = _called_ + 1 ;
       if rc >= 0 then _converged_ = 1 ;
        if rc < 0 & _printed_warning = 0 then  do ;
           print "Problem with optimization routine in _bsample_ = &bsample . Return code rc = " rc ;
            _printed_warning = 1;
        end;
        if rc = -5 | rc = -8 then do ; 
          
            if rc = -8 & _called_ = 2 then do;
              print "Warning: Optimization needed more than maximum number of function evaluations for two calls to nlpnms";
                 _done_ = 1 ;
            end;
            if rc = -5 & _called_ = 5 then do ;
              print "Warning: Optimization routine (nlpqn) was called 5 times and convergence was not obtained. Please "
               print "check bounds and starting point used and the log file for any warnings or errors";
               _done_ = 1 ;
            end;

            start = xr ;
          %if %bquote(&user_range)^= %then %do;             
              con[1,] = start - half_range;
              con[2,] = start + half_range;
          %end;
          %else %do;               
                con[1,] =  start - &srchbnd ;
                con[2,] =  start + &srchbnd ;
           %end;
       end;
       else do;          
          _done_ = 1 ;
       end;
   end;


 
   %do i = 1 %to &npsi;
     psi&i = xr[&i];
   %end;
   objfunc = mainf(xr);
%end; 
%else %if &est_method = 3 %then %do;
          t0 = 10;
          lbd = con[1,] ;
          ubd = con[2,] ;                     
          call find_t0("mainf",&npsi,&t0_eval,t0,lbd,ubd,xmin_sa,ymin_sa);

          %if %eval(&grid_search) = 1 %then %do;
                xr = xmin;                        
                if ymin_sa < ymin then xr = xmin_sa;
          %end;
          %else %if %eval(&grid_search)^= 1 %then xr = xmin_sa;;
     
                 t= anneal("mainf",&npsi,&sim_eval,xr,t0,lbd,ubd);
                    objfunc = mainf(xr); 
                    
                    psi1 = xr[1]; 
                    psi2 = 0;
               
                %if %eval(&npsi) = 2 %then psi2 = xr[2];;             
                     print 'Optimization results when using Simulated Annealing to estimate the paramters';
                 print t0 t xmin_sa ymin_sa psi1 %if &npsi = 2 %then psi2 ; objfunc ; 
                     
                    /* 95% CI for one dimensional psi */
                    %if %sysevalf(&npsi) = 1  & %eval(&conf_interval) = 1 %then %do;
                    start  mainf2(psi);
                           tmp = mainf(psi);
                           out = (tmp - &chi_level)**2.0;
                           return (out);
                         finish mainf2;
                     print 'Results for finding confidence interval'; 

                         lbd_lower = con[1,];
                         ubd_upper = con[2,];
                         lbd_upper = xr[1];
                         ubd_lower = xr[1];
                         call find_t0("mainf2",&npsi,&t0_eval,t0,lbd_lower,lbd_upper,xmin_sa,ymin_sa);
                         xrlow = xmin_sa;
                    t= anneal("mainf2",&npsi,&sim_eval,xrlow,t0,lbd_lower,lbd_upper);
                        objfunc_low = mainf2(xrlow);
                         psi1_low = xrlow[1];
                         
                     print  psi1_low objfunc_low; 
                         
                         call find_t0("mainf2",&npsi,&t0_eval,t0,ubd_lower,ubd_upper,xmin_sa,ymin_sa);
                         xrhigh = xmin_sa;
                    t= anneal("mainf2",&npsi,&sim_eval,xrhigh,t0,ubd_lower,ubd_upper);
                              
                        objfunc_high = mainf2(xrhigh);
                         psi1_high = xrhigh[1];
                         
                     print  psi1_high objfunc_high 
           
                    %end;

                    %else %do;
                         psi1_low = 0; psi1_high = 0;
                    %end;
%end;
%else %if &est_method = 4 %then %do;
   %find_root ;
   psi1 = xr[1];
%end;
  


 
%if %sysevalf(&npsi) = 1 AND &conf_interval = 1 %then %do;

      /* first find rough estimate of confidence interval by stepping abs(xr) in both
         directions */

      /* find lower estimate */

      start for_conf(x);
       tmp = mainf(x) -&chi_level;
        return (tmp);
      finish for_conf;

      delta = 1;
     testlow = objfunc;
     psilow = xr;
      countlow = 0;
      do while ( testlow < &chi_level & countlow < &countmax );
          psilow = psilow - delta;
          testlow = mainf(psilow);
          countlow = countlow + 1; 

      end;
      %if %eval(&optn_level)^= 0 %then print psilow testlow countlow;;
      /* find upper estimate */
      psihigh = xr;
      testhigh = objfunc;
      counthigh = 0;
      do while (testhigh < &chi_level & counthigh < &countmax );
        psihigh = psihigh + delta;
         testhigh = mainf(psihigh);
         counthigh = counthigh + 1;
      end;
      %if %eval(&optn_level)^= 0 %then print psihigh testhigh counthigh ;;
/* now find a better estimate */

     if (testhigh > &chi_level) & (testlow > &chi_level) then do;
       
 
        /* try a bisection method for finding bounds */

          left = xr;
          fleft = objfunc - &chi_level;

          right = psihigh;
          fright = testhigh - &chi_level ;

          middle = (left + right)/2.0;
         fmiddle = for_conf(middle);

          count = 0;
        diff = right - left ;
          do until(abs(fmiddle) < &conf_min | diff < 0.001 | count > &countmax );
            
            test = fmiddle * fleft ;
               if test < 0 then do;
                 right = middle;
                 fright = fmiddle;
            end;
               else do;
                 left = middle;
                 fleft = fmiddle;
               end;
               middle = (left + right)/2.0;
               fmiddle = for_conf(middle);
               count = count + 1;
                  
        diff = right - left ;
          end;

        psi_high = middle;
          objfunc_high = fmiddle + &chi_level;
        %if %eval(&optn_level)^= 0 %then print psi_high objfunc_high diff count;;
        psi1_high = psi_high;
/* lower bound */

          left = psilow;
          fleft = testlow - &chi_level;

          right = xr;
          fright = objfunc - &chi_level ;

          middle = (left + right)/2.0;
         fmiddle = for_conf(middle);
        
          count = 0;
        diff = right - left ;
          do until ( abs(fmiddle) < &conf_min | diff < 0.001  | count > &countmax );       
            test = fmiddle * fleft ;
               if test < 0 then do;
                 right = middle;
                 fright = fmiddle;
            end;
               else do;
                 left = middle;
                 fleft = fmiddle;
               end;
               middle = (left + right)/2.0;
               fmiddle = for_conf(middle);
            diff = right - left ;
               count = count + 1;                
          end;

        psi_low = middle;
          objfunc_low = fmiddle + &chi_level;
          %if %eval(&optn_level)^= 0 %then %do;
           print psi_low objfunc_low diff count ;
           print xr psi_low psi_high;
          %end;
          psi1_low = psi_low;
      end; 
       else do;
         print "can not find estimate of confidence interval using less than &countmax function evaluations.";
       end;
     
%end;

%if %eval(&mean_of_H) = 1  %then %do ;
    for_meanh = psi1 ;
    %do i = 2 %to &npsi;
       for_meanh = for_meanh || psi&i ;
     %end;
 
   tmp = mainf(for_meanh);
   mean_h = round(meanh,0.0001) ;
 

%end; 


 create _kk_ var {  %do i = 1 %to &npsi; psi&i   %end;    objfunc 
                    %if %eval(&mean_of_H) = 1 %then mean_h ;
                    %if %eval(&grid_search) = 1 %then %do;
                        grid_min   %do i = 1 %to &npsi;  grid_psi&i   %end;
                    %end; 
                    %if &npsi = 1 AND &conf_interval = 1 %then psi1_low psi1_high ; 
                    %if &check_func = 1 %then func ;
                    }; 

 
append ; 
quit;
%mend; 

 
%macro smooth_indicator ;
        
             smooth = 1;
              half = 1;
             band = 0;    
             store smooth half band;
 
            call nlpnms(rc,xr,"mainf",start,optn,con);  
            
               load goodflag;
                if goodflag = 1 then do;
                    start = xr;
                end;
           
                
             half = 0;
                band = 500.0;
                store half band;

             call nlpnms(rc,xr,"mainf",start,optn,con);  
             
                
                
             load goodflag;
                if goodflag = 1 then do;
                    start = xr;
                end;

                band = 100.0;
                store band;

             call nlpnms(rc,xr,"mainf",start,optn,con); 
                 
                
             load goodflag;
                if goodflag = 1 then do;
                    start = xr;
                end;
          
                band = 50.0;

                call nlpnms(rc,xr,"mainf",start,optn,con);             
             load goodflag;
                if goodflag = 1 then do;
                    start = xr;
                end;

                smooth = 0;
                half = 0;
                band = 0;
                store smooth half band ;
                
   %mend;


%macro where_list(arg1 = ,arg2 = );
   %local alpha numer all arg lp rp n new_words word len start_new
          special_char char is_number chk word_tmp arg1 arg2;
   %let alpha=ABCDEFGHIJKLMNOPQRSTUVWXYZ_ ;
   %let numer = 1234567890 ;
   %let all = &alpha.&numer ;
   %let arg = %upcase(&arg1);
   %let lp = ( ;
   %let rp = ) ;
   %let n = 1;
   %let new_words = ;
   %do %until (%qscan(&arg,%eval(&n),%str( ))=%str()); 
       %let word = %qscan(&arg,%eval(&n),%str( ));
       %let len = %length(&word);
       %let start_new = 1 ;
       %let special_char = %verify(&word,&all);
        %if %eval(&special_char) = 0 %then %do;
            %let char = %substr(&word,1,1);
            %let is_number = %verify(&char,&alpha) ;
            %if %eval(&is_number) = 0 %then %do; 
               %if &word ^=  %quote(LE) and
                      &word ^= %quote(LT) and
                      &word ^= %quote(NE) and
                      &word ^= %quote(GT) and
                      &word ^= %quote(GE) and
                      &word ^= %quote(EQ) and
                      &word ^= %quote(OR) and
                      &word ^= %quote(AND)
                   %then %let new_words = &new_words &word ;
           %end;
        %end;
        %else %do;
          %do i = 1 %to &len ;
             %let char = %substr(&word,%eval(&i),1);        
              %if &start_new = 1 %then %do;
                  %if %nrbquote(&char) = %nrbquote(&lp) or %nrbquote(&char) = %nrbquote(&rp) %then %let chk = 1 ;
                 %else %let chk = %verify(&char,&alpha);
                   %if &chk = 0 %then %do;
                       %let word_tmp = &char;
                        %let start_new = 0 ;
                 %end;
            %end; /* start = 1 */
            %else %do;
                %if %nrbquote(&char) = %nrbquote(&lp) or %nrbquote(&char) = %nrbquote(&rp) %then %let chk = 1 ;
                %else %let chk = %verify(&char,&all);
                 %if &chk = 0 %then %let word_tmp = &word_tmp.&char ;
                 %else %do;
                    %let start_new = 1;
                     %let new_words = &new_words &word_tmp ;
                       %let word_tmp = ;
                %end;
           %end; /* start = 0 */
       
       %end; /* i */
        %let new_words = &new_words &word_tmp ; /* add in last word_tmp */
        %let word_tmp = ; /* reset word_tmp */
     %end; 
     %let n = %eval(&n + 1 );
 %end; /* until */
 %let &arg2 = &new_words ;

%mend;
%macro error_code(dsn,code);
%local dsnid nobs rcc rc_code ;
%if %sysfunc(exist(est)) %then %do;
 %let dsnid = %sysfunc(open(est)) ;
 %let nobs = %sysfunc(attrn(&dsnid,nlobs)) ;
 %let rcc = %sysfunc(close(&dsnid)) ;
 
  %if %eval(&nobs) > 0 %then %do;
   data _null_;
   set est;
   call symput('rc_code',_status_);
   run;
   %let &code = %substr(&rc_code,1,1);
 %end;
 %else %let &code = -2 ; 

%end;
%else %do;
  %let &code = -1 ;
%end;
 
%mend ;



%macro default_rho;
proc iml;
start rho_d_dstar(psi,blip,dblip) ;
load    
         A   bliparg           
            %if &blip_Q = 0 %then B R pre_xi R1_R0_diff;
            %if &blip_Q = 1 %then newQ ;
             ;
          

/***************************************************************** 
 

THE CASE FOR ITT=1 ONLY WORKS WHEN USEOPT = 0. 

        for itt = 1 we have that 

           H_{K+1} = Y_{K+1} - *psi *a_1 * K   
       
WHEN ITT = 1 WE HAVE REDEFINED THE ARRAY A TO HOLD THE
CONSTANT VALUES OF A_1 FOR EACH SUBJECT. NOTHING ELSE NEEDS
TO BE CHANGED IN mainf.

***************************************************/
  
    p_time = nrow(bliparg);   
    dm1 = bliparg * t(psi) ; 
    dm = J(p_time,1,0);
     dm_deriv = J(p_time,1,0);
            
     %if &useopt = 1 %then %do;
          load smooth half band ;
             
          if smooth = 1 then do;
             if half = 1 then do;
                     dm = J(p_time,1,0.5);
                      dm_deriv = J(p_time,1,0);
             end;
             else do;
                    ttt = dm1/ band;
                    dm = probnorm(ttt);                 

                     dm0 = pdf('normal',ttt);
                     dm_deriv = ttt# dm0 + dm;
               end;
           end; /* smooth = 1 */
           else do; 
                dm = ( dm1[i] > 0)  ;                                           
           end;
             
      %end; /* useopt = 1 */
           
       blip = dm1#(dm - A );
      %if &blip_Q = 0 %then dblip = (dm - A)#bliparg ;;          
     
       finish rho_d_dstar;

       store module = rho_d_dstar;

       use _cero_;
       read all var { &treat} into A;
       ones = J(nrow(A),1,1);
       %if &npsi > 1 %then %do;
          read all var { &psivar } into preblip; 
               bliparg = ones || preblip ;
      %end;
       %else %do;
            bliparg = ones;
       %end;

       store A bliparg ;

       quit;

  %mend default_rho;
/**************/
%macro analytic_sol;  /* #### */
/* this approach is only valid for
     DR = 1, NEW_PARTS = 0 , blip_Q = 1, mult = 0 
     and user_rho = 1*/

data _cero_;
     set _cero_;
     if &outcome ^= . ;
run;
proc freq data= _cero_ noprint;
table &id / out=_tcount_(drop=percent);
run;
proc means data=_tcount_ n noprint; 
var count;
output out=_tmax n = count;
run;

data _tmax;
set _tmax end = _end_;;
if _end_ then do;
  call symput('individuals',count);
end;
run;

proc iml ; * worksize=5000000000;
     use _cero_;
     read all var { &outcome } into Y;
     %if &npsi > 1 OR &intercept = 0 %then %do;   
          read all var { &psivar } into preblip; 
     %end;
     read all var { &treat}  into A ;
     %if %eval(&dr) = 1 & %bquote(&covEH)^= %then read all var {&covEH } into W ;; /* model for H */
     read all  var { _pA_t };

     p_time = nrow(A);
     ones = J(p_time,1,1);

     %if %bquote(&user_regimes) = 1 %then %do;
          %let regime_tmp = %scan(&user_defined_regimes,%eval(&regime),%str(' '));
          read all var { &regime_tmp} into dm;
     %end;
     %else dm = J(p_time,1,0);;

     %if &intercept =1 %then %do;  
          %if &npsi > 1 %then %do;   
               bliparg = ones || preblip ;
          %end;
          %else %do;
               bliparg = ones ;
          %end;
     %end;
     %else %if &intercept = 0 %then bliparg = preblip; ;  

     residA = A - _pA_t;  

     use _tcount_;
     read all var {count} into tcount;
     %if %bquote(&cens)^= %then %do;
          use _cero_;
          read all var {pC_t} into pC1 ;
          %if %bquote(&cens2)^= %then read all var {pC2_t} into pC2 ;;
          %if %bquote(&cens3)^= %then read all var {pC3_t} into pC3 ;;
          pC_t = J(p_time,%eval(&ncens_var),0);
          %do i = 1 %to %eval(&ncens_var) ;
               pC_t[,&i] = pC&i ;
          %end;
          ipcw_all = J(p_time,%eval(&ncens_var),0);
          id = %sysevalf(&individuals) + 1;
          i = p_time;
          id_count = 0;
          count = 0; 
          do while ( i >= 1);
          if id_count = count then do;   
                    id_count = 1;
                    id = id - 1;
                    count = tcount[id];
                    tmp = pC_t[i,];
          end;
          else do;
                    id_count = id_count + 1;
                    tmp = tmp # pC_t[i,];
          end;
               ipcw_all[i,] = 1.0/tmp;
          i = i - 1;
          end;  
          ipcw = ipcw_all[,1];
          %do j = 2 %to %eval(&ncens_var);
          ipcw = ipcw # ipcw_all[,&j] ;
          %end; 
     %end;
     %else %do;
          ipcw = J(p_time,1,1);    
     %end;
     
     %if %bquote(&weight)^= %then ipcw = ipcw # weight ;;

     %if &blip_Q = 1 %then newQ = bliparg ;;
     %if %eval(&dr) = 0 %then %do;
          nc = ncol(bliparg);
          nr = nrow(bliparg);
          lhs = J(nc,nc,0);
          rhs = J(1,nc,0);
        
          blip = (dm - A) # bliparg ;
          %if %eval(&mean_of_H) = 1 %then B = J(p_time,&npsi,0); ;

          N = nrow(tcount);
          row = nr ;
          do id = N to 1 by -1;
              count = tcount[id];
              holder = J(1,nc,0);
   
              do i = 1 to count;
                    AQ_m = blip[row,] ;
                    Q_m = bliparg[row,] ;
                    holder = holder + AQ_m; /* sum[j = m to K] AQ  created in reverse order */
                    tmp_lhs = residA[row] * ipcw[row];
                    tmp_rhs  = Y[row] * tmp_lhs ;
                    tmp2 = tmp_lhs * holder ;
                    
                    lhs = lhs + t(Q_m) * tmp2 ;                    
                    rhs = rhs + tmp_rhs * Q_m ;

                    %if %eval(&mean_of_H) = 1 %then B[row,] =  holder ;; 

                    row = row - 1;
              end;
          end;

          psi = inv(lhs) * t( -1*rhs) ;          
     %end;
     %else %if %eval(&dr) = 1 %then %do;       
          %if %bquote(&covEH)^= %then W = ones||W;  
          %else W = ones ;;
          wtw = t(W)*(W # ipcw);   
          wtw_inv = inv(wtw);   
          pre_beta = wtw_inv  * t(W # ipcw) ;     
     

    

          bliparg_dif = (dm - A) # bliparg;   
          B = J(p_time,&npsi,0); 
          last = 0;

          do id = 1 to %eval(&individuals);   
                    count = tcount[id];
                    start = last + 1;
                    last =  start + count - 1;
                    Btmp = J(count, &npsi, 0);
                    do t = 0 to count-1;
                         j = last - t;
                         k = count - t;
                         dif_tmp = bliparg_dif[j:last, ]; 
                         dif_sum =  dif_tmp[+, ];
                         Btmp[k, ] = dif_sum;   
                    end;
                    B[start:last, ] = Btmp;
          end; /* id */
     
          i_resA = residA # ipcw;
          
          aux_B = pre_beta * B;
          aux_Y = pre_beta * Y;
          QB =  t(newQ) * ((B - W * aux_B) # i_resA );
          inv_QB =  inv(QB);
          psi = - inv_QB * t(newQ) * ((Y - W * aux_Y) # i_resA) ; 
     

     %end;

     %if %eval(&mean_of_H) = 1 %then %do ;
          H = Y + B * psi;
          sumH = 0;
          start = 1;
          do id = 1 to %eval(&individuals) - 1;   
               sumH = sumH + H[start];
               count = tcount[id];
               start = start + count;
          end;
          mean_h = sumH/&individuals;
     %end;

     %do i = 1 %to &npsi;
          psi&i = psi[&i];
     %end;

      

 create _kk_ var {  %do i = 1 %to &npsi; psi&i   %end;    
                    %if %eval(&mean_of_H) = 1 %then mean_h ;
                     
                    
                    }; 

 
append ; 

  

quit;

%mend analytic_sol;
/*****/
%macro ipcw;
data _neww_;
          set _cero_ end = _end_ ;
          by &id;
          %if %bquote(&k_step)^= %then _copyk = %eval(&k_step);;

          if first.&id then  do;
               _newind = 0; 
               _copy = count;          
          end;
          else do;
               _copy = _copy - 1;         
          end;
          %if %bquote(&eligible)^= %then  if &eligible = 1 then do;;
          %if %bquote(&k_step)^= %then %do;
               if _copyk > _copy then _copy_count = _copy;
               else _copy_count = _copyk;
          %end;
          %else %do;
               _copy_count = _copy;
          %end;
          do _i = 1 to _copy_count;
             _newind = _newind + 1;
              output;
          end;
          %if %bquote(&eligible)^= %then end;;
          retain _copy _newind ;
          if _end_ then  do; 
               call symput('_LengthW_',_N_); 
          end;
          run;

   
proc iml ;

%if %bquote(&Cens)^= %then %do;

    use _cero_;
    
     read all var {pC_t} into pC1 ;
     %if %bquote(&cens2)^= %then read all var {pC2_t} into pC2 ;;
     %if %bquote(&cens3)^= %then read all var {pC3_t} into pC3 ;;
   
     pC_t = J(nrow(pC1),%eval(&ncens_var),0);
     %do i = 1 %to %eval(&ncens_var) ;
          pC_t[,&i] = pC&i ;
     %end;


     close _cero_;

    use _tcount_;
        read all var {count} into tcount;
        close _tcount_;

   
    last = 0;
    row = 0;
    e_row = 0;

    use _neww_ ;
       read all var { in_eeq } ;
       close _neww_ ;

    nw = nrow(in_eeq) ;

    ipcw_all = J(nw,%eval(&ncens_var),0);

    do id = 1 to %sysevalf(&individuals);

          count = tcount[id];
          start = last + 1;
          last = start + count -1;

          pC_id = pC_t[start:last,];
          do m=1 to count;
               tmp = 1.0;
               e_row = e_row + 1;
               %if %bquote(&eligible)^=_eligible_ %then %do; 
                 if eligible[e_row] = 1 then do; %end;
                    c_tmp = count;
                    %if % bquote(&k_step)^= %then c_tmp = m + min(%eval(&k_step),count-m + 1) - 1 ; ;              
                    do k = m  to  c_tmp /* count */;
                         tmp = tmp # pC_id[k,];
                         row = row + 1;
                         ipcw_all[row,] = 1.0/tmp;
                    end;
               %if %bquote(&eligible)^=_eligible_ %then %do; end; /* end of eligible = 1 */ %end;
          end; /* m */
             
    end;

    ipcw = ipcw_all[,1];
    %do j = 2 %to %eval(&ncens_var);
          ipcw = ipcw # ipcw_all[,&j] ;
    %end;
%end;
%else %do;
   ipcw = J(nw,1,1.0);  /** if no censoring then set weights to be 1 */
%end;


/*ipcw = in_eeq # ipcw ; /* only want those who have non-missing Y_{k+1} */

quit;

data _ipcw_ft;
    merge _neww_ _ipcw_ft;  
    run;

data _ipcw_;
    set _ipcw_ft;
    by uniqueid &time;
    if last.&time | last.uniqueid;
    run;

%mend;

/*******/
/******/
%macro remove_interactions ;
     %local old_options ;
     %let old_options = %sysfunc(getoption(mprint)) %sysfunc(getoption(SYMBOLGEN)) %sysfunc(getoption(MLOGIC)) %sysfunc(getoption(NOTES)) ;
     options nomprint nomlogic nosymbolgen nonotes ;
  %local i var_list cov_list name _cov _inter non_int inter_tmp  ;

     %let cov_list = %nrstr(&covA_tmp  &covC_tmp  &covC2_tmp  &covC3_tmp &covEH &covVH &covDH);
     %let var_list = A_tmp C_tmp C2_tmp C3_tmp EH VH DH;
     

     %do i = 1 %to 7 ;
          
          %let name = %qscan(&var_list,%eval(&i),%str( ));
          %let _cov = %qscan(&cov_list,%eval(&i),%str( ));
         /*  %let _inter = %qscan(&inter_list,%eval(&i),%str( ));*/
          %let non_int = ;
          %let inter_tmp = ;
          %remove_interactions0(list1 = %unquote(&_cov),
                                list2 = inter_tmp ,
                                list3 = non_int );
          %let cov&name = &non_int ;
          %let inter&name = &inter_tmp ;
         
     %end;
     options &old_options ;
%mend ;

/******/
%macro remove_interactions0(list1 = original ,list2 = interactions , list3 = non_interactions)  ;
     %local i list2_tmp list3_tmp nl1 word int ;
     /* initialize output lists */
     %let list2_tmp = ; 
     %let list3_tmp  = ;
     %let nl1 = %numargs(&list1);

     %do i = 1 %to %eval(&nl1);
          %let word = %qscan(&list1,%eval(&i),%str( ));
          %let word = %cmpres(&word);
          %let ind = %index(&word,%str(*));
          %if %eval(&ind) = 0 %then %let list3_tmp  = &list3_tmp &word ;
          %else %let list2_tmp = &list2_tmp &word ;
      %end;
 
      %let &list2 = &list2_tmp;
      %let &list3 = &list3_tmp;

   

%mend;


/******/
/***/
%macro create_interactions ;

     %local old_options ;
     %let old_options = %sysfunc(getoption(mprint)) %sysfunc(getoption(symbolgen)) %sysfunc(getoption(mlogic)) %sysfunc(getoption(notes)) ;
     options nomprint nomlogic nosymbolgen nonotes ;

     %local i var_list cov_list name _cov _inter non_int inter_tmp already_done ;
     %local   products0 product_defs class_vars nprod0 nprod ;
     %local ininter nclass newlist_tmp newclass_tmp products  nprod0 prod ind word1 word2 word1_class;
     %local word2_class vartmp word1_levels word2_levels i levels levels1 levels2;
     %local var_index ;

     %let var_list = EH VH  DH   ;
     %let class_vars = &classH &classVH  &classDH ;
     %let cov_list = %nrstr( &covEH &covVH &covDH   );

     %let product_defs = ;
     %let already_done = ;
    
     %let nprod = 0;
     
     %let newclass = ;
     %let clist = &class_vars ;   

     %let nclass = %numargs(&clist);

     %do var_index = 1 %to 3 ;
          %let name = %scan(&var_list,%eval(&var_index),%str( ));
          %let _cov = %qscan(&cov_list,%eval(&var_index),%str( ));        
          %let inter = &&inter&name;
          %let ninter = %numargs(&inter);
          %let newlist_tmp = ;
          %let newclass_tmp = ;
          %let products = ;
          %let nprod0 = 0;

          %do i = 1 %to %eval(&ninter);
               %let prod = %cmpres(%qscan(&inter,%eval(&i),%str( )));
               %let ind = %index(&prod,%str(*));
               %let word1 = %qsubstr(&prod,1,%eval(&ind - 1));
               %let word2 = %qsubstr(&prod,%eval(&ind + 1));
               %let word1_class = 0;
               %let word2_class = 0;
               %do j = 1 %to %eval(&nclass);
                    %let vartmp = %scan(&clist,%eval(&j),%str( ));
                    %if &word1 = &vartmp %then %let word1_class = 1;
                    %if &word2 = &vartmp %then %let word2_class = 1 ;
               %end;
               /* for non-class variables we will indicate this with 1 level */
               %let word1_levels = 1;
               %let word2_levels = 1;
         
         
               %if %eval(&word1_class) = 0 AND %eval(&word2_class) = 0 %then %do;   
                    %let prod_tmp = _&word1.&word2 ;               
                    %let in_list = %check_list(&word1,&word2,&already_done);                
                    %if %eval(&in_list) = 0 %then %do;
                         %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                         %let nprod0 = %eval(&nprod0 + 1);
                         %let products =  &products %cmpres(_&word1.&word2) %cmpres(&word1) %cmpres(&word2)  ;
                         %let already_done = &already_done %cmpres(_&word1.&word2 ) ;
                    %end;
                    %else %do;
                         %if %eval(&in_list) = 1 %then %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                         %else %let newlist_tmp = &newlist_tmp %cmpres(_&word2.&word1) ;
                    %end; 
               %end;          
               %else %if %eval(&word1_class) = 1 | %eval(&word2_class) = 1 %then %do;
                    data _null_;
                    set _class_levels;
                    %if %eval(&word1_class) = 1 %then call symput('word1_levels',&word1);;
                    %if %eval(&word2_class) = 1 %then call symput('word2_levels',&word2);;
                    run;
                    %if %eval(&word1_levels) > 1 AND %eval(&word2_levels) = 1 %then %do;
                         %do levels = 1 %to %eval(&word1_levels - 1) ;
                              %let prod_tmp = _&word1.&levels.&word2 ;
                                  /* %let newlist_tmp = &newlist_tmp  %cmpres(&prod_tmp); */
                              %let in_list = %check_list(&word1.&levels, &word2,&already_done) ;
                              %if %eval(&in_list) = 0 %then %do;
                                   %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                                   %let nprod0 = %eval(&nprod0 + 1 );                                                
                                   %let products =  &products %cmpres(&prod_tmp)  %cmpres(_&word1.&levels) %cmpres(&word2) ;
                                   %let already_done = &already_done %cmpres(&prod_tmp) ;
                              %end;
                              %else %do;
                                   %if %eval(&in_list)= 1 %then %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                                   %else %let newlist_tmp = &newlist_tmp %cmpres(_&word2.&word1.&levels) ;
                              %end;
                         %end;
                   %end; 
                   %else %if %eval(&word1_levels) = 1 AND %eval(&word2_levels) > 1 %then %do;
                         %do levels = 1 %to %eval(&word2_levels - 1) ;
                              %let prod_tmp = _&word1.&word2.&levels ;
                           /*   %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp); */
                              %let in_list = %check_list(&word1, &word2.&levels ,&already_done) ;
                              %if %eval(&in_list) = 0 %then %do;
                                   %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                                   %let nprod0 = %eval(&nprod0 + 1 );                        
                                   %let products =  &products %cmpres(&prod_tmp) %cmpres(&word1) %cmpres(_&word2.&levels) ;
                                   %let already_done = &already_done %cmpres(&prod_tmp) ;
                              %end;
                              %else %do;
                                   %if %eval(&in_list) = 1 %then %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                                   %else %let newlist_tmp = &newlist_tmp  %cmpres(_&word2.&levels.&word1);
                              %end;
                         %end;
                    %end;
                    %else %if %eval(&word1_levels) > 1 AND %eval(&word2_levels) > 1 %then %do;
                         %do levels1 = 1 %to %eval(&word1_levels - 1) ;
                              %do levels2 = 1 %to %eval(&word2_levels - 1) ;
                                   %let prod_tmp = _&word1.&levels1.&word2.&levels2 ;
                                 /*  %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);*/
                                   %let in_list = %check_list(&word1.&levels1, &word2.&levels2 ,&already_done) ;
                                   %if %eval(&in_list) = 0 %then %do;
                                        %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                                        %let newclass_tmp = &newclass_tmp  %cmpres(&prod_tmp);
                                        %let nprod0 = %eval(&nprod0 + 1) ;                               
                                        %let products =  &products %cmpres(&prod_tmp) %cmpres(_&word1.&levels1)  %cmpres(_&word2.&levels2)  ;                               
                                        %let already_done = &already_done %cmpres(&prod_tmp) ;
                                  %end;
                                  %else %do;
                                        %if %eval(&in_list) =1 %then %let newlist_tmp = &newlist_tmp %cmpres(&prod_tmp);
                                        %else %let newlist_tmp = &newlist_tmp %cmpres(_&word2.&levsls2.&word1.&levels1);
                                  %end; 
                              %end;
                         %end;
                    %end;
               %end; 
           
          %end;  /* done all interactions in one model */
    
 
      
          %let cov&name = &&cov&name &newlist_tmp ;
          %let newclass = &newclass &newclass_tmp ;
          %let product_defs =  &product_defs &products ;
          %let nprod = %eval(&nprod + &nprod0) ;
 
     
 
     %end; /* var index for interaction in each model */
 
 
     %let nn = %eval(3 * &nprod) ;
     data _cero_;
     set _cero_ ;
     %if %bquote(&newclass)^= %then   length &newclass 3 ;;
     %do i = 1 %to %eval(3 * &nprod) %by 3;
          %let lhs = %scan(&product_defs,%eval(&i),%str( ));
          %let rhs1 = %scan(&product_defs,%eval(&i + 1),%str( ));
          %let rhs2 = %scan(&product_defs,%eval(&i + 2),%str( ));
          &lhs  = &rhs1  *  &rhs2 ;
     %end;
     run;

     proc datasets library = work nolist;
     delete _class_levels ;
     quit;
          
     options &old_options ;
%mend ;
/******/
/****/


/*****/
%macro check_list(word1,word2,list);
    %local test_word in_list ; 
    %let prod1 = %cmpres(_&word1&word2);
    %let prod2 = %cmpres(_&word2&word1);
    %let in_list = 0;
    
    %let n = 1;
     %if %bquote(&list)^= %then %do;
          %do %until (%qscan(&list,%eval(&n),%str( ))=%str() OR (%eval(&in_list) = 1));
           
               %let test_word = %qscan(&list,&n,%str( ));
               %if &prod1 = &test_word  %then %let in_list = 1;
               %else %if  &prod2 = &test_word %then %let in_list = 2 ;               
               %let n = %eval(&n + 1 );
          %end;

     %end;
     %eval(&in_list) /* there is no ; here since it will be used as %let a = %check_list(&word,&list) ;
                      and the ; is included at the end of this line  */  
%mend ; 

